unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, DBGrids, ExtCtrls, DBCtrls, Db, SqlitePassDbo, StdCtrls, Buttons,
  SqlitePassUtils, ComCtrls;

type
  TForm1 = class(TForm)
    DatasetCustomers: TSqlitePassDataset;
    DsCustomers: TDataSource;
    DBNavigator1: TDBNavigator;
    DBGrid1: TDBGrid;
    DatasetOrders: TSqlitePassDataset;
    DsOrders: TDataSource;
    PanelDatasetToolbar: TPanel;
    Bevel13: TBevel;
    SbDatasetShowSortDialog: TSpeedButton;
    SbDatasetFilterOnSelection: TSpeedButton;
    SbDatasetShowFilterDialog: TSpeedButton;
    Bevel14: TBevel;
    SbDatasetLocatePrior: TSpeedButton;
    Bevel15: TBevel;
    SbDatasetShowLocateDialog: TSpeedButton;
    Bevel16: TBevel;
    Bevel17: TBevel;
    SbDatasetFilterOnOff: TSpeedButton;
    SbDatasetLocateFirst: TSpeedButton;
    SbDatasetLocateLast: TSpeedButton;
    SbDatasetLocateNext: TSpeedButton;
    Bevel18: TBevel;
    LabelLocateRecordCount: TLabel;
    SbDatasetRefresh: TSpeedButton;
    Bevel19: TBevel;
    Panel1: TPanel;
    Splitter1: TSplitter;
    DBLookupListBox1: TDBLookupListBox;
    DBLookupComboBox1: TDBLookupComboBox;
    SbCleanDatabase: TSpeedButton;
    SqlitePassDb: TSqlitePassDatabase;
    SbOpenDatabase: TSpeedButton;
    Label1: TLabel;
    Label2: TLabel;
    EditCustomersRecordCount: TEdit;
    Label3: TLabel;
    EditOrdersRecordCount: TEdit;
    Label4: TLabel;
    Label5: TLabel;
    SpeedButton1: TSpeedButton;
    CheckBoxUseTransaction: TCheckBox;
    LabelEllapsedTime: TLabel;
    DatasetCustomersCustomer_Id: TIntegerField;
    DatasetCustomersCustomer_LastName: TStringField;
    DatasetCustomersCustomer_FirstName: TStringField;
    DatasetOrdersOrder_Id: TIntegerField;
    DatasetOrdersCustomer_Id: TIntegerField;
    DatasetOrdersOrder_Description: TStringField;
    ProgressBar: TProgressBar;
    DBGrid2: TDBGrid;
    DatasetOrdersCustomerName: TStringField;
    procedure SbCleanDatabaseClick(Sender: TObject);
    procedure DatasetCustomersGetFieldData(Field: TField;
      const NativeFieldData: String; FieldDataBuffer: Pointer;
      var Handled: Boolean);
    procedure SbOpenDatabaseClick(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure DBGrid1TitleClick(Column: TColumn);
  private
    procedure CreateDatabase(Sender: TObject);
  public
    { Dclarations publiques }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.CreateDatabase(Sender: TObject);
const
CustomersTable =   'CREATE TABLE "Customers" ("Customer_Id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,'
                 + '"Customer_LastName" TEXT(50), "Customer_FirstName" TEXT(50));';
OrdersTable =      'CREATE TABLE "Orders" ("Order_Id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,'
                 + '"Customer_Id" INTEGER,"Order_Description" TEXT(50));';
begin
With SqlitePassDb do
     begin
     Close;
     Database := GetExeDir + 'LookUpDatabase.db3';
     if Not FileExists(Database) then
        begin
        CreateDatabase('LookUpDatabase.db3', dbtSqliteExpert);
        TableDefs.CreateTable(CustomersTable);
        TableDefs.CreateTable(OrdersTable);
        end;
     Open;
     TableDefs.EmptyTable('Customers');
     TableDefs.EmptyTable('Orders');
     Compact;
     SbOpenDatabaseClick(Self);
     end;
end;

procedure TForm1.SbCleanDatabaseClick(Sender: TObject);
begin
CreateDatabase(Self);
end;

procedure TForm1.DatasetCustomersGetFieldData(Field: TField;
  const NativeFieldData: String; FieldDataBuffer: Pointer;
  var Handled: Boolean);
Var
s: String;
begin
s := NativeFieldData;
end;

procedure TForm1.SbOpenDatabaseClick(Sender: TObject);
begin
 SqlitePassDb.Open;
// DatasetCustomers.DatasetName := 'Customers';
 DatasetCustomers.Open;
// DatasetOrders.DatasetName := 'Orders';
 DatasetOrders.Open;

end;

procedure TForm1.SpeedButton1Click(Sender: TObject);
var
i, j, CustomersRecordCount, OrdersRecordCount: Integer;
begin
CustomersRecordCount := StrToIntDef(EditCustomersRecordCount.Text,0);
OrdersRecordCount := StrToIntDef(EditOrdersRecordCount.Text,0);
try
  DatasetOrders.DisableControls;
  DatasetCustomers.DisableControls;
  ProgressBar.Max := CustomersRecordCount;
  SqlitePassDb.Engine.StartChrono;
  if CheckBoxUseTransaction.Checked
     then SqlitePassDb.Engine.Transaction.Start;

  For i := 1 to CustomersRecordCount do
      begin
      { Optional - you can comment this line,
        then the Customer_Id will be set automatically }
      DatasetCustomersCustomer_Id.Value := i;
      DatasetCustomersCustomer_LastName.Value := 'Customer last name = Number ' + IntToStr(i);
      DatasetCustomersCustomer_FirstName.Value := 'Customer first name = Number ' + IntToStr(i);
      DatasetCustomers.Append;
      For j := 1 to OrdersRecordCount do
          begin
          DatasetOrdersCustomer_Id.Value := i;
          DatasetOrdersOrder_Description.Value := 'Description for orders ' + IntToStr(j) + ' from customer ' + DatasetOrdersCustomer_Id.AsString;
          DatasetOrders.Append;
          end;
      ProgressBar.Position := i;
      end;
finally
  SqlitePassDb.Engine.Transaction.Commit;
  SqlitePassDb.Engine.StopChrono;
  LabelEllapsedTime.Caption := 'Ellapsed time : ' + FormatFloat('0.0000', SqlitePassDb.Engine.ElapsedTime);
  DatasetCustomers.EnableControls;
  DatasetOrders.EnableControls;
end;

end;

procedure TForm1.DBGrid1TitleClick(Column: TColumn);
var
i, SortPos: Integer;
SortDirection: String;

begin
Try
DatasetCustomers.DisableControls;

For i := 0 to Pred(DBGrid1.Columns.Count)
    do DBGrid1.Columns[i].Title.Font.Color := clBlack;

SortDirection := ' ASC';
SortPos := Pos('<Asc> ', Column.Title.Caption);
if SortPos > 0
   then SortDirection := ' DESC'
   else begin
        SortPos := Pos('<Desc> ', Column.Title.Caption);
        if SortPos > 0 then SortDirection := ' ASC';
        end;
Column.Title.Caption := StringReplace(Column.Title.Caption, '<Asc> ', '', []);
Column.Title.Caption := StringReplace(Column.Title.Caption, '<Desc> ', '', []);
if SortDirection = ' ASC'
   then Column.Title.Caption := '<Asc> ' + Column.Title.Caption
   else Column.Title.Caption := '<Desc> ' + Column.Title.Caption;

For i := 0 to Pred(DBGrid1.Columns.Count)
    do if DBGrid1.Columns[i] <> Column then
          begin
          DBGrid1.Columns[i].Title.Caption := StringReplace(DBGrid1.Columns[i].Title.Caption, '<Asc> ', '', []);
          DBGrid1.Columns[i].Title.Caption := StringReplace(DBGrid1.Columns[i].Title.Caption, '<Desc> ', '', []);
          end;

Column.Title.Font.Color := clRed;
DatasetCustomers.SortedBy := '"' + Column.FieldName + '" ' + SortDirection;
DatasetCustomers.Sort;
finally
DatasetCustomers.EnableControls;
end;
end;

end.
