{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2007-2008
    Last update : 01/06/2008

  --------------------------------------------------------------------------- }

unit SqlitePassChooseDatasetDialog;
{$IFDEF FPC}
 {$mode DELPHI}
{$ENDIF}

interface

uses
 {$IFDEF FPC}
  LResources,
 {$ENDIF}
 {$IFDEF WIN32}
  Windows,
 {$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, DB, ComCtrls, SqlitePassDbo, ImgList;

type

  { TSqlitePassChooseDatasetDialog }

  TSqlitePassChooseDatasetDialog = class(TForm)
    PanelBottom: TPanel;
    BtOk: TButton;
    BtCancel: TButton;
    PanelTop: TPanel;
    LabelAvailableDatasets: TLabel;
    Tv: TTreeView;
    TreeViewImageList: TImageList;

    procedure TvDblClick(Sender: TObject);
   Public
   DatabaseNode, TablesNode, QueriesNode, ViewsNode, TriggersNode, SystemTablesNode,
    SQLNode, SQLCreateNode, SQLCreateTableNode, SQLSelectNode, SQLInsertNode,
    SQLUpdateNode, SQLDeleteNode, SQLOthersNode: TTreeNode;
   constructor Create(AOwner: TComponent; Dataset: TSqlitePassDataset); reintroduce;
  end;

var
  SqlitePassChooseDatasetDlg: TSqlitePassChooseDatasetDialog;

implementation

{$IFNDEF FPC}
 {$R *.DFM}
{$ENDIF}

constructor TSqlitePassChooseDatasetDialog.Create(AOwner: TComponent;
  Dataset: TSqlitePassDataset);
const
EmptyImageIndex = -1;
SelectedImageIndex = 1;
DatabaseImageIndex = 0;
TableImageIndex = 2;
QueryImageIndex = 3;
WiewImageIndex = 4;
TriggerImageIndex = 5;
SQLImageIndex = 6;
SQLItemImageIndex = 7;
SystemTableImageIndex = 8;
ItemImageIndex = 9;
AttachedItemImageIndex = 11;

var
i: Integer;
TvItem: TTreeNode;
CurrentSQLNode: TTreeNode;
Db: TSqlitePassDatabase;

 function AddNode(ParentNode: TTreeNode; Caption: String; ImageIndex, SelectedImageIndex: Integer): TTreeNode;
 begin
 if Assigned(ParentNode)
     then Result := Tv.Items.AddChild(ParentNode,Caption)
     else Result := Tv.Items.Add(nil,'Database');
 Result.ImageIndex := ImageIndex;
 Result.SelectedIndex := SelectedImageIndex;
 end;

 function GetItemImageIndex(Attached: Boolean): Integer;
 begin
  If Attached
     then Result := AttachedItemImageIndex
     else Result := ItemImageIndex;
 end;

 function GetSQLNode(var Node: TTreeNode; NodeText: String): TTreeNode;
 begin
 if Not Assigned(Node)
    then Node := AddNode(SQLNode, NodeText,SQLItemImageIndex,SelectedImageIndex);
 Result := Node;
 end;

begin
{ -------------------------------------------------------------- }
inherited Create(AOwner);
if Dataset = nil then exit;
Db := Dataset.Database;
if Db = nil then exit;
{ -------------------------------------------------------------- }

Try
Tv.Items.BeginUpdate;

Tv.Items.Clear;
DatabaseNode := nil;
TablesNode   := nil;
QueriesNode  := nil;
ViewsNode    := nil;
TriggersNode := nil;
SQLNode      := nil;
SQLCreateNode := nil;
SQLCreateTableNode := nil;
SQLSelectNode := nil;
SQLInsertNode := nil;
SQLUpdateNode := nil;
SQLDeleteNode := nil;
SQLOthersNode := nil;
CurrentSQLNode := nil;

DatabaseNode := AddNode(nil,'Database',DatabaseImageIndex,DatabaseImageIndex);
TablesNode   := AddNode(DatabaseNode,'Tables',TableImageIndex,TableImageIndex);
QueriesNode  := AddNode(DatabaseNode,'Queries',QueryImageIndex,QueryImageIndex);
ViewsNode    := AddNode(DatabaseNode,'Views',WiewImageIndex,WiewImageIndex);
TriggersNode := AddNode(DatabaseNode,'Triggers',TriggerImageIndex,TriggerImageIndex);
SQLNode := AddNode(DatabaseNode,'SQL',SQLImageIndex,SQLImageIndex);
SystemTablesNode := AddNode(DatabaseNode,'System Tables',SystemTableImageIndex,SystemTableImageIndex);

for i:= 0 to Pred(Db.TableDefs.Count)
    do if Db.IsSystemTable(Db.TableDefs[i].TableName)
       then AddNode(SystemTablesNode, Db.TableDefs[i].TableFullName,GetItemImageIndex(Db.TableDefs[i].Attached),SelectedImageIndex)
       else AddNode(TablesNode,Db.TableDefs[i].TableFullName,GetItemImageIndex(Db.TableDefs[i].Attached),SelectedImageIndex);

for i:= 0 to Pred(Db.QueryDefs.Count)
    do AddNode(QueriesNode,Db.QueryDefs[i].QueryFullName,GetItemImageIndex(Db.QueryDefs[i].Attached), SelectedImageIndex);

for i:= 0 to Pred(Db.Views.Count)
    do AddNode(ViewsNode,Db.Views[i].ViewFullName,GetItemImageIndex(Db.Views[i].Attached), SelectedImageIndex);

for i:= 0 to Pred(Db.Triggers.Count)
    do AddNode(TriggersNode, Db.Triggers[i].TriggerFullName,GetItemImageIndex(Db.Triggers[i].Attached), SelectedImageIndex);

for i:= 0 to Pred(Db.SQLStmtDefs.Count)
    do begin
       case Db.SQLStmtDefs[i].SQLType of
            stSelect     : CurrentSQLNode := GetSQLNode(SQLSelectNode, '(Select)');
            stCreate     : CurrentSQLNode := GetSQLNode(SQLCreateNode, '(Create)');
            stCreateTable: CurrentSQLNode := GetSQLNode(SQLCreateTableNode, '(Create Table)');
            stInsert     : CurrentSQLNode := GetSQLNode(SQLInsertNode, '(Insert');
            stUpdate     : CurrentSQLNode := GetSQLNode(SQLUpdateNode, '(Update)');
            stDelete     : CurrentSQLNode := GetSQLNode(SQLDeleteNode, '(Delete)');
            else CurrentSQLNode := GetSQLNode(SQLOthersNode, '(Others)');
            end; { Case }
       //AddNode(CurrentSQLNode,Db.SQLStmts[i].StmtName,GetItemImageIndex(Db.SQLStmts[i].Attached),SelectedImageIndex);
       AddNode(CurrentSQLNode,Db.SQLStmtDefs[i].SQLStmtName,ItemImageIndex,SelectedImageIndex);
       end;

finally
Tv.FullExpand;
if Tv.Items.Count > 0
   then Tv.TopItem := DatabaseNode;
Tv.Items.EndUpdate;
end;
end;

procedure TSqlitePassChooseDatasetDialog.TvDblClick(Sender: TObject);
begin
ModalResult := mrOk;
end;

initialization
 {$IFDEF FPC}
  {$I SqlitePassSortByDialog.lrs}
 {$ENDIF}
end.
