{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2006 - 2009

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 05.01.2009

  --------------------------------------------------------------------------- }




{ TSqlitePassDatasetVersionInfo }

constructor TSqlitePassDatasetVersionInfo.Create;
begin
inherited Create;
FComponent := SqlitePassDatasetVersion;
FPackage := SqlitePassPackageVersion;
end;

procedure TSqlitePassDatasetVersionInfo.SetFComponent(
  const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;

procedure TSqlitePassDatasetVersionInfo.SetFPackage(const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;


{ --- TSqlitePassDataset --- }

constructor TSqlitePassDataset.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFNDEF FPC}
  { ObjectView is set to True to speed up Fields Binding }
  ObjectView := True;
  {$ENDIF}
  if Assigned(AOwner) and (AOwner is TSqlitePassDatabase)
     then FDatabase := TSqlitePassDatabase(AOwner)
     else FDatabase := nil;
  FDatasetName := '';
  FDatasetType := dtUnknown;
  FInInternalOpen := False;
  FReadOnly := False;
  FWriteMode := wmDirect;
  FRecordsCacheCapacity := 100;
  FRecordset := TSqlitePassRecordset.Create(Self);
  { Locate }
  FLocateSmartRefresh := False;
  FLocateRecordset := TSqlitePassRecordset.Create(Self);
  FLocateFoundRecords := TList.Create;
  { LookUp }
  FLookUpCache := False;
  FLookUpMaxRecordCount := 0;
  FLookUpOptions := [];
  FLookUpSmartRefresh := False;
  FLookUpRecordset := TSqlitePassRecordset.Create(Self);
  FLookUpResultFields := TList.Create;
  { Sorting }
  FSortCalcFields := False;
  FSortMode := smQuickSort;
  FSortedFields := TList.Create;
  SetFSortedBy('');
  { }
  FProcessMessages := False;

  FSQLChanges := [];
  FSQL := TStringList.Create;
  FSQL.Clear;
  FSQLSelectStmt := TSqlitePassSelectStmt.Create(Self);
  FExportMethod := VisibleFields;
  FIndexDefs := TSqlitePassDatasetIndexDefs.Create(Self);
  FVersionInfo := TSqlitePassDatasetVersionInfo.Create;
  FDatabaseAutoActivate := True;
  FParams := TParams.Create(Self);
  { Filtering }
  FOnFilterRecord := nil;
  FDetailFields := TList.Create;
  FMasterLink := TMasterDataLink.Create(Self);
  FMasterLink.OnMasterChange := MasterChanged;
  FMasterLink.OnMasterDisable := MasterDisabled;
  FFilterRecordLowerLimit := 0;
  FFilterRecordUpperLimit := 0;
  FMasterAutoActivate   := True;
  { BookmarkSize is used by TDataset }
  BookmarkSize := SizeOf(Integer);
  FRefetchRows := False;
  { Ensure correct info to FSQLStmt with FSQL.assign or gettext/settext }
  FSQL.OnChange := UpdateFSQL;
  FOnExportProgress := nil;
  FOnImportProgress := nil;
End;

destructor TSqlitePassDataset.Destroy;
Begin
  Close;

  { Force Database.Datasets to remove this dataset from the datasets list }
  Database := nil;

  FVersionInfo.Free;
  FIndexDefs.Free;
  FParams.Free;
  OnFilterRecord := nil;
  FMasterLink.OnMasterChange := nil;
  FMasterLink.OnMasterDisable := nil;
  FMasterLink.Free;
  FDetailFields.Free;
  FSQLSelectStmt.Free;
  FSQL.OnChange := nil;
  FSQL.Free;
  FLocateRecordset.Free;
  FLocateFoundRecords.Free;
  FLookUpRecordset.Free;
  FLookUpResultFields.Free;
  FRecordset.Free;
  ClearSortedFields;
  FSortedFields.Free;
   // TODO Check FreeNotification / RemoveFreeNotification
  if Assigned(FDatabase) then FDatabase.FreeNotification(Self);
  {$IFDEF FPC}
    //FDatabase.RemoveFreeNotification(Self);
  {$ENDIF}
  inherited Destroy;
End;

Function TSqlitePassDataset.GetFDatabase:TSqlitePassDatabase;
Begin
 Result:=FDatabase;
End;

Procedure TSqlitePassDataset.SetFDatabase(Value:TSqlitePassDatabase);
Begin
 if Active then FDatabase.DatabaseError.RaiseException(Msg2011);

 { Previous FDatabase }
 if Assigned(FDatabase) then
    begin
    { We remove the previous dataset from the Database datasets list }
    FDatabase.Datasets.Remove(Self);
    // TODO : Check FSQLSelectStmt.FDatabase assignement
    FSQLSelectStmt.FDatabase:= nil;
    FRecordset.FDatabase := nil;
    FLocateRecordset.FDatabase := nil;
    FLookUpRecordset.FDatabase := nil;
    end;

 FDatabase := Value;

 { New Database }
 if Assigned(FDatabase) then
    begin
    { We add this dataset to the Database datasets list }
    FDatabase.Datasets.Add(Self);
    { Re-Links FDatabase to the new one }
    FSQLSelectStmt.FDatabase:= FDatabase;
    FRecordset.FDatabase := FDatabase;
    FLocateRecordset.FDatabase := FDatabase;
    FLookUpRecordset.FDatabase := FDatabase;
    end;
End;


Procedure TSqlitePassDataset.SetFDatasetName(const Value: String);
var
WasActive: Boolean;
Begin
if (Value = FDatasetFullName) and (Value <> '')
   then Exit;

if (csDesigning in ComponentState) and (FDatasetName <> '') and (Fields.Count > 0)
   then if MessageDlg(Msg2028, mtWarning, [mbYes, mbNo], 0) = mrNo
           then Exit;

FDatasetFullName := Value;
FDatasetName := Copy(Value, Pos('.',Value)+1,Length(Value));

{ Resets properties which are specicific to a dataset
  and avoid to call Dataset.Close / Open twice if not csLoading }
if not (csLoading in ComponentState) then
   begin
   WasActive := Active;
   if WasActive
       then Close;
   DestroyFields;
   Filtered := False;
   Filter := '';
   FilterRecordLowerLimit := 0;
   FilterRecordUpperLimit := 0;
   Sorted := False;
   SortedBy := '';
   ResetLocateList;
   FLookUpRecordset.Close;
   MasterSource := nil;
   SetFDatasetType;
   Active := WasActive;
   end
   else SetFDatasetType;
End;

procedure TSqlitePassDataset.SetFDatasetType;
var
WasActive: Boolean;
i: integer;
SelectedFields, FieldIdentifier, TableIdentifier: string;
begin
FDatasetType := dtUnknown;
if (Database = nil) or (Not Database.Connected) then exit;

Try
{ disable notification }
FSQL.OnChange := nil;
WasActive := Active;
if WasActive
   then Close;

if FDatasetName = ''
   then FDatasetType := dtSQLDirect
   else if FDatabase.TableDefs.FindTable(FDatasetFullName) <> nil
           then FDatasetType := dtTable
           else if FDatabase.QueryDefs.FindQuery(FDatasetFullName) <> nil
                   then FDatasetType := dtQuery
                   else if FDatabase.Views.FindView(FDatasetFullName) <> nil
                        then FDatasetType := dtView
                        else if FDatabase.SQLStmtDefs.FindSQLStmt(FDatasetFullName) <> nil
                             then FDatasetType := dtSQLSelect;

Case FDatasetType of
     dtTable:   begin
                UpdateIndexDefs;
                TableIdentifier := FDatasetFullName;
                FSQLSelectStmt.QuoteString(TableIdentifier);
                { We check if the designtime designer is active.
                  If so we select all the table fields to display them in the field designer
                  Else we select only the wanted fields to save memory }
                If Assigned(Fields) and (Fields.Count > 0) and not (csDesigning in ComponentState)
                   then begin
                        For i := 0 to Pred(Fields.Count)
                            do if Fields[i].FieldKind = fkData
                                  then begin
                                       FieldIdentifier := Fields[i].FieldName;
                                       FSQLSelectStmt.QuoteString(FieldIdentifier);
                                       SelectedFields  := SelectedFields + FieldIdentifier + ', ';
                                       end;
                        System.Delete(SelectedFields,Length(SelectedFields)-1,2);
                        end
                   else SelectedFields := '*';

                   FSQL.Text := 'SELECT ' +  SelectedFields + ' FROM ' + TableIdentifier + ';'
                end;
     dtQuery:   begin
                FSQL.Text := FDatabase.QueryDefs.QueryByName(FDatasetFullName).Sql;
                UpdateParamsList;
                end;
     dtView:    FSQL.Text := FDatabase.Views.ViewByName(FDatasetName).SqlSelectStmt;
     dtSQLSelect:FSQL.Text := FDatabase.SQLStmtDefs.SQLStmtByName(FDatasetName).SQL;
     else       begin { Dataset name doesn't exist in database ?}
                FDatasetFullName := '';
                FDatasetName := '';
                Include(FSQLChanges, scSelect);
                UpdateFSQLStmt;
                end;
     end;
Finally
{ Enable notification }
FSQL.OnChange := UpdateFSQL;

Active := WasActive;

{ if designtime and dataset is closed, force fielddefs update to display infos}
if (Not Active) and (csDesigning in ComponentState) and (aoRead in FSQLSelectStmt.AvailableOperations)
   then InternalInitFieldDefs;
end;
end;

procedure TSqlitePassDataset.SetFSQL(const Value: TStringList);
begin
{ This is necessary to avoid error at designtime... }
if Value.Text <> FSQL.Text
   then FSQL.Assign(Value);
end;

{ This procedure is called whenever a change is made to FSQL
  via TStringList(FSQL).OnChange := UpdateFSQL,
  to ensure correct assignement with FSQL.assign or gettext/settext }
Procedure TSqlitePassDataset.UpdateFSQL(Sender: TObject);
Begin
if (csLoading in ComponentState) then exit;
{ We consider, once the SQL statement is changed by end user,
  that the Dataset doesn't match a known database entry (table, query or view)
  Setting the DatasetName to '' will call SetFDatasetName and SetFDatasetType }
DatasetName := '';
Include(FSQLChanges, scSelect);
End;

{ This procedure must be called whenever a change is made to FSQL
  to ensure correct info to FSQLStmt }
procedure TSqlitePassDataset.UpdateFSQLStmt;
var
i: Integer;
WhereStmt, MasterDetailStmt, FilterStmt, OrderbyStmt, FieldName, FieldValue: string;

begin
if (FSQLChanges = []) or ((FSortMode = smQuickSort) and (FSQLChanges = [scOrderBy])) then Exit;

Try
{ Select part }
if scSelect in FSQLChanges then
   begin
   FSQLSelectStmt.SQL := FSQL.Text;
   FSQLSelectStmt.Prepare;
   Exclude(FSQLChanges, scSelect);
   FRefetchRows := True;
   end;

{ Where part }
if scWhere in FSQLChanges then
   begin
   { Restore the "Where" part of the original SQL Stmt }
   WhereStmt := FSQLSelectStmt.OriginalSQLSections.GetSectionText(kwWhere);

   { Master/Detail Filtering }
   MasterDetailStmt := '';
   if (MasterFields <> '') and (FInInternalOpen or Active) and Assigned(FDetailFields)
      and Assigned(FMasterLink.Datasource) and (FMasterLink.Active) and (FDetailFields.Count > 0) then
      begin
      MasterDetailStmt := kwBraceOpen;
      for i := 0 to Pred(FDetailFields.Count) do
          begin
          FieldName := TField(FDetailFields[i]).FieldName;
          FSQLSelectStmt.QuoteString(FieldName);
          FieldValue := TField(FMasterLink.Fields[i]).AsString;
          FSQLSelectStmt.QuoteString(FieldValue);
          MasterDetailStmt := MasterDetailStmt + FieldName + kwSpace + kwEqual + KwSpace + FieldValue;
          if i < Pred(FDetailFields.Count)
             then MasterDetailStmt := MasterDetailStmt + KwSpace + kwAnd + KwSpace;
          end;
      MasterDetailStmt := MasterDetailStmt + kwBraceClose;
      end;

   { Filter Filtering }
   if Filtered
      then FilterStmt := kwSpace + kwBraceOpen + Filter + kwBraceClose
      else FilterStmt := '';

   if MasterDetailStmt <> '' then
      if WhereStmt = ''
         then WhereStmt := kwWhere + kwSpace + MasterDetailStmt
         else WhereStmt := WhereStmt + kwAnd + kwSpace + MasterDetailStmt;

   if FilterStmt <> '' then
      if WhereStmt = ''
         then WhereStmt := kwWhere + kwSpace + FilterStmt
         else WhereStmt := WhereStmt + kwAnd + kwSpace + FilterStmt;

   FSQLSelectStmt.SQLSections.ReplaceSectionText(kwWhere, WhereStmt);
   Exclude(FSQLChanges, scWhere);
   FRefetchRows := True;
   end;

{ OrderBy part - Sorting }
if scOrderBy in FSQLChanges then
   begin
   OrderByStmt := FSQLSelectStmt.OriginalSQLSections.GetSectionText(kwOrderBy);
   if Sorted and (Length(FSortedBy) > 0) then
      begin
      if OrderByStmt = ''
         then OrderByStmt := kwOrderBy + kwSpace + FSortedBy
         else OrderByStmt := OrderByStmt + kwAnd + kwSpace + MasterDetailStmt;
      end;

   FSQLSelectStmt.SQLSections.ReplaceSectionText(KwOrderBy, OrderByStmt);
   Exclude(FSQLChanges, scOrderBy);
   FRefetchRows := True;
   end;
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('SQL Statement : ' + FSQLSelectStmt.SQL);
{$ENDIF}

Except
 On EDatabaseError do
    FDatabase.DatabaseError.RaiseExceptionFmt(Msg3013, [CRLF, FSQLSelectStmt.SQL]);
end;
end;

Procedure TSqlitePassDataset.UpdateInternalFieldsInfos;
begin
{ Just an empty call }
Inherited;
end;

Function TSqlitePassDataset.GetFReadOnly: Boolean;
Begin
If Assigned(FDatabase) and FDatabase.ReadOnly
   then FReadOnly := True;

If Active
   and (not (DatasetType in [dtTable, dtQuery, dtSQLSelect, dtSQLDirect]))
   or (FSqlSelectStmt.AvailableOperations = [aoRead])
   then FReadOnly := True;

Result  := FReadOnly;
End;

Procedure TSqlitePassDataset.SetFReadOnly(const Value:Boolean);
Begin
  FReadOnly := Value;
End;

Procedure TSqlitePassDataset.Loaded;
begin
try
  {DatasetType cannot be set properly while TSqlitePassDataset
   is loading from DFM or LFM. So we do it now }
  SetFDatasetType;
  inherited Loaded;
except
  Application.HandleException(Self)
end;
end;

Procedure TSqlitePassDataset.Notification(AComponent: TComponent; Operation: TOperation); //Ok
Begin
 If (Operation = opRemove) and (AComponent = FDatabase) then FDatabase := Nil;
 Inherited Notification(AComponent, Operation);
End;


function TSqlitePassDataset.CheckCanOpen(VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Boolean;
begin
Result := Active or FInInternalOpen;

if Not Result then
   begin
   // TODO : Handle FDatabase.DatabaseError.RaiseException even if FDatabase is not Assigned...
   if not Assigned(FDatabase)
      then DatabaseError(Msg2018);

   if FDatabaseAutoActivate and (Not FDatabase.Connected)
      then FDatabase.Connected:=True;

   if Not FDatabase.Connected
      then if csDesigning in ComponentState
              then FDatabase.DatabaseError.RaiseException(Msg2021, -1, Self, VerboseLevel)
              else FDatabase.DatabaseError.RaiseException(Msg2020, -1, Self, VerboseLevel);

   if (FDatasetName='') And (FSQL.Count=0)
      then FDatabase.DatabaseError.RaiseException(Msg2019, -1, Self,  VerboseLevel);
    end;
end;

{procedure InternalOpen; virtual; abstract;

 TDataset.Open -> DoInternalOpen -> InternalOpen
------------------------------------------
Open the dataset. You must call InternalInitFieldDefs;
if DefaultFields is True, then you must call CreateFields,
which will create the necessary TFields from the fielddefs.}
Procedure TSqlitePassDataset.InternalOpen;
Begin

CheckCanOpen;

try
  FInInternalOpen := True;
  try
  { Init FieldDefs }
  InternalInitFieldDefs;

  { If fields definitions are stored in *.lfm or *.dfm using design time fields editor,
    then they are created at runtime when the form is created, even if the dataset is closed.
    If no persistent fields have been created in the Fields Editor
    then FPC uses the TDataset.CreateFields (It calls FieldDefs.CreateField) to create fields dynamically. }
  if DefaultFields then CreateFields;

  { Activates or deactivates Lookup Cache (TField.CacheLookUp) }
  ActivateLookUpCache;

  { TDataset.BindFields sets up all sorts of internal
    things like field offsets, as well as creating any calculated
    or lookup fields. It is within BindFields that CalcFieldsSize
    is actually set for later use in InitBufferPointers }
  BindFields(True);
  
  { Now Fields are defined, we can process MasterFields }
  if Assigned(MasterSource) then
     begin
     if FMasterAutoActivate and Assigned(MasterSource.Dataset)
        then MasterSource.Dataset.Open;
     ProcessMasterDetailFields;
     end;

  { Execute the Query }
  InternalRefresh;
 //FDatabase.Idle;
  except
  On EDatabaseError do
     begin
     FRecordset.Close;
     FSQLSelectStmt.FAvailableOperations := [];
     FDatabase.DatabaseError.RaiseException(Msg2029);
     end;
  end;
finally
  FInInternalOpen := False;
end;
end;

{procedure InternalRefresh; virtual; abstract;
-------------------------------------------
Refresh the dataset.}
procedure TSqlitePassDataset.InternalRefresh;
var
  SavedState: TDataSetState;
begin
Try
   UpdateFSQLStmt;

   If Not (aoRead in FSQLSelectStmt.AvailableOperations)
      then Exit;

   if Filtered and Assigned(OnFilterRecord)
      then SavedState := SetTempState(dsFilter);

   {  Queries data using SQL sentence : Refetch the records from Sqlite database engine }
   Try
   if FRefetchRows then
      begin
      FRecordset.Open(FSQLSelectStmt.SQL, FSQLSelectStmt.FPrimaryKeyCount);
      FRefetchRows := False;
      end;

   { Sorts the recordset using QuikSort in case we need to sort Calculated or LookUp fields }
   If FSQLChanges = [scOrderBy]
      then SortRecords(SortedBy);

   RefreshLocateList;
   RefreshLookUpRecordset;

   if State = dsFilter
      then RestoreState(SavedState);

   Except
    Try
    { if an error occurs, we try to get back to the original SQL Stmt }
      FSQLSelectStmt.RestoreOriginalStmt;
      Filtered := False;
      UpdateFSQLStmt;
      FRecordset.Open(FSQLSelectStmt.SQL, FSQLSelectStmt.FPrimaryKeyCount);
    Except
      Raise;
    end;
   end;
Except
 Close;
 Raise;
end;
end;

procedure TSqlitePassDataset.RefreshActive;
begin
if Active then Refresh;
end;

Procedure TSqlitePassDataset.InternalClearFieldDefs;
begin
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('InternalClearFieldDefs');
{$ENDIF}
{ Clear the internal items }
While FieldDefs.Count > 0
  do FieldDefs[0].Free;
{ Clear the list pointers }
FieldDefs.Clear;
end;

{procedure InternalClose; virtual; abstract;
-------------------------------------------
Closes the dataset. Any resources allocated in InternalOpen should be freed
here.}
Procedure TSqlitePassDataset.InternalClose;
Begin
FDatasetType := dtUnknown;
FSQLSelectStmt.Unprepare;
FRecordset.Close;
BindFields(False);
if DefaultFields
   then DestroyFields;
End;


{procedure InternalGotoBookmark(ABookmark: Pointer); virtual; abstract;
----------------------------------------------------------------------
Set the record position on the position that is associated with the
ABookMark data. The ABookMark data is the data that is acquired through
the GetBookMarkData call, and should be kept for each record.}
Procedure TSqlitePassDataset.InternalGotoBookmark(ABookmark: Pointer);
begin
 FRecordset.CurrentRecordIndex := Integer(ABookmark^);
End;

Function TSqlitePassDataset.BookmarkValid(Bookmark: TBookmark): Boolean;
Begin
// TODO should test Pointer is not Nil and Record was not deleted
Result := (Integer(Bookmark^) > -1) and (Integer(Bookmark^) < RecordCount);
End;

{ TDataset: function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; virtual; abstract;
  --------------------------------------------------------------------------
  Returns the BookmarkFlag (bfCurrent, bfBOF, bfEOF, bfInserted)
  associated with a recordBuffer.
  -------------------------------------------------------------------------- }
Function TSqlitePassDataset.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
 Result := FRecordSet.GetBookMarkFlag(PInteger(Buffer)^);
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('GetBookmarkFlag (bfCurrent, bfBOF, bfEOF, bfInserted) : Record = ' + IntToStr(PInteger(Buffer)^)
   + ' BookmakFlag := ' + IntToStr(Ord(FRecordSet.GetBookMarkFlag(PInteger(Buffer)^))));
{$ENDIF}
End;

{TDataset: procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); virtual; abstract;
-----------------------------------------------------------------------------
Set the BookmarkFlag 'Value' (bfCurrent, bfBOF, bfEOF, bfInserted)
of the record Buffer.
----------------------------------------------------------------------------- }
Procedure TSqlitePassDataset.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
Begin
FRecordSet.SetBookMarkFlag(PInteger(Buffer)^,Value);
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('> SetBookmarkFlag (bfCurrent, bfBOF, bfEOF, bfInserted) : Record = ' + IntToStr(PInteger(Buffer)^));
   //TODO + ' BookmakFlag := ' + IntToStr(Ord(Integer(Buffer)^).BookMarkFlag)));
{$ENDIF}
End;

{procedure GetBookmarkData(Buffer: PChar; Data: Pointer); virtual; abstract;
---------------------------------------------------------------------------
Puts the bookmarkdata for Buffer into the area pointed to by Data.}
Procedure TSqlitePassDataset.GetBookmarkData(Buffer: PChar; Data: Pointer);
Begin
PInteger(Data)^ := FRecordset.GetBookMarkData(PInteger(Buffer)^);
{$IFDEF DEBUG}
FDatabase.Debugger.Display('GetBookmarkData : Record = ' + IntToStr(PInteger(Buffer)^));
  //TODO + ' BookmakFlag := ' + IntToStr(Ord(TSqlitePassMemRecord(PInteger(Buffer)^).BookMarkFlag)));
{$ENDIF}
End;

{procedure SetBookmarkData(Buffer: PChar; Data: Pointer); virtual; abstract;
---------------------------------------------------------------------------
Move the 'Data' value to the BookmarkData associated with Buffer}
Procedure TSqlitePassDataset.SetBookmarkData(Buffer: PChar; Data: Pointer);
Begin
FRecordSet.SetBookMarkData(PInteger(Buffer)^,PCardinal(Data)^);
{$IFDEF DEBUG}
FDatabase.Debugger.Display('> SetBookmarkData : Record = ' + IntToStr(PInteger(Buffer)^));
{$ENDIF}
End;


Function TSqlitePassDataset.CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer;
const
  ResultCodes: array[Boolean, Boolean] of ShortInt = ((2,-1),(1,0));
begin
  { Tests for nil BookMarks First }
  Result := ResultCodes[Bookmark1 = nil, Bookmark2 = nil];
  If Result = 2 then
     begin
     Result := 0;
     if Integer(Bookmark1^) < Integer(Bookmark2^)
        then Result := -1
        else if Integer(Bookmark1^) > Integer(Bookmark2^)
             then Result :=  1;
     End;
end;

{ ----- RecNo ----- }

function TSqlitePassDataset.GetRecordCount: Longint;
begin
   CheckActive;
   Result := FRecordset.RecordsCount;
end;

function TSqlitePassDataset.GetRecNo: Integer;
begin
  Result := GetActiveRecord;
end;

procedure TSqlitePassDataset.SetRecNo(Value: Integer);
begin
Dec(Value);
if Value <> GetActiveRecord then
   begin
   CheckBrowseMode;
   DoBeforeScroll;
   FRecordset.CurrentRecordIndex := Value;
   Resync([rmCenter]);
   DoAfterScroll;
   end;
end;

Procedure TSqlitePassDataset.InternalInitIndexDefs;
Begin
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('InternalInitIndexDefs');
{$ENDIF}
 UpdateIndexDefs;
End;

Procedure TSqlitePassDataset.UpdateIndexDefs;
Begin
 FIndexDefs.RefreshFromCache;
End;

{procedure InternalInitFieldDefs; virtual; abstract;
-----------------------------------------------------------------------
This method is called from InternalOpen, and
initialize FieldDef definitions for all fields in a record.
It adds these definitions to the FFielddefs object.
----------------------------------------------------------------------}
Procedure TSqlitePassDataset.InternalInitFieldDefs;
var
i, CheckColumnCount, ColumnCount: Integer;
StmtFieldDef: TSqlitePassSelectStmtFieldDef;
NewFieldDef: TFieldDef;
WasActive: Boolean;

begin
{$IFDEF DEBUG}
   FDatabase.Debugger.Display('InternalInitFieldDefs');
{$ENDIF}

try
CheckCanOpen(vlSilent);
WasActive := Active;

if FDatasetType = dtUnknown
   then SetFDatasetType;

InternalClearFieldDefs;

{ First, we need to process the SQL Statement }
Include(FSQLChanges, scSelect);
UpdateFSQLStmt;

ColumnCount := SqliteDbv3_column_count(SqlSelectStmt.StmtHandle) - SqlSelectStmt.FPrimaryKeyCount;
CheckColumnCount := 0;

{ We build the TFieldDefs list from the SqlSelectStmt.FieldDefs list }
For i := 0 to Pred(SqlSelectStmt.FieldDefs.Count) do
    begin
    StmtFieldDef := SqlSelectStmt.FieldDefs[i];

    if (StmtFieldDef.DataType = ftString) and (StmtFieldDef.Size = 0)
       then StmtFieldDef.FSize := 30;

    NewFieldDef := TFieldDef.Create(FieldDefs, GetUniqueFieldName(StmtFieldDef),StmtFieldDef.DataType,
                                    StmtFieldDef.Size, False {= Not Required}, i+1);

    { Then we fill the following properties }
    With NewFieldDef do
      begin
      Precision := StmtFieldDef.Precision;

    { AutoInc Fields are not always ReadOnly so we don't set them as ReadOnly
      if StmtFieldDef.FAutoInc
         then Attributes := Attributes + [faReadOnly]; }

     { This will also set TFieldDef.Attributes +/- [faRequired] }
      Required := StmtFieldDef.NotNull and Not (faHiddenCol in Attributes)
                  And Not (faReadOnly in Attributes) and Not StmtFieldDef.PrimaryKey;

      Inc(CheckColumnCount);
      end; { With }
     end;{ For i }

{ Raises an Exception if Sqlite Engine doesn't return the same ColumCount
  as parsed SQL Statement ...}
If ColumnCount <> CheckColumnCount
   then FDatabase.DatabaseError.RaiseException(Msg2025, -1, Self);

InternalInitIndexDefs;

if not WasActive
   then close;
except
Raise;
end;
end;

{procedure InternalHandleException; virtual; abstract;
-----------------------------------------------------
not needed yet. Just implement an empty call.}
Procedure TSqlitePassDataset.InternalHandleException;
begin
 Application.HandleException(Self);
end;


Function TSqlitePassDataset.GetActiveRecord: Integer;
Begin
  Try
    Case State of
         dsBrowse      : if IsEmpty  // Ok
                            then Result := -1
                            else Result := PInteger(ActiveBuffer)^;
         dsBlockRead,
         dsNewValue,
         dsCurValue,
         dsEdit,
         dsInsert      : Result := PInteger(ActiveBuffer)^;

         dsCalcFields  : Result := PInteger(CalcBuffer)^;

         dsOldValue    :;// Result := FRecordset.OldValueRecordIndex;
         dsFilter      : Result := FRecordset.CurrentRecordIndex;
         else Result := -1;
         end;
  Except
   Result:=-1;
  End;
End;


{function GetFieldData(Field: TField; Buffer: Pointer): Boolean; virtual; abstract;
----------------------------------------------------------------------------------
Puts the data for field 'Field' from the active buffer into Buffer.
This is called whenever a field value is demanded by a TField, so it must be
efficient.
the TDataset class doesnt know anything about how our record buffer is organized.
The next question then becomes that if the TDataset class doesnt know about our
record structure, how it does it pull out field values from this structure?
How it does it put them back in when the user is editing a record?
The answer is that it uses the GetFieldData and SetFieldData methods which a
custom dataset class must override.
These methods are called by the Tdataset.fields when it needs to retrieve
or set a specific field value from the current record buffer.
GetFieldData is prototyped as follows:

function TDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;

The Field parameter is the field for which the value needs to be retrieved
We send back the needed value in the buffer.
Returns False if a null value was retrieved}
Function TSqlitePassDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
StrSize, FieldPos, ActiveRecIndex: Integer;
StrValue: string;

begin
Result := False;
ActiveRecIndex := GetActiveRecord;
FieldPos := Field.Tag;

{ When buffer is nil, the TDataset or TField.GetIsNull verifies if the current field IsNull }
Result := (ActiveRecIndex > -1) and not FRecordset.FieldIsNull(ActiveRecIndex, FieldPos);

if Result and Assigned(Buffer) then
   begin
   Case Field.DataType of
        ftBoolean    : Move(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, Buffer^, SizeOf(WordBool));
        ftWord,
        ftSmallint,
        ftInteger    : Move(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, Buffer^, SizeOf(Integer));

        ftAutoInc,
        ftLargeInt   : Move(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, Buffer^, SizeOf(Int64));

        ftBCD,
        ftCurrency,
        ftFloat      : Move(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, Buffer^, SizeOf(Double));
        ftDate       : PDateTimeRec(Buffer)^.Date := PInteger(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^;
        ftTime       : PDateTimeRec(Buffer)^.Time := PInteger(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^;
        ftDateTime   : PDateTimeRec(Buffer)^.DateTime := PDouble(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^;

        ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString
                     : begin
                       StrValue := FRecordset.GetFieldAsString(ActiveRecIndex, FieldPos);
                       StrSize := Min(Length(StrValue)+1,Field.DataSize);
                       Move(PChar(StrValue)^, Buffer^, StrSize);
                       end;
        end; { Case }
  end; { if Result and Assigned(Buffer) }

{$IFDEF DEBUG}
 FDatabase.Debugger.Display(PChar('< GetFieldData >'
                                + ' Dataset = ' + DatasetName
                                + ' ActiveRecordIndex = ' + IntToStr(ActiveRecIndex)
                                + ' Field = ' + Field.DisplayName
                                + ' FieldPos  = ' + IntToStr(FieldPos)));
{$ENDIF}
End;

{procedure SetFieldData(Field: TField; Buffer: Pointer); virtual; abstract;
--------------------------------------------------------------------------
Move the data associated with Field from Buffer to the activebuffer.}
Procedure TSqlitePassDataset.SetFieldData(Field: TField; Buffer: Pointer);
var
FieldPos, ActiveRecIndex: Integer;

begin
if not (State in dsWriteModes)
   then FDatabase.DatabaseError.RaiseException(SNotEditing, -1, Self);

if Field.ReadOnly
   then DatabaseErrorFmt(Msg2032, [Field.FieldName]);

ActiveRecIndex := GetActiveRecord;

if (ActiveRecIndex > -1) and Assigned(Buffer) then
   begin
   Field.Validate(Buffer);
   FieldPos := Field.Tag;
   case Field.DataType of
        ftBoolean    : Move(Buffer^, FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, SizeOf(WordBool));

        ftWord,
        ftSmallint,
        ftInteger    : Move(Buffer^, FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, SizeOf(Integer));

        ftAutoInc,
        ftLargeInt   : Move(Buffer^, FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, SizeOf(Int64));

        ftBCD,
        ftCurrency,
        ftFloat      : Move(Buffer^, FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^, SizeOf(Double));
        ftDate       : PInteger(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^ := PDateTimeRec(Buffer)^.Date;
        ftTime       : PInteger(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^ := PDateTimeRec(Buffer)^.Time;
        ftDateTime   : PDouble(FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos))^ := PDateTimeRec(Buffer)^.DateTime;

        ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString
                     : FRecordset.SetFieldAsString(ActiveRecIndex, FieldPos, String(PChar(Buffer)));
        end; { Case }

  FRecordset.SetFieldNullValue(ActiveRecIndex, FieldPos, False);
  FRecordset.SetFieldWasModified(ActiveRecIndex, FieldPos, True);

  { The DataEvent is used to set TDataset.SetModified to True.
    Then TDataset.CheckBrowseMode will be able to Post the record if necessary
    when scrolling the dataset }
  if not (State in [dsInternalCalc, dsCalcFields, dsFilter, dsNewValue])
     then DataEvent(deFieldChange, Longint(Field));

  end; { if (ActiveRecIndex > -1) and Assigned(Buffer) }

 {$IFDEF DEBUG}
 FDatabase.Debugger.Display(PChar('< GetFieldData >'
                                + ' Dataset = ' + DatasetName
                                + ' ActiveRecordIndex = ' + IntToStr(ActiveRecIndex)
                                + ' Field = ' + Field.DisplayName
                                + ' FieldPos  = ' + IntToStr(FieldPos)));
 {$ENDIF}
end;

procedure TSqlitePassDataset.UpdateAutoIncValues;
var
i, ActiveRecIndex: Integer;
FieldPos: Integer;
begin
ActiveRecIndex := GetActiveRecord;
for i := 0 to Pred(FSQLSelectStmt.FAutoIncFieldDefs.Count) do
  begin
  FieldPos :=  FSQLSelectStmt.FFieldDefs[i].FRecordFieldIndex;
//  if FRecordset.FieldIsNull(ActiveRecIndex, FieldPos)
//     then begin
          Move(FSQLSelectStmt.FFieldDefs[i].FFieldDefs.FTableDef.FLastInsertedRowId,
               FRecordset.GetFieldBuffer(ActiveRecIndex, FieldPos)^,
               SizeOf(Int64));
          FRecordset.SetFieldNullValue(ActiveRecIndex, FieldPos, False);
//          end;
  end;
end;


{procedure InternalFirst; virtual; abstract;
-------------------------------------------
This is called when 'TDataset.First' is called;
After this method, TDataset calls 'Getrecord' asking for the next record.
'Next' will inc(CurrentRecordIndex). Finally 'GetRecord' will retrieve
Record[FMemRecordset.CurrentRecordIndex] i.e. Record[0]
or return grBOF if we are already at BOF}
Procedure TSqlitePassDataset.InternalFirst;
Begin
FRecordset.First;
End;

{procedure InternalLast; virtual; abstract;
------------------------------------------
This is called when 'TDataset.Last' is called;
After this method, TDataset calls 'Getrecord' asking for the prior record.
'Prior' will Dec(CurrentRecordIndex). Finally 'GetRecord' will retrieve
Record[FMemRecordset.CurrentRecordIndex] i.e. Record[Pred(RecordCount)]
or return grEOF if we are already at EOF}
Procedure TSqlitePassDataset.InternalLast;
Begin
FRecordset.Last;
End;

{procedure InternalSetToRecord(Buffer: PChar); virtual; abstract;
----------------------------------------------------------------
Set the current record to the record in Buffer; if bookmark data
is specified in this buffer, that data can be used to determine which
record this should be.}
Procedure TSqlitePassDataset.InternalSetToRecord(Buffer: PChar);
Begin
FRecordset.CurrentRecordIndex := PInteger(Buffer)^;
End;

procedure TSqlitePassDataset.InternalInsert;
begin
{ This is just an empty call - could also be empty}
inherited InternalInsert;
end;

Procedure TSqlitePassDataset.InternalEdit;
Begin
inherited InternalEdit;
FRecordset.SaveCurrentRecordOldValues;
End;

Procedure TSqlitePassDataset.InternalCancel;
Begin
inherited InternalCancel;
Case State of
     dsEdit:   FRecordset.RestoreCurrentRecordOldValues;
     dsInsert: FRecordset.Cancel;
     end;
End;

{procedure InternalPost; virtual; abstract;
------------------------------------------
Post the data in the active buffer to the underlying dataset. }
Procedure TSqlitePassDataset.InternalPost;
begin
try
Case State of
     dsEdit   : FSQLSelectStmt.Post(ptUpdate);
     dsInsert : begin
                FSQLSelectStmt.Post(ptInsert);
                FRecordset.PostRecord;
                UpdateAutoIncValues;
                end;
     end;
Except
FDatabase.DatabaseError.RaiseException(Msg2022);
end;
End;


{procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); virtual; abstract;
---------------------------------------------------------------------------------
Adds a record to the dataset. The record's data is in Buffer and Append
indicates whether the record should be appended (True) or Inserted (False).
Note that for SQL based datasets, this has no meaning.}
Procedure TSqlitePassDataset.InternalAddRecord(Buffer: Pointer; Append: Boolean);
Begin
// Used when end user wants to add an entire record to the dataset
// TODO Check (TDataSet.AppenRecord/TDataSet.InsertRecord/TDataSet.AddRecord(const Values: array of const; Append: Boolean);
End;


{procedure InternalDelete; virtual; abstract;
--------------------------------------------
Deletes the current Record.}
Procedure TSqlitePassDataset.InternalDelete;
Begin
try
  FSQLSelectStmt.Post(ptDelete);
  FRecordset.DeleteRecord(FRecordset.CurrentRecordIndex);
except
  FDatabase.DatabaseError.RaiseException(Msg2024);
end;
End;


{function IsCursorOpen: Boolean; virtual; abstract;
--------------------------------------------------
This function should return True if data is available, even if the dataset
is not active.}
Function TSqlitePassDataset.IsCursorOpen: Boolean;
Begin
  Result := FRecordset.Active and (aoRead in FSQLSelectStmt.AvailableOperations);
End;

{function AllocRecordBuffer: PChar; virtual; abstract;
-----------------------------------------------------
Must allocate enough memory to store a complete record in the dataset.
Optionally, this buffer must contain enough memory to store bookmarkdata.
The descendent must be able to construct a bookmark from this buffer.}
Function TSqlitePassDataset.AllocRecordBuffer: PChar;
Begin
{ test added to avoid memory leak
  when destroying component a [uncessary?] call is sent from TDataSet.SetBufListSize...}
if csDestroying in ComponentState then exit;
GetMem(Result, GetRecordSize);
End;


{procedure FreeRecordBuffer(var Buffer: PChar); virtual; abstract;
-----------------------------------------------------------------
Must free the memory allocated in the AllocRecordBuffer call.}
Procedure TSqlitePassDataset.FreeRecordBuffer(var Buffer: PChar);
Begin
FreeMem(Buffer, GetRecordSize);
End;


{procedure InternalInitRecord(Buffer: PChar); virtual; abstract;
---------------------------------------------------------------
This method is called to initialize a new field buffer when the dataset
is put into insert or append mode. We ask the InternalRecorset to add a new record }
Procedure TSqlitePassDataset.InternalInitRecord(Buffer: PChar);
begin
PInteger(Buffer)^ := FRecordset.AddTempRecord;
{$IFDEF DEBUG}
FDatabase.Debugger.Display('InternalInitRecord : ' + IntToStr(PInteger(Buffer)^));
{$ENDIF}
End;


Function TSqlitePassDataset.GetCanModify: Boolean;
Begin
 Result := Active and (DatasetType in [dtTable, dtQuery, dtSQLSelect, dtSQLDirect]) and (not ReadOnly);
End;

{function GetRecordSize: Word; virtual; abstract;
------------------------------------------------
Since we just use pointers to keep track of TSqlitePassMemRecords, the dataset just needs
to allocate enough size to hold a pointer value (i.e. longint or integer}
Function TSqlitePassDataset.GetRecordSize: Word;
begin
  Result := SizeOf(Pointer);
End;


{function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; virtual; abstract;
-----------------------------------------------------------------------------------
This method must do 3 things:
1) Get the record data for the next/current/previous record, depending
   on the GetMode value. It should return
    grOK    if all was OK.
    grBOF   if the previous record was requested, and we are at the start.
    grEOF   if the next record was requested, and we are at the end.
    grError if an error occurred.

2) If DoCheck is True, and the result is grError, then an exception must be
    raised.

3) It should initialize bookmark data for this record with flag 'bfCurrent'
   This data can be stored in the buffer, if space was allocated for it with
   AllocRecordBuffer.
}
Function TSqlitePassDataset.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
Begin

case GetMode of
  gmCurrent: FRecordset.MoveBy(0);
  gmNext:    FRecordset.Next;
  gmPrior:   FRecordset.Prior;
  end;

Result := FRecordset.MoveState;

Case Result of
    grOk :  begin
            PInteger(Buffer)^ := FRecordSet.FCurrentRecordIndex;
            { we update the calculated / Lookup fields values }
            if FHasCalcFields then GetCalcFields(Buffer);
            end;
    grError:If DoCheck then FDatabase.DatabaseError.RaiseException(FDatasetName + ' : ' + Msg1040);
    end;
{$IFDEF DEBUG}
 With FDatabase.Debugger do
      begin
      Display('[ --------- GetRecord ------- ]');
      Display('Getmode (gmCurrent, gmNext, gmPrior) = ' + IntToStr(Ord(GetMode)));
      Display('CurrentRecordIndex = ' + IntToStr(FRecordset.CurrentRecordIndex));
      Display('Dataset.state (dsInactive, dsBrowse, dsEdit, dsInsert, dsSetKey, dsCalcFields, dsFilter, dsNewValue, dsOldValue, dsCurValue, dsBlockRead, dsInternalCalc) = ' + IntToStr(Ord(Self.State)));
      Display('Dataset.Movestate (grOK, grBOF, grEOF, grError)= ' + IntToStr(Ord(Result)));
      Display('[ ------   <Get Record>  ---- ]');
      Display('');
      end;
{$ENDIF}
end;


{ ----- Filtering ----- }

{ Filtering with Master/Detail }

function TSqlitePassDataset.GetMasterDataSource: TDataSource;
begin
  Result := FMasterLink.DataSource;
end;

procedure TSqlitePassDataset.SetMasterDataSource(Value: TDataSource);
begin
  if IsLinkedTo(Value)
     then FDatabase.DatabaseError.RaiseException(Msg2026, -1, Self);
  FMasterLink.DataSource := Value;
  if Value = nil
     then MasterFields := '';
end;

function TSqlitePassDataset.GetMasterFields: string;
begin
  Result := FMasterFields;
end;

procedure TSqlitePassDataset.SetMasterFields(const Value: string);
begin
FMasterFields := Value;

{ We parse MasterFields/DetailFields only when dataset.fields are defined }
if not (csLoading in ComponentState) and Active
   then ProcessMasterDetailFields;
end;

procedure TSqlitePassDataset.ProcessMasterDetailFields;
var
JoinedFieldEnd, PosEqual: Integer;
ParsedValue, DetailFieldName, CurrentJoinedFieldsNames, MasterFields: String;
begin
FDetailFields.Clear;
ParsedValue := '';
if FMasterFields <> '' then
   begin
   MasterFields := FMasterFields;
     Repeat
     JoinedFieldEnd := Pos(';', MasterFields);

     If JoinedFieldEnd = 0
        then JoinedFieldEnd := Length(MasterFields)+1;

     If JoinedFieldEnd > 0 then
        begin
        CurrentJoinedFieldsNames := System.Copy(MasterFields, 1, JoinedFieldEnd-1);
        System.Delete(MasterFields, 1, JoinedFieldEnd);
        PosEqual := Pos('=', CurrentJoinedFieldsNames);
        If PosEqual = 0
           then PosEqual := Length(CurrentJoinedFieldsNames)+1;
        If PosEqual > 0 then
           begin
           DetailFieldName := System.Copy(CurrentJoinedFieldsNames, PosEqual+1, JoinedFieldEnd - PosEqual);
           FDetailFields.Add(Fields.FindField(DetailFieldName));
           System.Delete(CurrentJoinedFieldsNames, PosEqual, JoinedFieldEnd - PosEqual);
           ParsedValue := ParsedValue + CurrentJoinedFieldsNames + ';';
           end;
        end;
      Until MasterFields = '';
    if ParsedValue[Length(ParsedValue)] = ';'
       then System.Delete(ParsedValue, Length(ParsedValue), 1);
    end;

FMasterLink.FieldNames := ParsedValue;
Include(FSQLChanges, scWhere);
RefreshActive;
end;

procedure TSqlitePassDataset.MasterChanged(Sender: TObject);
begin
 Include(FSQLChanges, scWhere);
 RefreshActive;
end;

procedure TSqlitePassDataset.MasterDisabled(Sender: TObject);
begin
 Include(FSQLChanges, scWhere);
 RefreshActive;
end;

procedure TSqlitePassDataset.DoOnNewRecord;
var
  i: Integer;
begin
{ Set default detail fields values from master fields if needed }
if MasterFields <> '' then
   begin
   For i := 0 to Pred(FDetailFields.Count)
       do TField(FDetailFields[i]).Value := TField(FMasterLink.Fields[i]).Value;
   end;

inherited DoOnNewRecord;
end;


{ Filtering with other filters }

procedure TSqlitePassDataset.CheckFiltered(ApplyFilter: Boolean = False);
begin
  Include(FSQLChanges, scWhere);
  ApplyFilter := ApplyFilter or (Filtered and ((Filter <> '') or (FFilterRecordLowerLimit <> 0) or (FFilterRecordUpperLimit <> 0)));
  If ApplyFilter
     then RefreshActive;
end;

{ This procedure is kept to give a public access to force applying a filter }
procedure TSqlitePassDataset.ApplyFilter;
begin
 Filtered := True;
 CheckFiltered(Filtered);
end;

procedure TSqlitePassDataset.SetFiltered(Value: Boolean);
begin
  If Value <> Filtered then
     begin
     if (Not Filtered) and (RecordCount = 0) then Exit; 
     Inherited SetFiltered(Value);
     CheckFiltered(True);
     end;
end;

procedure TSqlitePassDataset.SetFilterText(const Value: string);
begin
  Inherited SetFilterText(Value);
  CheckFiltered(Filtered);
end;

procedure TSqlitePassDataset.SetFFilterRecordLowerLimit(
  const Value: Cardinal);
begin
  FFilterRecordLowerLimit := Value;
  CheckFiltered(Filtered);
end;

procedure TSqlitePassDataset.SetFFilterRecordUpperLimit(
  const Value: Cardinal);
begin
  FFilterRecordUpperLimit := Value;
  CheckFiltered(Filtered);
end;

{ Implementation of virtual method TDataset.FindRecord }
function TSqlitePassDataset.FindRecord(Restart, GoForward: Boolean): Boolean;
begin
SetFound(False);
if GoForward
   then if Restart
           then First
           else Next
   else if Restart
           then Last
           else Prior;
SetFound((FRecordset.MoveState <> grError) and (FRecordset.FRecordsCount > 0));
Result := Found;
end;

{ --------- Locate --------- }

{ Locate

  On peut procder par
   1 - Faire une requete SQL en reprenant la requete actuelle du Dataset et en filtrant sur
     les critres de slection. On DOIT alors faire correspondre le recordset filtr
     (locateRecordset) avec le recorsdet en cours, Les rsultats (Seulement le RecordIndex) sont conservs
     dans une TList pour permettre LocateFirst, LocateNext, LocatePrior, LocateLast
     LocateRecordCount (pour indiquer le nombre d'enregistrements trouvs) tant
     que Dataset est Open ou que l'on a pas appel Refresh.

     Le point dlicat est de gnrer la liste des enregistrements le + rapidement possible :

             1a - on peut faire deux boucles imbriques pour parcourir le recordset courant
             (qui est forcment + grand ou qual  celui filtr (locateRecordset) donc limiter
             la boucle  1 seule passe
             puis pour chaque enregistrement, vrifier dans locateRecordset la prsence
             de l'enregistrement (le critre de correspondance tant la recherche du ou
             des mme Rowid dans les deux recordset).
             On aura intrt  trier locateRecordset sur les mmes critres que le recordset courant.
             Ainsi, le point de dpart de la boucle interne a est incrment chaque fois
             qu'un enregistrement correspond. On diminue au fur et  mesure le nombre d'enre  parcourir.
             On sort de la boucle principale ds que tous les enregistrements de LocateRecordset
             on trouvs une correspondance pour ne pas parcourir entirement le recordset courant
             si ce n'est pas ncessaire).

             1b - SI le recordset courant est TRIE sur RowId
                  on peut parcourir le locaterecordset et rechercher
                  l'enregistrement correspondant dans le recordset courant en utilisant
                  une recherche dichotomique.
                  Une fois l'enregistrement trouv, on vrifie la correspondance en comparant
                  les cls primaires des deux enregistrements.

     Quelle mthode est la plus rapide ?
            pour 1a, si le recordset courant contient 10000 enre et LocateRecordset 100 enre
            on peut avoir un max d'environ 10000 + (100 * 10000) = +/- 10.000.000 d'itrations
            pour 1b,  condition que le recordset courant soit tri sur RowId, on aurait un max de 100 + (100 x 9) = +/- 1000 itrations.
            Donc, dans ce cas, choix de la mthode 1b.

            Exemple de recherche Dichotomique :

            function RechDicho (var f: TypeFichier; v: t_auteur_livre) : integer;
            var G, D, m : integer;
            trouve : boolean;
            courant : t_livre;
            Begin
            G := 0; D := filesize (f) - 1; // bornes de lensemble considr
            trouve := false;
            while (G <= D and not trouve) do
            begin
            m := (G + D) div 2; // calcul de la valeur moyenne
            seek (f, m); // positionnement du repre de fichier
            read (f, courant); // lecture de lenregistrement
            if courant.auteur = v then trouve := true // condition de sortie
            else
            if courant.auteur > v then D := m-1 // trop  droite
            else G := m+1; // trop  gauche
            end; // while
            if trouve then RechDicho := m // dtermine ltat de sortie du prg
            else RechDicho := -1;
            end; // RechDico
     }

procedure TSqlitePassDataset.LocateMoveToRecord;
begin
  FLocateMoveState:= grError;
  if (FLocateCurrentItem > -1) and (FLocateCurrentItem < FLocateFoundRecords.Count) then
     begin
     MoveBy(PInteger(FLocateFoundRecords.Items[FLocateCurrentItem])^ - PInteger(ActiveBuffer)^);
     FLocateMoveState := grOk;
     if FLocateCurrentItem = 0
        then FLocateMoveState := grBOF
        else if FLocateCurrentItem = Pred(FLocateFoundRecords.Count)
             then FLocateMoveState := grEOF;
     end;
end;

function TSqlitePassDataset.LocateFirst: Boolean;
begin
 FLocateCurrentItem := 0;
 LocateMoveToRecord;
 Result := (FLocateMoveState <> grError);
end;

function TSqlitePassDataset.LocateNext: Boolean;
begin
 if FLocateCurrentItem < Pred(FLocateFoundRecords.Count)
    then Inc(FLocateCurrentItem);
 LocateMoveToRecord;
 Result := (FLocateMoveState = grOk);
end;

function TSqlitePassDataset.LocatePrior: Boolean;
begin
 if (FLocateCurrentItem > 0)
    then Dec(FLocateCurrentItem);
 LocateMoveToRecord;
 Result := (FLocateMoveState = grOk);
end;

function TSqlitePassDataset.LocateLast: Boolean;
begin
 FLocateCurrentItem := Pred(FLocateFoundRecords.Count);
 LocateMoveToRecord;
 Result := (FLocateMoveState <> grError);
end;

function TSqlitePassDataset.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
var
SqlStmt: String;

  { If the first PrimayKey is found checks the others primarykeys  }
  function CheckOthersPrimaryKeys(CurrentRecord, LocateCurrentRecord: Integer): Boolean;
  var
  i: Integer;
  begin
  Result := True;
  for i := 1 to Pred(FSQLSelectStmt.TableDefs.Count) do
     begin
     Result := FRecordset.GetRowId(CurrentRecord, i) = FLocateRecordset.GetRowId(LocateCurrentRecord, i);
     if Not Result then Break;
     end;
  end;

  { Finds the first matching PrimaryKey :
    Scans FRecordset and check if the record is also in FLocateRecordset }
  function FindKeyValues(LocateCurrentRecordIndex: Integer): Integer;
  var
  i: Integer;
  begin
  Result := -1;
  for i := 0 to Pred(FRecordset.RecordsCount) do
      begin
      if FRecordset.GetRowId(i,0) = FLocateRecordset.GetRowId(LocateCurrentRecordIndex,0) then
         if CheckOthersPrimaryKeys(i, LocateCurrentRecordIndex) then
            begin
            Result := i;
            Break;
            end;
      end; {For i}
  end;

  { Builds the link table between FLocateRecordset and FRecordset }
  procedure RefreshFoundRecordsList;
  var
  i, RecordIndex: integer;
  begin
  FLocateFoundRecords.Clear;
  For i := 0 to Pred(FLocateRecordset.RecordsCount) do
      begin
      RecordIndex := FindKeyValues(i);
      if RecordIndex > -1
         then FLocateFoundRecords.Add(Pointer(RecordIndex));
      end;
  end;

begin
  { TODO - Optimize with Dichotomique search when FRecordsert
    is sorted on PrimaryKey }
  Result := False;

  { Saves Locate infos to enable automatic list update }
  FLocateKeyFields := KeyFields;
  FLocateKeyValues := KeyValues;
  FLocateOptions := Options;

  { Retrieves the corresponding records from Sqlite db }
  { We only support locate from a unique table in current version so, TableCount is set to '1' }
  SqlStmt := CreateLocateSQL(KeyFields, KeyValues, Options);
  if SqlStmt <> '' then
     begin
     FLocateRecordset.Open(SqlStmt, 1);
     RefreshFoundRecordsList;
     { Necessary call to notify lookupcontrols to update display }
     DataEvent(deDataSetChange, 0);
     FLocateRecordset.Close;
     { Moves to the first matching record }
     Result := LocateFirst;
     end
     else Result := False;

end;

procedure TSqlitePassDataset.RefreshLocateList;
begin
  if FLocateSmartRefresh and (FLocateKeyFields <> '')
     then Locate(FLocateKeyFields, FLocateKeyValues, FLocateOptions);
end;

procedure TSqlitePassDataset.ResetLocateList;
begin
 FLocateKeyFields := '';
 FLocateKeyValues := null;
 FLocateOptions := [];
 FLocateFoundRecords.Clear;
 FLocateCurrentItem := -1;
 FLocateMoveState := grError;
end;

function TSqlitePassDataset.GetLocateRecordCount: Integer;
begin
 Result := FLocateFoundRecords.Count;
end;

function TSqlitePassDataset.CreateLocateSQL(const KeyFields: string;
  const KeyValues: Variant; LocateOptions: TLocateOptions; MaxRecordCount: Integer = 0): String;
var
  i: Integer;
  WhereStmt, LocateStmt, CompareStr, PartialStr: String;
  LocateFields: TList;
  WasCaseSensitiveLike: Boolean;
  LocateSQL: TSqlitePassSelectStmt;
begin
  Result := '';
  LocateFields := TList.Create;
  LocateSQL := TSqlitePassSelectStmt.Create(Self);
  LocateStmt := '';

  WasCaseSensitiveLike := FDatabase.Options.CaseSensitiveLike;
  if loPartialKey in LocateOptions
     then begin
          CompareStr := kwLike;
          PartialStr := '%'
          end
     else begin
          CompareStr := '=';
          PartialStr := '';
          end;

  CompareStr := kwSpace + CompareStr + kwSpace;

  Try
  GetFieldList(LocateFields, KeyFields);

  if (LocateFields.Count = 0) or (FSQLSelectStmt.SQL = '')
     then Exit;

  { Prepares the SQL Statement with the additional "Where" clause}
  if LocateFields.Count > 1
     then begin
          for i := 0 to Pred(LocateFields.Count) do
              LocateStmt := LocateStmt + TField(LocateFields[i]).FieldName + CompareStr + '"' + PartialStr + VarAsType(KeyValues[i], VarString) + PartialStr + '" AND ';
          System.Delete(LocateStmt, Length(LocateStmt)-3, 4);
          end
     else LocateStmt :=  TField(LocateFields[0]).FieldName + CompareStr + '"' + PartialStr + VarAsType(KeyValues, VarString) + PartialStr + '"';

  With LocateSQL do
      begin
      SQL := FSQLSelectStmt.SQL;
      WhereStmt := SQLSections.GetSectionText(kwWhere);
      if WhereStmt = ''
         then WhereStmt := kwWhere + kwSpace + LocateStmt
         else begin
              WhereStmt := StringReplace(WhereStmt, kwSQLEnd, KwSpace, [rfIgnoreCase]);
              WhereStmt := WhereStmt + kwSpace + KwAnd + KwSpace + LocateStmt;
              end;
       SQLSections.ReplaceSectionText(kwWhere, WhereStmt);

       if MaxRecordCount > 0
          then SQLSections.ReplaceSectionText(kwLimit, kwLimit + kwSpace + IntToStr(MaxRecordCount));

       Result := SQL;
      end;

  finally
  LocateFields.Free;
  LocateSQL.Free;
  FDatabase.Options.CaseSensitiveLike := WasCaseSensitiveLike;
  end;
end;

{ --------- LookUp --------- }

{Lookup rcupre les infos des champs souhaits selon que certains champs correspondent aux critres de recherche

On peut procder par
 1 - Faire une requete SQL en reprenant la requete actuelle du Dataset et en filtrant sur
     les critres de slection. Il n'est pas utile de faire correspondre le recordset filtr
     (lookupRecordset) avec le recordset en cours, mais on doit conserver les rsultats dans
     le LookUpRecordset pour permettre LookUpFirst, LookUpNext, LookUpPrior, LookUpLast
     LookUpRecordCount (pour indiquer le nombre d'enregistrements trouvs) tant
     que Dataset est Open ou que l'on a pas appel Refresh.
     }

function TSqlitePassDataset.LookupFirst: Variant;
begin
FLookUpRecordset.First;
FLookUpRecordset.Next;
Result := GetLookUpResultValues;
end;

function TSqlitePassDataset.LookupNext: Variant;
begin
FLookUpRecordset.Next;
Result := GetLookUpResultValues;
end;

function TSqlitePassDataset.LookupPrior: Variant;
begin
FLookUpRecordset.Prior;
Result := GetLookUpResultValues;
end;

function TSqlitePassDataset.LookupLast: Variant;
begin
FLookUpRecordset.Last;
Result := GetLookUpResultValues;
end;

function TSqlitePassDataset.GetLookUpResultValues: Variant;
var
i: integer;

  function GetResultFieldValue(ResultField: TField): Variant;
  var
  FieldPos: Integer;
  begin
  Result := Null;
  if FLookUpRecordset.RecordsCount > 0 then
     begin
     FieldPos := ResultField.Tag;
     Result := FLookUpRecordset.GetFieldAsVariant(FLookUpRecordset.CurrentRecordIndex, FieldPos, ResultField);
     end;
  end;
  
begin
if FLookUpResultFields.Count > 1 then
   begin
   Result := VarArrayCreate([0, FLookUpResultFields.Count - 1], VarVariant);
   for i := 0 to Pred(FLookUpResultFields.Count)
       do Result[i] := GetResultFieldValue(TField(FLookUpResultFields[i]));
   end
   else Result := GetResultFieldValue(TField(FLookUpResultFields[0]));
end;

function TSqlitePassDataset.Lookup(const KeyFields: string; const KeyValues: Variant;
  const ResultFields: string): Variant;
begin
  Result := Null;
 { Retrieves the corresponding records from Sqlite db }
 { We only support locate from a unique table in current version so, TableCount is set to '1' }
  FLookUpRecordset.Open(CreateLocateSQL(KeyFields, KeyValues, FLookUpOptions, FLookUpMaxRecordCount), 1);
  FLookUpResultFields.Clear;
  GetFieldList(FLookUpResultFields, ResultFields);
  Result := LookUpFirst;
end;

procedure TSqlitePassDataset.RefreshLookUpRecordset;
begin
// TODO
// If FLookUpSmartRefresh then ...
end;

procedure TSqlitePassDataset.SetFLookUpCache(Const Value: Boolean);
begin
if Active then FDatabase.DatabaseError.RaiseException(Msg2030);
FLookUpCache := Value;
end;

procedure TSqlitePassDataset.ActivateLookUpCache;
var
i: Integer;
begin
if FLookUpCache then
   for i := 0 to Pred(FieldCount) do
       if Fields[i].FieldKind = fkLookup
          then Fields[i].LookupCache := True;
end;

{ ----- Sorting ----- }
procedure TSqlitePassDataset.SetFSortCalcFields(const Value: Boolean);
begin
  FSortCalcFields := Value;
  If FSortCalcFields
     then FSortMode := smQuickSort;
end;

procedure TSqlitePassDataset.SetFSortMode(const Value: TSqlitePassSortMode);
begin
  FSortMode := Value;
  If FSortMode = smSQL
     then FSortCalcFields := False;
end;

procedure TSqlitePassDataset.SetFSorted(const Value: Boolean);
begin
  if FSortedBy = ''
     then FSorted := False
     else FSorted := Value;
  if FSorted
     then RefreshActive;
end;

procedure TSqlitePassDataset.SetFSortedBy(const Value: String);
begin
if Value <> FSortedBy then
   begin
   FSortedBy := Value;
   if FSortedBy = ''
      then FSorted := False;
   Include(FSQLChanges, scOrderBy);
   if Sorted
      then RefreshActive;
   end;
end;

Procedure TSqlitePassDataset.Sort;
var
WasSorted:  Boolean;
Begin
 WasSorted := FSorted;
 { Sorts the dataset }
 Sorted := True;
 { Back to previous state }
 FSorted := WasSorted;
End;

{ --- (Enable sorting on Calculated and Lookup Fields) --- }
Procedure TSqlitePassDataset.FillLookupFields;
var
i: Integer;
Buffer: PChar;

begin
if FHasCalcFields then
   begin
   GetMem(Buffer, SizeOf(Pointer));
     try
       { we update all the calculated / Lookup fields values }
       for i:= 0 to Pred(FRecordset.RecordsCount) do
           begin
           PInteger(Buffer)^ := i;
           GetCalcFields(Buffer);
           end;
     finally;
       FreeMem(Buffer, SizeOf(Pointer));
     end;
   end;
end;

procedure TSqlitePassDataset.ClearSortedFields;
var
i: Integer;

begin
{ Clear the internal items }
for i := 0 to Pred(FSortedFields.Count)
  do TSqlitePassSortedField(FSortedFields.Items[i]).Free;
{ Clear the list pointers }
FSortedFields.Clear;
end;


procedure TSqlitePassDataset.GetSortedFields(SortedBySql: String);
var
SqlStmt: TSqlitePassSqlStmt;
SortedField: TSqlitePassSortedField;

begin
SqlStmt := TSqlitePassSqlStmt.Create(FDatabase);
SqlStmt.Tokenizer.Text := SortedBySql;
ClearSortedFields;
SqlStmt.Tokenizer.Next(ttIdentifier);

While Not SqlStmt.Tokenizer.EOF do
 begin
 SortedField := TSqlitePassSortedField.Create;
 with SortedField do
       begin
       FieldName := SqlStmt.Tokenizer.Token.Text;
       FieldIndex := FieldByName(FieldName).Index;
       SqlStmt.Tokenizer.Next(ttIdentifier);
       Descendant := Uppercase(SqlStmt.Tokenizer.Token.Text) = 'DESC';
       end;
 FSortedFields.Add(SortedField);
 SqlStmt.Tokenizer.Next(ttIdentifier);
 end;
SqlStmt.Free;
end;

Procedure TSqlitePassDataset.SortRecords(SortedBySql: String);
var
  Save_Cursor:TCursor;

begin
  If IsEmpty then Exit;

  Save_Cursor := Screen.Cursor;
  Screen.Cursor := crHourglass;
  try
    if FSortCalcFields
       then FillLookupFields;

    GetSortedFields(SortedBySql);
    if FSortedFields.Count > 0
       then FRecordset.SortRecords(FSortedFields);
  finally
    Screen.Cursor := Save_Cursor;
    Exclude(FSQLChanges, scOrderBy);
  end;
end;

{ ----- Sorting - END ----- }


{ Returns a Unique field name, adding a number if necessary to duplicate names }
function TSqlitePassDataset.GetUniqueFieldName(FieldDef: TSqlitePassSelectStmtFieldDef): String;

 function AddUniqueNumber(FieldName: String): String;
 var
 i: Integer;
 NewName: String;
 begin
 i := 0;
 Result := FieldDef.FieldName;
 if FieldDefs.IndexOf(FieldName) <> -1
    then begin
         Repeat
         Inc(i);
         NewName := FieldName + '_' + IntToStr(i);
         Until FieldDefs.IndexOf(NewName) = -1;
         Result := NewName;
         end;
 end;
 
begin
Result := FieldDef.FieldName;
if FieldDefs.IndexOf(Result) <> -1
   then Result := FieldDef.FieldFullName;
if FieldDefs.IndexOf(Result) <> -1
   then Result := AddUniqueNumber(FieldDef.FieldFullName);

end;


{ ----- WriteMode ----- }

procedure TSqlitePassDataset.SetFWriteMode(Const Value: TSqlitePassWriteMode);
begin
if Active then FDatabase.DatabaseError.RaiseException(Msg2033);
FWriteMode := Value;
end;

procedure TSqlitePassDataset.ApplyChanges;
var
RecordState: TSqlitePassRecordState;

begin
// TODO, should be improved using FRecordset.FModifiedRecords Array...
if WriteMode = wmPostponed then
   begin
   FDatabase.FEngine.Transaction.Start;
   try
     First;
     While Not Eof do
      begin
      RecordState := FRecordset.GetRecordState(FRecordset.FCurrentRecordIndex);
      Case RecordState of
           rsInserted : FSQLSelectStmt.Post(ptInsert);
           rsModified : FSQLSelectStmt.Post(ptUpdate);
           rsDeleted  : FSQLSelectStmt.Post(ptDelete);
           end;
      Next;
      end;
      FDatabase.FEngine.Transaction.Commit;
   except
   FDatabase.FEngine.Transaction.Rollback;
   end;
   end;
end;

procedure TSqlitePassDataset.CancelChanges;
begin
  FRecordset.ClearModifiedRecords
end;

procedure TSqlitePassDataset.EmptyTable;
begin
  if DatasetType = dtTable
     then FDatabase.Engine.ExecSQL('DELETE FROM "' + DatasetName + '";');
  FRefetchRows := True;   
end;

{ ----- Queries params ----- Begin }

Procedure TSqlitePassDataset.SetParamsList(Value: TParams);
begin
 FParams.AssignValues(Value);
end;

Procedure TSqlitePassDataset.UpdateParamsList;
var
 NewParamsList: TParams;
begin
if not (csReading in ComponentState) then
    if ParamCheck or (csDesigning in ComponentState) then
    begin
        NewParamsList := TParams.Create(Self);
        try
            NewParamsList.ParseSQL(FSQLSelectStmt.SQL, True);
            NewParamsList.AssignValues(FParams);
            FParams.Clear;
            FParams.Assign(NewParamsList);
        finally
            NewParamsList.Free;
        end;
    end;
end;

Function TSqlitePassDataset.GetParamsCount: Integer;
begin
 Result := FParams.Count;
end;

Procedure TSqlitePassDataset.DefineProperties(Filer: TFiler);
  Function WriteData: Boolean;
  begin
      if Filer.Ancestor <> nil then
          Result := not FParams.IsEqual(TSqlitePassDataset(Filer.Ancestor).FParams)
      else
          Result := FParams.Count > 0;
  end;
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end;

Procedure TSqlitePassDataset.ReadParamData(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FParams);
end;

Procedure TSqlitePassDataset.WriteParamData(Writer: TWriter);
begin
  Writer.WriteCollection(Params);
end;

{ ----- Queries Params ----- End }

procedure TSqlitePassDataset.FreeRecordPointers(Buffer: PChar);
begin
{ TODO : Doesn't seem necessary to implement this procedure
  since no call was found to this procedure in VCL  }
end;

procedure TSqlitePassDataset.AllocateBLOBPointers(Buffer: PChar);
begin
{ TODO : Doesn't seem necessary to implement this procedure
  since no call was found to this procedure in VCL  }
end;

procedure TSqlitePassDataset.FreeBlobPointers(Buffer: PChar);
begin
{ TODO : TSqlitePassDataset.FreeBlobPointers
  Doesn't seem necessary to implement this procedure
  since Blobs are cleared by TSqlitePassMemRecord.Clear
  and no call was found to this procedure in VCL  }
end;

Procedure TSqlitePassDataset.ClearCalcFields(Buffer: PChar);
begin
FRecordset.ClearCalcFields(PInteger(Buffer)^);
{$IFDEF DEBUG}
FDatabase.Debugger.Display('[ --------- ClearCalcFields ------- ]');
{$ENDIF}
end;


Function TSqlitePassDataset.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
Begin
  Result:=TSqlitePassBlobStream.Create(TBlobField(Field),Mode);
End;

procedure TSqlitePassDataset.CloseBlob(Field: TField);
begin
{ TSqlitePassDataset.CloseBlob
  This procedure doesn't need to be implemented
  Blobs are kept in memory and freed by TSqlitePassRecordset }
end;


{ TSqlitePassDatasetIndex }

constructor TSqlitePassDatasetIndex.Create(Owner: TSqlitePassDatasetIndexDefs);
begin
inherited Create;
FIndexDefs := Owner;
end;

destructor TSqlitePassDatasetIndex.Destroy;
begin
FIndexDefs := nil;
inherited Destroy;
end;

procedure TSqlitePassDatasetIndex.DeleteIndex;
begin
 FindexDefs.FDataset.Database.Engine.ExecSQL('Drop index ' + FIndexName);
 FIndexDefs.FDataset.FDatabase.IndexDefs.Refresh;
end;

procedure TSqlitePassDatasetIndex.Reindex;
begin
 FIndexDefs.FDataset.FDatabase.Engine.ExecSQL('Reindex ' + FIndexName);
end;


{ TSqlitePassDatasetIndexDefs }

constructor TSqlitePassDatasetIndexDefs.Create(Owner: TSqlitePassDataset);
begin
 inherited Create;
 FDataset := Owner;
end;

destructor TSqlitePassDatasetIndexDefs.Destroy;
begin
 FDataset := nil;
 inherited Destroy;
end;

function TSqlitePassDatasetIndexDefs.GetItem(Index: Integer): TSqlitePassIndex;
begin
 Result:=TSqlitePassIndex(Inherited Items[Index]);
end;

procedure TSqlitePassDatasetIndexDefs.SetItem(Index: Integer;
  const Value: TSqlitePassIndex);
begin
 Put(Index,Value);
end;

procedure TSqlitePassDatasetIndexDefs.RefreshFromCache;
var
i: Integer;
TableName: String;
begin
{ Clears only the list, not the items which are pointing to Database indexdefs }
Clear;
TableName := FDataset.FDatasetName;
if TableName <> ''
   then begin
        for i := 0 to Pred(FDataset.FDatabase.IndexDefs.count)
         do begin
            if AnsiCompareText(FDataset.FDatabase.IndexDefs[i].TableName, TableName) = 0
               then Add(FDataset.FDatabase.IndexDefs[i]);
            end;
   end;
end;


{ TBlobStream }

Constructor TSqlitePassBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
var
FBlobFieldIndex: Integer;
Buffer: PChar;
Begin
 FField := Field;
 FDataset := TSqlitePassDataset(FField.DataSet);
 FActiveRecIndex := FDataset.GetActiveRecord;
 if FActiveRecIndex = -1 then Exit;

 if FDataset.FRecordset.FieldIsNull(FActiveRecIndex, Field.Tag)
    then FBlobFieldIndex := FDataset.FRecordset.AddMemoryStream(FActiveRecIndex, Field.Tag)
    else begin
         Buffer := FDataset.FRecordset.GetFieldBuffer(FActiveRecIndex, Field.Tag);
         FBlobFieldIndex := PInteger(Buffer)^;
         end;

 FRecBlobStream := TMemoryStream(FDataset.FRecordset.FMemoryStreams[FBlobFieldIndex]);
 FOpened := True;

 Case Mode of
      bmRead  : LoadFromStream(FRecBlobStream);
      bmWrite : begin
                FRecBlobStream.Clear;
                FModified := True;
                end;
      end;
end;

Destructor TSqlitePassBlobStream.Destroy;
Begin
 if FModified then
   try
     SaveToStream(FRecBlobStream);
     FDataset.FRecordset.SetFieldNullValue(FActiveRecIndex, FField.Tag, FRecBLobStream.Size = 0);
     FDataset.FRecordset.SetFieldWasModified(FActiveRecIndex, FField.Tag, True);
     FField.Modified := True;
     FDataSet.DataEvent(deFieldChange, Longint(FField));
     FOpened := False;
     FModified := False;
   Except
     Application.HandleException(Self);
   End;
inherited Destroy;
end;

function TSqlitePassBlobStream.Read(var Buffer; Count: Longint): Longint;
begin
 Result := 0;
 if FOpened
    then Result := inherited Read(Buffer, Count);
end;

Function TSqlitePassBlobStream.Write(const Buffer; Count: Longint): Longint;
Begin
 Result := 0;
 if FOpened then
    begin
    Result := inherited Write(Buffer, Count);
    FModified := True;
    end;
end;





