
{ Unit SqlitePassDbo.

  This unit defines all the interface definitions for the SqlitePass Project


  ---------------------------------------------------------------------------

  This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    @Author(Luc DAVID 2006-2009 luckylazarus@free.fr)
    @Author(http://source.online.free.fr )

    Major changes are indicated in the \Documentation\Changes.pdf file
    @created(2006)
    @lastmod(05.01.2009)

  --------------------------------------------------------------------------- }


unit SqlitePassDbo;
{$i SqlitePassDbo.inc}

{ Turns on/off debug output to Debuger Event Log}
{.$DEFINE DEBUG}
{.$DEFINE DEBUG_SQLStmt}

Interface

uses
 Classes, SysUtils, Db, Forms, Dialogs, Controls, Math,
 SqlitePassConst,
 SqlitePassErrorLang,
 SqlitePassKexiDef,
 SqlitePassApi_v3,
 SqlitePassUtils,

{$IFDEF DEBUG}
 SqlitePassDebug,
{$ENDIF}

{$IFDEF FPC}
 Variants, LCLIntf, DbConst,
{$ELSE}
  {$IFDEF Delphi6}
   Variants,
  {$ENDIF}
  DbConsts,
  DBCommon,
{$ENDIF}

{$IFDEF WIN32}
 Windows;
{$ELSE}
 Dynlibs;
{$ENDIF}

Const
{$IFDEF WIN32}
  {$DEFINE extdecl:=stdcall}
  DefaultSQLiteLibrary = 'sqlitepass3.dll';
{$ELSE}
  {$DEFINE extdecl:=cdecl}
  DefaultSQLiteLibrary = 'libsqlitepass3.so';
{$ENDIF}

 JulianEpoch = -2415018.5;
 UnixEpoch = JulianEpoch + 2440587.5;

Type
 { forward declarations }
 TSqlitePassEngine = class;
 TSqlitePassDatabase = Class;
 TSqlitePassDataset = class;
 TSqlitePassDatasets = class;
 TSqlitePassTableDef = class;
 TSqlitePassTableDefs = class;
 TSqlitePassDatabaseIndexDefs = class;
 TSqlitePassQueryDefs = class;
 TSqlitePassViews = class;
 TSqlitePassTriggers = class;
 TSqlitePassTableFieldDef = class;
 TSqlitePassTableFieldDefs = class;
 TSqlitePassIndexColumns = class;
 TSqlitePassIndex = class;
 TSqlitePassDatasetIndex = class;
 TSqlitePassTranslator = class;
 TSqlitePassDatabasesAttached = class;
 TSqlitePassSelectStmtFieldDefs = class;
 TSqlitePassSelectStmtTableDef = class;
 TSqlitePassSelectStmtTableDefs = class;
 TSqlitePassRecordset = class;
 TSqlitePassSQLStmtDefs = class;
 TSqlitePassSelectStmt = class;
 TSqlitePassSQLStmt = class;
 TSqlitePassSQLTokenizer = class; 

 TSqlitePassDatabaseDataTypeOptions = class;
 TSqlitePassVerboseLevel = vlSilent..vlLogAndShow;
 TSqlitePassAvailableOperations = set of (aoRead, aoInsert, aoUpdate, aoDelete);
 TSqlitePassSQLChanges = set of (scSelect, scWhere, scGroupBy, scHaving, scOrderBy, scLimit, scRefresh);
 TSqlitePassDatabaseObjectTypes =(dboAll, dboTableDefs, dboQueryDefs, dboIndexDefs, dboViews, dboTriggers);
 TSqlitePassDatabaseType = (dbtSqlitePass, dbtKexi, dbtSqliteAdmin, dbtSqlite4Fpc, dbtSqliteExpert, dbtUnknown);
 TSqlitePassDatabaseState = set of (dbsInternalOpen, dbsRefreshingDefinitions);
 TSqlitePassPostType = (ptInsert, ptUpdate, ptDelete);
 TSqlitePassQuoteStyle = (qsSingleQuote, qsDoubleQuote, qsBracket, qsNone);
 TSqlitePassSortMode = (smQuickSort, smSQL);
 TSqlitePassWriteMode = (wmDirect, wmPostponed);
 PPointer= ^Pointer;
 TSqlitePassExecCallBack=function(acolumns: Integer; aColumnValues, aColumnNames:Ppchar):integer of object;
 PSqlitePassExecCallBack=^TSqlitePassExecCallBack;
 TSqlitePassGetPragmaCallback=function(UserData: Pointer; ColumnCount: Integer; ColumnValues, ColumnNames: PPointer): Integer; cdecl;
 TSqlitePassRecordState=(rsUnModified, rsModified, rsInserted, rsDeleted);

  PCardinal = ^Cardinal;
  PDateTimeRec = ^TDateTimeRec;
  PByteBool = ^ByteBool;
  PWordBool = ^WordBool;

 { TVersionInfo }
 TSqlitePassDatabaseVersionInfo = Class(TPersistent)
     Private
         FComponent   : String;
         FDatabase    : TSqlitePassDatabase;
         FPackage     : String;
         FDbVersion   : String;
         FSchema    : Integer;
         FSqliteLib   : String;
         FUserTag     : Integer;
         function GetFDbSchema: Integer;
         function GetFSqliteEngine: String;
         function GetFUserTag: Integer;
         procedure SetFDbVersion(const Value: String);
         procedure SetFPackage(const Value: String);
         procedure SetFDbSchema(const Value: Integer);
         procedure SetFSqliteEngine(const Value: String);
         procedure SetFUserTag(const Value: Integer);
         procedure SetFComponent(const Value: String);
     Public
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
     Published
         Property Component     : String  Read FComponent       Write SetFComponent;
         Property Schema        : Integer Read GetFDbSchema     Write SetFDbSchema;
         Property DbVersion     : String  Read FDbVersion       Write SetFDbVersion;
         Property Package       : String  Read FPackage         Write SetFPackage;
         Property SqliteLibrary : String  Read GetFSqliteEngine Write SetFSqliteEngine;
         Property UserTag       : Integer Read GetFUserTag      Write SetFUserTag;
     end;

TSqlitePassAutoVacuumType = 0..2; { PRAGMA auto_vacuum =  0 | none | 1 | full | 2 | incremental }
TSqlitePassSynchronous = (Off, Normal, Full);
TSqlitePassTempStore = (Default, ToFile, ToMemory);
TSqlitePassEncoding = (UTF_8, UTF_16, UTF_16le, UTF_16be);
TSqlitePassPageSize = 512..32768;
TSqlitePassRecordsCacheCapacity = 1..MaxInt;

 { TSqlitePassDatabaseOptions }
 TSqlitePassDatabaseOptions = Class(TPersistent)
    Private
         FDatabase: TSqlitePassDatabase;
         FAutoVacuum: TSqlitePassAutoVacuumType;
         FCacheSize: Integer;
         FDefaultCacheSize: Integer;
         FCaseSensitiveLike: Boolean;
         FCountChanges: Boolean;
         FEmptyResultCallback: Boolean;
         FEncoding: TSqlitePassEncoding;
         FFullColumnNames: Boolean;
         FLogErrors: Boolean;
         FQuoteStyle: TSqlitePassQuoteStyle;
         FQuoteOpen: Char;
         FQuoteClose: Char;
         FShortColumnNames: Boolean;
         FPageSize: TSqlitePassPageSize;
         FSynchronous : TSqlitePassSynchronous;
         FTemporaryStorage: TSqlitePassTempStore;
         function GetFAutoVacuum: TSqlitePassAutoVacuumType;
         function GetFCacheSize: Integer;
         function GetFCountChanges: Boolean;
         function GetFDefaultCacheSize: Integer;
         function GetFEmptyResultCallBack: Boolean;
         function GetFEncoding: TSqlitePassEncoding;
         function GetFFullColumnNames: Boolean;
         function GetFPageSize: TSqlitePassPageSize;
         function GetFSynchronous: TSqlitePassSynchronous;
         function GetFTemporaryStorage: TSqlitePassTempStore;
         procedure SetFAutoVacuum(const Value: TSqlitePassAutoVacuumType);
         procedure SetFCacheSize(const Value: Integer);
         procedure SetFCaseSensitiveLike(const Value: Boolean);
         procedure SetFCountChanges(const Value: Boolean);
         procedure SetFDefaultCacheSize(const Value: Integer);
         procedure SetFEmptyResultCallBack(const Value: Boolean);
         procedure SetFEncoding(const Value: TSqlitePassEncoding);
         procedure SetFFullColumnNames(const Value: Boolean);
         procedure SetFPageSize(Value: TSqlitePassPageSize);
         procedure SetFQuoteStyle(const Value: TSqlitePassQuoteStyle);
         procedure SetFSynchronous(const Value: TSqlitePassSynchronous);
         procedure SetFTemporaryStorage(const Value: TSqlitePassTempStore);
    Public
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
    Published
         Property AutoVacuum: TSqlitePassAutoVacuumType Read GetFAutoVacuum Write SetFAutoVacuum;
         Property CacheSize: Integer Read GetFCacheSize Write SetFCacheSize;
         Property CaseSensitiveLike: Boolean Read FCaseSensitiveLike Write SetFCaseSensitiveLike;
         Property CountChanges: Boolean Read GetFCountChanges Write SetFCountChanges;
         Property DefaultCacheSize: Integer Read GetFDefaultCacheSize Write SetFDefaultCacheSize;
         Property EmptyResultCallback: Boolean Read GetFEmptyResultCallBack Write SetFEmptyResultCallBack;
         Property Encoding: TSqlitePassEncoding Read GetFEncoding Write SetFEncoding;
         Property FullColumnNames: Boolean Read GetFFullColumnNames Write SetFFullColumnNames;
         Property LogErrors: Boolean Read FLogErrors Write FLogErrors;
         Property PageSize: TSqlitePassPageSize Read GetFPageSize Write SetFPageSize;
         Property QuoteStyle: TSqlitePassQuoteStyle Read FQuoteStyle Write SetFQuoteStyle;
         Property Synchronous : TSqlitePassSynchronous Read GetFSynchronous Write SetFSynchronous;
         Property TemporaryStorage: TSqlitePassTempStore Read GetFTemporaryStorage Write SetFTemporaryStorage;
    end;

 { Datatypes detection and conversion }
 TSqlitePassBooleanStorage = (asInteger, asText);
 TSqlitePassDateTimeStorage = (dtsDateTime, dtsJulian, dtsText, dtsUnix, dtsMac);
 TSqlitePassDateStorage = TSqlitePassBooleanStorage;
 TSqlitePassTimeStorage = TSqlitePassDateStorage;
 TSqlitePassDataTypeDetectionMode = (dmTypeName, dmDbSpecific, dmCustom, dmForceStr);
 TSqlitePassDataTypeMappingMode = (mmExact, mmExactNoCase, mmPartial, mmPartialNoCase, mmAll);
 TSqlitePassDataTypeStorageOptions = set of (soProperties, soTranslationRules, soCustomFieldDefs, soManual);


 { TSqlitePassFieldTypeTranslationRule.
   Database FieldTypes mapping to Pascal FieldTypes }
 TSqlitePassFieldTypeTranslationRule = Class
 Private
 FDatatypeName: String;
 FMappingMode: TSqlitePassDataTypeMappingMode;
 FFieldType: TFieldType;
 Public
 Property DatatypeName: String Read FDatatypeName Write FDatatypeName;
 Property MappingMode: TSqlitePassDataTypeMappingMode Read FMappingMode Write FMappingMode;
 Property FieldType: TFieldType Read FFieldType Write FFieldType;
 end;

 { TSqlitePassFieldTypesTranslationRules }
 TSqlitePassFieldTypesTranslationRules = Class(TSqlitePassObjectList)
  Private
    FDataTypeOptions: TSqlitePassDatabaseDataTypeOptions;
  Protected
    Function GetItem(Index: Integer): TSqlitePassFieldTypeTranslationRule;
    Procedure SetItem(Index: Integer; FieldTypeTranslationRule: TSqlitePassFieldTypeTranslationRule);
  Public
    constructor Create(DataTypeOptions: TSqlitePassDatabaseDataTypeOptions);
    Destructor Destroy; override;
    Function Add(FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule): Integer;
    procedure ClearAndFreeItems;
    Procedure SetDefaultTranslationRules(DatabaseType: TSqlitePassDatabaseType);
    procedure SetFieldDataType(FieldDef: TSqlitePassTableFieldDef);
    Property Items[Index: Integer]: TSqlitePassFieldTypeTranslationRule read GetItem write SetItem; default;
  end;
                         
 { TSqlitePassCustomFiedlDef }
 TSqlitePassCustomFieldDef = Class
 Private
 FFieldName: String;
 FFieldType: TFieldType;
 FFieldSize: Integer;
 FFieldPrecision: Integer;
 FTableName: String;
 Public
 Property TableName: String Read FTableName Write FTableName;
 Property FieldName: String Read FFieldName Write FFieldName;
 Property FieldType: TFieldType Read FFieldType Write FFieldType;
 Property FieldSize: Integer Read FFieldSize Write FFieldSize;
 Property FieldPrecision: Integer Read FFieldPrecision Write FFieldPrecision;
 end;

 { TSqlitePassCustomFieldDefs }
 TSqlitePassCustomFieldDefs = Class(TSqlitePassObjectList)
  Protected
    Function GetItem(Index: Integer): TSqlitePassCustomFieldDef;
    Procedure SetItem(Index: Integer; CustomFieldDef: TSqlitePassCustomFieldDef);
  Public
    Constructor Create;
    Destructor Destroy; override;
    Function Add(CustomFieldDef: TSqlitePassCustomFieldDef): Integer;
    procedure ClearAndFreeItems;
    Property Items[Index: Integer]: TSqlitePassCustomFieldDef read GetItem write SetItem; default;
  end;

 {
  The TSqlitePassDatabaseDataTypeOptions is used by the TSqlitePassDatabase.DatatypeOptions property and
  gives you the opportunity to set custom or default behaviors for a given database.
  It lets you also define how fields.datatype will be retrieved and translated
  into pascal datatypes using TranslationRules and CustomFieldDefs.}
  TSqlitePassDatabaseDataTypeOptions = Class(TPersistent)
      Private
         {Year-Month-Day}
         FDateFormatPattern : Array[0..2, 0..1] of Word;
         {Hour-Minute-Sec-Msec}
         FTimeFormatPattern : Array[0..3, 0..1] of Word;
         {Year-Month-Day-Hour-Minute-Sec-Msec}
         FDateTimeFormatPattern : Array[0..6, 0..1] of Word;
         {True Values - False Values}
         FBooleanTrueValues    : TStringList;
         FBooleanFalseValues   : TStringList;
         FBooleanFormat        : String;
         FBooleanStorage       : TSqlitePassBooleanStorage;
         FCustomFieldDefs      : TSqlitePassCustomFieldDefs;
         FDatabase             : TSqlitePassDatabase;
         FDateFormat           : String;
         FDateStorage          : TSqlitePassDateStorage;
         FDateTimeFormat       : String;
         FDateTimeStorage      : TSqlitePassDateTimeStorage;
         FDecimalSeparator     : Char;
         FDefaultFieldType     : TFieldType;
         FDetectionMode        : TSqlitePassDataTypeDetectionMode;
         FFieldTypesTranslationRules : TSqlitePassFieldTypesTranslationRules;
         FLoadOptions          : TSqlitePassDataTypeStorageOptions;
         FSaveOptions          : TSqlitePassDataTypeStorageOptions;
         FTimeFormat           : String;
         FTimeStorage          : TSqlitePassTimeStorage;
         FStorageVersion       : String;
         procedure FindSubString(Chr: Char; LowerStr: String; Var StartPos, EndPos: Word);
         procedure SetFDateFormat(const Value: String);
         procedure SetFTimeFormat(const Value: String);
         procedure SetFDateTimeFormat(const Value: String);
         procedure SetFBooleanFormat(const Value: String);
         procedure ReadTranslationRules(Reader: TReader);
         procedure WriteTranslationRules(Writer: TWriter);
         procedure ReadCustomFieldDefs(Reader: TReader);
         procedure WriteCustomFieldDefs(Writer: TWriter);
        Protected
         procedure DefineProperties(Filer: TFiler); override;
        Public
          { ----- Constructor / Destructor ----- }
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
         { --- }
         Procedure ApplyCustomFieldDefs;
         function  GetOwner: TPersistent; override;
         procedure SetDefaultValues(DatabaseType: TSqlitePassDatabaseType; SetTranslationRules: Boolean=false);
         Procedure LoadFromDatabase(_LoadOptions: TSqlitePassDataTypeStorageOptions = []);
         Procedure SaveToDatabase(_SaveOptions: TSqlitePassDataTypeStorageOptions = []);
       Published
         { ----- Properties ----- }
         // TODO -> Disable format changes when database is active

           { a string used as a Boolean formating template.
           The BooleanFormat is only active when the BooleanStorage property is set to "asText".
           In this case it is used to read or write boolean values stored as text to the database
           BooleanFormat is separated into two sections :
           The left part represents the text statements used for True,
           a separator '-'
           The right part represents the text statements used for False

           @italic(
           Usualy BooleanFormat takes one of the following value :
           'Yes,True,On,1,-No,False,Off,0')

           When BooleanFormat is set to "asText"
           the first value of the "True" section ('Yes' in the example)
           and the first value of the "False" sections ('No' in the example)
           are used as text representation to write boolean value to database}
         Property BooleanFormat        : String Read FBooleanFormat Write SetFBooleanFormat;

         {BooleanStorage determines how the Boolean values are stored in the database.
          It can take one of the following values :
          asInteger : the field value is stored as an integer.
          asText : the field value is stored as formated text, using the BooleanFormat property setting}
         Property BooleanStorage       : TSqlitePassBooleanStorage Read FBooleanStorage Write FBooleanStorage;

         {CustomFieldDefs : A collection of custom fielddefs.

         A custom fielddef can be used to overwrite the fielddefs translation
         already done by the DetectionMode and the TranslationRules properties.

         A custom fielddef is made of :
           A TableName and a FieldName to identify the field
           A FieldType: TfieldType (ftinteger, ftstring...etc)
           A FieldSize
           A FieldPrecision

         This property can be very usefull when a fielddef cannot be built on the datatype.
         @italic(For example, you created a table with two BLOB fields declared with the
         same datatype (CLOB) in the table create statement.
         If you want to store MemoText in the first blob and store graphic data in the second, you just
         need to create two custom fielddefs using the designtime editor
         or the sqliteToolbox program and store then in your database using LoadOptions and SaveOptions.)

         At design time, a specific editor lets you define the translation rules.}
         Property CustomFieldDefs      : TSqlitePassCustomFieldDefs Read FCustomFieldDefs Write FCustomFieldDefs;

         { a string used as a Date formating template.
           The DateFormat is only active when the DateStorage property is set to asText.
           In this case it is used to read or write date values stored as text to the database
           DateFormat accepts @bold(only Uppercases) to match Day (D), Month (M), Year (Y)
           Any other character is treated as literal and is included is the formated string

           @italic(
           usualy DateFormat takes one of the following value :
           DD-MM-YYYY
           YYYY-MM-DD
           DD/MM/YYYY
           YYYY/MM/DD)
            }
         Property DateFormat           : String Read FDateFormat Write SetFDateFormat;

         {DateStorage determines how the Date values are stored in the database.
          It can take one of the following values :
          asInteger : the field value is stored as an integer.
          asText : the field value is stored as formated text, using the DateFormat property setting}
         Property DateStorage          : TSqlitePassDateStorage Read FDateStorage Write FDateStorage;

         { a string used as a DateTime formating template.
           The DateTimeFormat is only active when the DateTimeStorage property is set to "asText".
           In this case it is used to read or write datetime value stored as text to the database
           DateFormat accepts @bold(only Uppercases for the Date part and only lowercases for the Time part)
           to match Days (D), Months (M), Years (Y), Hours (h), Minutes (m), Secondes (s) and Msecs (z)
           Any other character is treated as literal and is included is the formated string

           @italic(
           usualy DateFormat takes one of the following value :
           DD-MM-YYYY hh:mm:ss
           DD/MM/YYYY hh:mm:ss
           DD-MM-YYYY hh:mm:ss:zzz
           DD/MM/YYYY hh:mm:ss:zzz)
           Note : Kexi databases format DateTime as DD-MM-YYYYThh:mm:ss
           )
            }
         Property DateTimeFormat       : String Read FDateTimeFormat Write SetFDateTimeFormat;

         {DateTimeStorage determines how the DateTime values are stored in the database.
          It can take one of the following values :
          dtsDateTime : the field value is stored as a double value.
          dtsJulian : the field value is stored as a double value converted to a julian date.
          dtsText : the field value is stored as formated text, using the DateTimeFormat property setting
          dtsUnix : the field value is stored as a int64 value converted to an Unix date.
          dtsMac : the field value is stored as a int64 value converted to an Apple Mac date.}
         Property DateTimeStorage      : TSqlitePassDateTimeStorage Read FDateTimeStorage Write FDateTimeStorage;

         { Decimal separator : not used}
         Property DecimalSeparator     : Char Read FDecimalSeparator Write FDecimalSeparator;

         { Determines the default field type used when field type translations fails.
           It is usually set to ftUnknow or ftString}
         Property DefaultFieldType     : TFieldType Read FDefaultFieldType Write FDefaultFieldType;

         {The DetectionMode controls how the TsqlitePassDatabase component will behave when it tries to retrieve the database fields definition.
         It can have on of the following values :

         @bold(dmTypeName)@br
         FieldTypes are retrieved from the initial SQL CREATE statement of the table for example :
         'CREATE TABLE cars_names (id INTEGER PRIMARY KEY, companycode Integer, name Text(200))'
         Then the database uses the TranslationRules (a collection of translation Rule) to match the datatypes names found in the database table,  for example 'integer' with a pascal datatype (ftInteger in this example).

         @bold(dmDbSpecific)@br
         Field.datatypes are set directly by sqlitepass, depending on the Database.Databasetype value.
         Note : This can only be used with the 'dbtKexi' Databasetype.

         @bold(dmCustom)@br
         Field.datatypes are first preset directly by sqlitepass, depending on the Database.Databasetype value.
         If the first match fails, it uses the TranslationRules if you defined some
         Finally, it fires the OnDataTypeConversion event letting you modified directly the Field.Datatype Of course, the OnDataTypeConversion Event must be assigned...

         @bold(dmForceStr)@br
         Converts any Datatype to ftString;}
         Property DetectionMode        : TSqlitePassDataTypeDetectionMode Read FDetectionMode Write FDetectionMode;

         {if SoManual is not selected (default behavior) your optional settings are retrieved directly from
          a special table named "SQLitePass__DbSettings" every time you open a database

           The LoadOptions property determines which parts of the database DataTypeOptions are retrieved from the table
           It can take one or more of the following values :
           soProperties : It retrieves your settings for
                        BooleanExtension;
                        BooleanStorage;
                        DateFormat;
                        DateStorage;
                        DateTimeFormat;
                        DateTimeStorage;
                        DecimalSeparator;
                        DefaultFieldType;
                        DetectionMode;
                        TimeFormat;
                        TimeStorage;

           soTranslationRules : It retrieves and fills the TranslationRules collection.
           soCustomFieldDefs : It retrieves and fills the CustomFieldDefs collection.

           soManual : The selected parts are not retrieved automatically.
           You will then have to call LoadFromDatabase to load your database settings.
           }
         Property LoadOptions          : TSqlitePassDataTypeStorageOptions Read FLoadOptions Write FLoadOptions;

         { a string used as a Time formating template.
           The TimeFormat is only active when the DateStorage property is set to "asText".
           In this case it is used to read or write Time values stored as text to the database
           DateFormat accepts @bold(only lowercases) to match Hours (h), Minutes (m), Secondes (s) and Msecs (z)
           Any other character is treated as literal and is included is the formated string
           @italic(
           usualy TimeFormat takes one of the following value :
           hh:mm:ss
           hh:mm:ss:zzz
           hh-mm-ss
           hh-mm-ss-zzz)}
         Property TimeFormat           : String Read FTimeFormat Write SetFTimeFormat;

         {TimeStorage dtermines how the Time values are stored in the database.
          It can take one of the following values :
          asInteger : the field value is stored as an integer.
          asText : the field value is stored as formated text, using the TimeFormat property setting}
         Property TimeStorage          : TSqlitePassTimeStorage Read FTimeStorage Write FTimeStorage;

         {if SoManual is not selected (default behavior) your optional settings are stored directly to
          a special table named "SQLitePass__DbSettings" every time you close a database.

           The SaveOptions property determines which parts of the database DataTypeOptions are saved from the table
           It can take one or more of the following values :
           soProperties : It saves your settings for
                        BooleanExtension;
                        BooleanStorage;
                        DateFormat;
                        DateStorage;
                        DateTimeFormat;
                        DateTimeStorage;
                        DecimalSeparator;
                        DefaultFieldType;
                        DetectionMode;
                        TimeFormat;
                        TimeStorage;

           soTranslationRules : It saves the the TranslationRules collection to the database.
           soCustomFieldDefs : It saves the CustomFieldDefs collection to the database.

           soManual : The selected parts are not saved automatically.
           You will then have to call SaveToDatabase to save your database settings.
           }
         Property SaveOptions          : TSqlitePassDataTypeStorageOptions Read FSaveOptions Write FSaveOptions;

         {TranslationRules : A collection of translation rules.
         TranslationRules are stored in *.DFM or *.LFM file but can also be stored directly in the database using
         the LoadOptions and SaveOptions properties.

         A translation rule is made of :
           A datatype name (whatever you want depending on how fields datatypes are named inside the database or the table create statement)
           A matching rule (mmExact, mmExactNoCase, mmPartial, mmPartialNoCase, mmAll);
           A resulting datatype (ft... pascal fieldtype)


         The matching rule can be set for each translation rule and can be one of the following values:

         @bold(mmExact)
         the datatype found in the database table definition must match exactly
         the rule datatype name. It is case sensitive.

         @italic(For example if the 'Int' name is found in the database table definition
         then it will not match the ('Integer',mmExact,ftInteger) rule thus ftUnknown
         will be returned)

         @bold(mmExactNoCase)
         Same as mmExact but not case sensitive

         @bold(mmPartial)
         the datatype found in the database table definition can partially match the rule datatype name.
         It is case sensitive.
         @italic(For example if the 'Int' name is found in the database table definition then
         it will match the ('Integer',mmExact,ftInteger) rule and will return ftInteger)

         @bold(mmPartialNoCase)
         Same as mmPartial but not case sensitive

         @bold(mmAll (Default))
         Tries to determine the datatype using mmExact.
         If it fails then another try is done using mmPartial
         If the matching rule failed, the DefaultFieldType is returned (ftUnknown as default).

         At design time, a specific editor lets you define the translation rules.}
         Property TranslationRules     : TSqlitePassFieldTypesTranslationRules Read FFieldTypesTranslationRules Write FFieldTypesTranslationRules;
      end;

 { }
 TSqlitePassDatabaseError = class
 Private
   FDatabase: TSqlitePassDatabase;
   FErrorCount: Integer;
   FErrorStringList: TStringList;
   FLastErrorCode: Integer;
   FLastErrorMsg: String;
   FLastErrorTime: TDateTime;
   procedure LogError(Msg: string; const Args: array of const; const ErrorCode: Integer = -1; Component: TComponent = nil);
   procedure RaiseException(const Msg: String; const ErrorCode: Integer = -1;  Component: TComponent = nil; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow);
   procedure RaiseExceptionFmt(const Msg: string; const Args: array of const; const ErrorCode: Integer = -1; Component: TComponent = nil; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow);
   function  ShowError(const Msg: string; DlgType: TMsgDlgType = mtWarning; DlgButtons: TMsgDlgButtons = [mbOk]; HelpCtx: Longint = 0): Word;
 Public
   { ----- Constructor / Destructor ----- }
   Constructor Create(Database: TSqlitePassDatabase);
   Destructor Destroy; override;
   { ----- Procedures / Functions ----- }
   procedure Clear;
   procedure GetLastError(var ErrorCode: Integer;var Msg: String;var Time: TDateTime);
   procedure SaveToFile(FileName: String);
   property ErrorCount: Integer Read FErrorCount;
   property ErrorList: TStringList Read FErrorStringList;
  end;


 { TSqlitePassDatabase Events }
 TConnectEvent = procedure(Database: TSqlitePassDatabase) of object;
 TDataTypeConversion = procedure (Database: TSqlitePassDatabase; TableDef: TSqlitePassTableDef;
                                  FiedDef: TSqlitePassTableFieldDef) of object;

 {The TSqlitePassDatabase component is the main link between your
  application and the sqlite library.
  It currently supports sqlite engine version 3.xx and gives you direct acces to
  a wide range of sqlite databases created using database management progams like kexi,
  sqlite administrator, sqliteToolbox...etc.
  With proper setting, you can also use it to read and write to your own custom sqlite databases.
  }
 TSqlitePassDatabase = Class(TComponent)
       Private
         {$IFDEF DEBUG}
         Debugger            : TSqlitePassDebugger;
         {$ENDIF}
         FConnected          : Boolean;
         FCollatingOrder     : String;
         FDatabase           : String;
         FDatabaseError      : TSqlitePassDatabaseError;
         FDatasets           : TSqlitePassDatasets;
         FTableDefs          : TSqlitePassTableDefs;
         FIndexDefs          : TSqlitePassDatabaseIndexDefs;
         FQueryDefs          : TSqlitePassQueryDefs;
         FViews              : TSqlitePassViews;
         FTriggers           : TSqlitePassTriggers;
         FSQLStmtDefs        : TSqlitePassSQLStmtDefs;
         { Temporary List used by RefreshDefinition to get Attached Databases info }
         FDatabasesList      : TList;
         { AttachedDatabases Object List }
         FDatabases          : TSqlitePassDatabasesAttached;
         FDatabaseType       : TSqlitePassDatabaseType;
         FDataTypeOptions    : TSqlitePassDatabaseDataTypeOptions;
         FEngine             : TSqlitePassEngine;
         FTranslator         : TSqlitePassTranslator;
         FOptions            : TSqlitePassDatabaseOptions;
         FQueryTimeout       : Integer;
         FReadOnly           : Boolean;
         FShowSysObjects     : Boolean;
         FSmartOpen          : Boolean;
         FState              : TSqlitePassDatabaseState;
         FUser               : Integer;
         FVersionInfo        : TSqlitePassDatabaseVersionInfo;
         FSQLiteLibrary      : String;

         { ----- Events ---- }
         FAfterConnect       : TConnectEvent;
         FAfterDisconnect    : TConnectEvent;
         FBeforeConnect      : TConnectEvent;
         FBeforeDisconnect   : TConnectEvent;
         FDataTypeConversion : TDataTypeConversion;

         { Private Functions and Procedures }
         function CheckCanOpen(VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Boolean;
         function DatabaseTypeFromFileName(FileName: String): TSqlitePassDatabaseType;
         Procedure SetFDatabase(Value : String);
         procedure SetFDatabaseType(const Value: TSqlitePassDatabaseType);
         Procedure SetFDatabaseVersion(Value : String);
         procedure SetFUser(const Value: Integer);
         Procedure SetFVersionInfo(Value : String);
         Function  GetFCollatingOrder:String;
         Procedure SetFShowSysObjects(Value : Boolean);
         Procedure SetFSQLiteLibrary(Value : String);
         Procedure SetFReadOnly(Value : Boolean);
         Procedure SetFConnected(Value : Boolean);
         function  GetFEngine: TSqlitePassEngine;
         procedure ClearDefinitions;
         procedure OpenAttachedDatabases;
         procedure CloseAttachedDatabases;
         { ------ Pragma ------ }
         procedure SetPragma(PragmaName:string; value:Boolean);overload;
         procedure SetPragma(PragmaName:string; value:integer);overload;
         procedure SetPragma(PragmaName, value:string);overload;
         function  GetStrPragma(PragmaName:string):String;
         function  GetIntPragma(PragmaName:string):integer;
       Protected
         Procedure Loaded; override;
       Public
         { ----- Constructor / Destructor ----- }

         Constructor Create(AOwner : TComponent); override;
         Destructor Destroy; override;

         { ----- Procedures / Functions ----- }

         Procedure Idle;

         {Same as Connected := True}
         Procedure Open;

         {Same as Connected := False }
         Procedure Close;


         Function Compact: Integer;
         Function CreateDatabase(DbName: String; DbType: TSqlitePassDatabaseType;
                                  Encoding: TSqlitePassEncoding = UTF_8; PageSize: TSqlitePassPageSize = 4096;
                                  AutoVacuum: TSqlitePassAutoVacuumType = 0): Boolean;
         Function DeleteDatabase(Const DbName: String): Boolean;

         {Enables to attach one or several foreign databases to the current one.
          The attached databases must be compatible with the current one
          (the databases must have been created with the same database manager application).
          Once a database is attached, its content becomes available as part of the current database.
          Then you can access to tables, queries... as if they were part of the main database.
          Usage : AttachDatabase(DatabasePath: String);}
         Function AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;

         {Detaches a previously attached database.}
         Function DetachDatabase(DatabaseAlias: String): Boolean;

         { Updates the database definition content (TableDefs, QueryDefs...) from file }
         procedure RefreshDefinitions;

         { Returns True if the Table is a system table }
         Function IsSystemTable(TableName: String): Boolean;
         { ----- Properties ----- }

         { A utility object to log errors }
         Property DatabaseError: TSqlitePassDatabaseError Read FDatabaseError;

         { Internal object to access the sqlite engine }
         Property Engine: TSqlitePassEngine Read GetFEngine;

         { translation object used to get or set data that depend on
           the DatabaseType property }
         Property Translator: TSqlitePassTranslator Read FTranslator;

         { A collection of attached database to the current one }
         property Databases: TSqlitePassDatabasesAttached Read FDatabases;

         { A collection of all the datasets that use the current database component }
         Property Datasets: TSqlitePassDatasets Read FDatasets Write FDatasets;

         {The IndexDefs property is a collection that
          gives you access to the indexes definitons stored in the database.}
         Property IndexDefs: TSqlitePassDatabaseIndexDefs  Read FIndexDefs Write FIndexDefs;

         {The QueryDefs property is a collection that
          gives you access to the queries definitons stored in the database.}
         property QueryDefs: TSqlitePassQueryDefs Read FQueryDefs Write FQueryDefs;

         property SQLStmtDefs:  TSqlitePassSQLStmtDefs Read FSQLStmtDefs Write FSQLStmtDefs;

         {The TableDefs property is a collection that
          gives you access to the tables definitons stored in the database.}
         property TableDefs: TSqlitePassTableDefs Read FTableDefs Write FTableDefs;

         {The ViewDefs property is a collection that
         gives you access to the views definitons stored in the database.}
         property Views: TSqlitePassViews Read FViews Write FViews;

         {The TriggerDefs property is a collection that
         gives you access to the Triggers definitons stored in the database.}
         property Triggers: TSqlitePassTriggers Read FTriggers Write FTriggers;
       Published
         { ----- Properties ----- }

         {Set this property to True to connect the database defined in the database property.
          Set it to False to disconnect the database and all the datasets associated with it.
          You can also verify if a database is connected using :
          @code(if MyDatabase.Connected then...)}
         Property Connected            : Boolean Read FConnected           Write SetFConnected Default False;

         {Represents the physical database file you want to connect. At design time, shows up a file selection dialog box.}
         Property Database             : String  Read FDatabase            Write SetFDatabase;


         {Represents the database type detected from the database file extension you are using.
          (*.kexi for kexi database for example).
          Once the database type is recognised, the TsqlitePassDatabase component sets this property and creates
          an internal translator to take care of the database specifications.
          You can also define this property by yourself if the file extension doesn't match the correct database type.}
         Property DatabaseType         : TSqlitePassDatabaseType Read FDatabaseType Write FDatabaseType;


         {One of the main difficulties when working with Sqlite databases is to detect
          and translate properly the fields.datatypes since sqlite datatype are not formely defined.
          SqlitePass implements several ways to define a datatype.
          The Database.DatatypeOptions property gives you the opportunity to set custom or
          default behaviors for a given database and lets you also define how fields.datatype
          will be retrieved and translated into pascal datatypes using TranslationRules and CustomFieldDefs}
         Property DatatypeOptions      : TSqlitePassDatabaseDataTypeOptions Read FDatatypeOptions Write FDatatypeOptions;

         {Represents the database optional settings.}
         Property Options              : TSqlitePassDatabaseOptions Read FOptions Write FOptions;


         Property QueryTimeout         : Integer Read FQueryTimeout        Write FQueryTimeout;

         { Same as classic dataset behavior }
         Property ReadOnly             : Boolean Read FReadOnly            Write SetFReadOnly default False;

         { }
         Property ShowSystemObjects    : Boolean Read FShowSysObjects      Write SetFShowSysObjects;

         {When set to True, the TsqlitePassDataset will automatically takes care of
          opening the database connection if the database is not connected.
          Set it to False to manually control the database connection.}
         Property SmartOpen            : Boolean Read FSmartOpen           Write FSmartOpen;

         {Represents an alternative library file to be used instead of the default one.
          By default, sqlitepassDatabase tries to use the sqlitepass3.dll or libsqlitepass3.so file
          located in the system path directory (..\windows\system32\ or ..\user\lib for example).
          Enter a complete library file path to use a different library.
          From version 0.28 TsqlitePassDatabase needs the sqlite library compiled
          with the ENABLE_METADATA precompiler directive.
          These libraries are available from http://source.online.free.fr
          or you can compile your own following the tutorial available on the same internet site.}
         Property SQLiteLibrary        : String  Read FSQLiteLibrary       Write SetFSQLiteLibrary;

         Property UserTag              : Integer Read FUser                Write SetFUser;
         Property VersionInfo          : TSqlitePassDatabaseVersionInfo  Read FVersionInfo   Write FVersionInfo;
          { ----- Events ----- }
         Property OnAfterConnect       : TConnectEvent Read FAfterConnect        Write FAfterConnect;
         Property OnAfterDisconnect    : TConnectEvent Read FAfterDisconnect     Write FAfterDisconnect;
         Property OnBeforeConnect      : TConnectEvent Read FBeforeConnect       Write FBeforeConnect;
         Property OnBeforeDisconnect   : TConnectEvent Read FBeforeDisconnect    Write FBeforeDisconnect;
         Property OnDataTypeConversion : TDataTypeConversion Read FDataTypeConversion Write FDataTypeConversion;
      end;


TSqlitePassGenericFieldDef = Class
  Private
   FNoConstraints: Boolean;
   FAutoInc: Boolean;
   FUnique: Boolean;
   FPrimaryKey: Boolean;
//   FInternalPrimaryKey: Boolean;
   FForeignKey: Boolean;
   FNotNull: Boolean;
   FNotEmpty: Boolean; { only legal for string-like and blob fields }
   FIndexed: Boolean;
   FDatabaseOrigineNameIndex: Integer;
   FTableOrigineNameIndex: Integer;
   FFieldName: String;
   FDisplayName: String;
   FSize: Word;
   FPrecision: Integer;
   FNativeDataType: String;
   FNativeDataTypeCode: Integer;
   FDataType: TFieldType;
   FDefaultValue: String;
   FTableFieldNo: Integer;
   FRecordFieldIndex: Integer;
   FHint: String;
   procedure Assign(Source: TSqlitePassGenericFieldDef);
  Public
   Property AutoInc: Boolean Read FAutoInc;
   { write is enabled to force a DataType conversion }
   Property DataType: TFieldType Read FDataType Write FDataType;
   Property DefaultValue: String Read FDefaultValue;
   { write is enabled to enable changing default display name }
   Property DisplayName: String Read FDisplayName Write FDisplayName;
   Property FieldName: String Read FFieldName;
   Property TableFieldNo: Integer Read FTableFieldNo;
   Property ForeignKey: Boolean Read FForeignKey;
   { write is enabled to enable changing default hint message }
   Property Hint: String Read FHint Write FHint;
   Property Indexed: Boolean Read FIndexed;
//   Property InternalPrimaryKey: Boolean Read FInternalPrimaryKey;
   Property NativeDataType: String Read FNativeDataType;
   Property NativeDataTypeCode: Integer Read FNativeDataTypeCode Write FNativeDataTypeCode;
   Property NotEmpty: Boolean Read FNotEmpty;
   Property NoConstraints: Boolean Read FNoConstraints;
   Property NotNull: Boolean Read FNotNull;
   Property Precision: Integer Read FPrecision;
   Property PrimaryKey: Boolean Read FPrimaryKey;
   { The internal FieldIndex(Field Position in TSqlitePassRecordset Record) }
   Property RecordFieldIndex: Integer Read FRecordFieldIndex;
   Property Size: Word Read FSize;
   Property Unique: Boolean Read FUnique;
  end;

TSqlitePassTableFieldDef = Class(TSqlitePassGenericFieldDef)
  Private
  FFieldDefs: TSqlitePassTableFieldDefs;
  function GetFFieldFullName: String;
  Public
  Constructor Create(Owner: TSqlitePassTableFieldDefs);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableFieldDef);
  Property FieldFullName: String Read GetFFieldFullName;
  Property FieldDefs: TSqlitePassTableFieldDefs Read FFieldDefs;
  end;

TSqlitePassSelectStmtFieldDef = Class(TSqlitePassGenericFieldDef)
  Private
  FFieldDefs: TSqlitePassSelectStmtFieldDefs;
   { Indicate wether or not the field need to be used in Update Stmt }
  FWasModified: Boolean;
  function GetFFieldFullName: String;
  Public
  Constructor Create(Owner: TSqlitePassSelectStmtFieldDefs);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableFieldDef);
  Property FieldFullName: String Read GetFFieldFullName;
  Property FieldDefs: TSqlitePassSelectStmtFieldDefs Read FFieldDefs;
  end;

TSqlitePassTableFieldDefs = class(TSqlitePassObjectList)
  private
  FTableDef: TSqlitePassTableDef;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassTableFieldDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassTableFieldDef);
  public
  Constructor Create(Owner: TSqlitePassTableDef);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableFieldDefs);
  function FieldDefByName(const Value: string): TSqlitePassTableFieldDef;
  function FindFieldDef(const Value: string): TSqlitePassTableFieldDef;
  property Count;
  property Items[Index: Integer]: TSqlitePassTableFieldDef read GetItem write SetItem; default;
  property TableDef: TSqlitePassTableDef read FTableDef;
  end;


TSqlitePassSelectStmtFieldDefs = class(TSqlitePassObjectList)
  private
  FTableDef: TSqlitePassSelectStmtTableDef;
  function GetItem(Index: Integer): TSqlitePassSelectStmtFieldDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSelectStmtFieldDef);
  procedure ClearAndFreeItems;
  public
  Constructor Create(Owner: TSqlitePassSelectStmtTableDef);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableFieldDefs);
  property Count;
  property Items[Index: Integer]: TSqlitePassSelectStmtFieldDef read GetItem write SetItem; default;
  property TableDef: TSqlitePassSelectStmtTableDef read FTableDef;
  end;
  

{ TSqliteAttachedDatabases : Maintain a list of all attached databases }

TSqlitePassDatabaseAttached = Class(TObject)
       Private
         FAliasName          : String;
         FDatabase           : String;
         FDatabasesList      : TSqlitePassDatabasesAttached;
       Public
         { ----- Constructor / Destructor ----- }
         Constructor Create(Databases: TSqlitePassDatabasesAttached);
         Destructor Destroy; override;
         { ----- Properties ----- }
         Property AliasName: String Read FAliasName;
         Property Database: String  Read FDatabase Write FDatabase;
      end;

TSqlitePassDatabasesAttached = Class(TSqlitePassObjectList)
  Private
   FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDatabaseAttached;
    Procedure SetItem(Index: Integer; Database: TSqlitePassDatabaseAttached);
    procedure DetachAll;
  public
    constructor Create(Owner: TSqlitePassDatabase);
    destructor Destroy; override;
    function DatabaseByName(Const Name: String): TSqlitePassDatabaseAttached;
    function FindDatabase(const Value: string): TSqlitePassDatabaseAttached;
    property Items[Index: Integer]: TSqlitePassDatabaseAttached read GetItem write SetItem; default;
    property Count;
  end;


{ TableDefs }

TSqlitePassGenericTableDef = class
  private
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  FAliasName: String;
  FCanModify: Boolean;
  FDatabaseOrigineNameIndex: Integer;
  FOnConflict: String;
  FSql: String;
  FTableName: String;
  FTemporary: Boolean;
  function GetAttached: Boolean;
  function GetTableFullName: String;
  Public
  procedure Assign(Source: TSqlitePassGenericTableDef; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  Property AliasName: String Read FAliasName;
  property Attached: Boolean Read GetAttached;
  Property CanModify: Boolean Read FCanModify;
  Property OnConlict: String Read FOnConflict;
  property Sql: String Read FSql;
  property TableName: String Read FTableName;
  property TableFullName: String Read GetTableFullName;
  Property Temporary: Boolean Read FTemporary;
  end;

TSqlitePassTableDef = class(TSqlitePassGenericTableDef)
  private
  { Owner }
  FTableDefs: TSqlitePassTableDefs;
  { FieldDefs }
  FFieldDefs: TSqlitePassTableFieldDefs;
  Public
  Constructor Create(Owner: TSqlitePassTableDefs);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableDef; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure DeleteTable(Silent: Boolean = False);
  procedure RenameAs(NewName: String);
  procedure Reindex;
  property FieldDefs: TSqlitePassTableFieldDefs Read FFieldDefs;
  end;

TSqlitePassSelectStmtTableDef = class(TSqlitePassGenericTableDef)
  private
  { Owner }
  FTableDefs: TSqlitePassSelectStmtTableDefs;
  { Prepared sql statements to speed up operations }
  FInsertStmt: Pointer;
  FDeleteStmt: Pointer;
  FInsertStmtText: String;
  FDeleteStmtText: String;
  { InternalPrimaryKey }
  FLastInsertedRowId: Int64;
//  FPrimaryKeyField: TSqlitePassSelectStmtFieldDef;
  FAvailableOperations: TSqlitePassAvailableOperations;
  { FieldDefs }
  FFieldDefs: TSqlitePassSelectStmtFieldDefs;
  procedure BufferToSqliteValue(Const PreparedStmt: Pointer; Const DataType: TFieldType; RecordIndex, RecordFieldIndex, StmtFieldNo: Integer);
  public
  Constructor Create(Owner: TSqlitePassSelectStmtTableDefs);
  Destructor Destroy; override;
//  procedure Assign(Source: TSqlitePassSelectStmtTableDef; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure Assign(Source: TSqlitePassTableDef);
  procedure PrepareInsertDeleteSQLStmts;
  procedure UnprepareSQLStmts;
  procedure PostInsert;
  procedure PostUpdate(RowId: Int64);
  procedure PostDelete(RowId: Int64);
  property AvailableOperations: TSqlitePassAvailableOperations Read FAvailableOperations;
  property FieldDefs: TSqlitePassSelectStmtFieldDefs Read FFieldDefs;
//  Property PrimaryKeyField: TSqlitePassSelectStmtFieldDef Read FPrimaryKeyField;
  end;

TSqlitePassTableDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassTableDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassTableDef);
  public
  constructor Create(Owner: TSqlitePassDatabase);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTableDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddTableDefs(Source: TSqlitePassTableDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure CreateTable(Sql: String);
  procedure DeleteTable(Const TableName: String; Silent: Boolean = False);
  procedure EmptyTable(Const TableName: String);
  function TableByName(const Value: string): TSqlitePassTableDef;
  function FindTable(const Value: string): TSqlitePassTableDef;
  function FindFieldDef(TableName, FieldName: String): TSqlitePassTableFieldDef;
  function FindFieldTableName(FieldName: String; TableNames: TStringList): String;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassTableDef read GetItem write SetItem; default;
  end;

TSqlitePassSelectStmtTableDefs = class(TSqlitePassObjectList)
  private
  FSQLStmt: TSqlitePassSelectStmt;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassSelectStmtTableDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSelectStmtTableDef);
  public
  constructor Create(Owner: TSqlitePassSelectStmt);
  destructor Destroy; override;
  function FindTable(const Value: string): TSqlitePassSelectStmtTableDef;
  property Count;
  property Items[Index: Integer]: TSqlitePassSelectStmtTableDef read GetItem write SetItem; default;
  end;


  { QueryDefs }

TSqlitePassQueryDef = class
  private
  { Owner }
  FQueryDefs: TSqlitePassQueryDefs;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  FQueryName: String;
  FSql: String;
  FFieldDefs: TFieldDefs;
  FParams: Tparams;
  function GetAttached: Boolean;
  function GetQueryFullName: String;
  public
  Constructor Create(Owner: TSqlitePassQueryDefs);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassQueryDef; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  property Attached: Boolean Read GetAttached;
  procedure DeleteQuery;
  procedure RenameAs(NewName: String);
  property QueryName: String Read FQueryName;
  property QueryFullName: String Read GetQueryFullName;
  property Sql: String Read FSql;
  property Params: TParams Read FParams;
  property Fields: TFieldDefs Read FFieldDefs;
  end;

TSqlitePassQueryDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassQueryDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassQueryDef);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassQueryDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddQueryDefs(Source: TSqlitePassQueryDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure CreateQuery(Sql: String);
  procedure DeleteQuery(QueryName: String);
  function QueryByName(const Value: string): TSqlitePassQueryDef;
  function FindQuery(const Value: string): TSqlitePassQueryDef;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassQueryDef read GetItem write SetItem; default;
  end;


{ Indexes management }

TSqlitePassIndexDirection = (cidAscending, cidDescending, cidUnknown);

TSqlitePassIndexColumn = class
 public
 Position: Integer;
 CollatingSequence: String;
 ColumnName: String;
 Direction: TSqlitePassIndexDirection;
 procedure Assign(Source: TSqlitePassIndexColumn);
 end;

TSqlitePassIndexColumns = class(TSqlitePassObjectList)
  private
  FIndex: TSqlitePassIndex;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassIndexColumn;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndexColumn);
  public
  constructor Create(Owner: TSqlitePassIndex);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassIndexColumns);
  procedure AddColumns(Source: TSqlitePassIndexColumns);
  function ColumnByName(const Value: string): TSqlitePassIndexColumn;
  function FindColumn(const Value: string): TSqlitePassIndexColumn;
  property Items[Index: Integer]: TSqlitePassIndexColumn read GetItem write SetItem; default;
  property Count;
  end;

TSqlitePassIndex = class
 private
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  FTableName: String;
  FIndexName: String;
  FOverwriteIfExists: Boolean;
  FColumns : TSqlitePassIndexColumns;
  FSql: String;
  FUnique : Boolean;
  function GetIndexFullName: String;
  function GetAttached: Boolean;
  procedure SetFIndexName(value: String);
 public
  constructor Create;
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassIndex; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  property Attached: Boolean Read GetAttached;
  property Columns: TSqlitePassIndexColumns read FColumns write FColumns;
  property IndexName: String Read FIndexName Write SetFIndexName;
  property IndexFullName: String Read GetIndexFullName;
  property Sql: String Read FSql;
  property TableName: String Read FTableName Write FTableName;
  property Unique: Boolean read FUnique write FUnique;
 end;

TSqlitePassDatabaseIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatabaseIndexDefs;
  public
  constructor Create(Owner: TSqlitePassDatabaseIndexDefs);
  Destructor Destroy; override;
  procedure Reindex;
  procedure DeleteIndex;
  end;

TSqlitePassDatabaseIndexDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassIndex;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndex);
  public
  constructor Create(Owner: TSqlitePassDatabase);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassDatabaseIndexDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddIndexes(Source: TSqlitePassDatabaseIndexDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  function IndexByName(const Value: string): TSqlitePassIndex;
  function FindIndex(const Value: string): TSqlitePassIndex;
  procedure CreateIndex(Sql: String); overload;
  procedure CreateIndex(Index: TSqlitePassDatasetIndex); overload;
  procedure DeleteIndex(IndexName: String);
  procedure Reindex;
  procedure Refresh;
  procedure RenameIndex(OldIndexName, NewIndexName: String);
  property Items[Index: Integer]: TSqlitePassIndex read GetItem write SetItem; default;
  property Count;
  end;


{ Views }

TSqlitePassView = class
 private
 FAttachedDatabase: TSqlitePassDatabaseAttached;
 FViews: TSqlitePassViews;
 FSqlCreateStmt: String;
 FSqlSelectStmt: String;
 FViewName: String;
 FTemp: Boolean;
 function GetAttached: Boolean;
 function GetViewFullName: String;
 public
 Constructor Create(Owner: TSqlitePassViews);
 Destructor Destroy; override;
 procedure Assign(Source: TSqlitePassView; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
 property Attached: Boolean Read GetAttached;
 procedure DeleteView;
 property SqlCreateStmt: String Read FSqlCreateStmt;
 property SqlSelectStmt: String Read FSqlSelectStmt;
 property ViewName: String Read FViewName;
 property ViewFullName: String Read GetViewFullName;
 end;


{ TSqlitePassViews }

TSqlitePassViews = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassView;
  procedure SetItem(Index: Integer; const Value: TSqlitePassView);
  public
  constructor Create(Owner: TSqlitePassDatabase);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassViews; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddViews(Source: TSqlitePassViews; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure CreateView(Const ViewName, SQL: String; Const Temp: Boolean = False; Const Silent: Boolean = True);
  procedure DeleteView(ViewName: String);
  function ViewByName(const Value: string): TSqlitePassView;
  function FindView(const Value: string): TSqlitePassView;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassView read GetItem write SetItem; default;
  end;


{ Triggers }

TSqlitePassTrigger = class
 private
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  FTriggers: TSqlitePassTriggers;
  FSql: String;
  FTemporary: Boolean;
  FTriggerName: String;
  function GetAttached: Boolean;
  function GetTriggerFullName: String;
 public
  constructor Create(Owner: TSqlitePassTriggers);
  Destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTrigger; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure DeleteTrigger;
  property Attached: Boolean Read GetAttached;
  property Sql: String Read FSql;
  property Temporary: Boolean Read FTemporary Write FTemporary;
  property TriggerName: String Read FTriggerName Write FTriggerName;
  property TriggerFullName: String Read GetTriggerFullName;
 end;

TSqlitePassTriggers = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  Protected
  procedure ClearAndFreeItems;
  Function GetItem(Index: Integer): TSqlitePassTrigger;
  Procedure SetItem(Index: Integer; Value: TSqlitePassTrigger);
  public
  constructor Create(Owner: TSqlitePassDatabase);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassTriggers; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddTriggers(Source: TSqlitePassTriggers; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure CreateTrigger(Sql: String);
  procedure DeleteTrigger(TriggerName: String);
  function TriggerByName(const Value: string): TSqlitePassTrigger;
  function FindTrigger(const Value: string): TSqlitePassTrigger;
  property Items[Index: Integer]: TSqlitePassTrigger read GetItem write SetItem; default;
  property Count;
  procedure Refresh;
  end;


{TSqlitePassDatasets : Maintain a list of all datasets connected to the database}

TSqlitePassDatasets = Class(TSqlitePassObjectList)
  Private
    FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDataset;
    Procedure SetItem(Index: Integer; Dataset: TSqlitePassDataset);
  public
    constructor Create(Owner: TSqlitePassDatabase);
    destructor Destroy; override;
    function DatasetByName(Const Name: String): TSqlitePassDataset;
    function FindDataset(const Value: string): TSqlitePassDataset;
    property Items[Index: Integer]: TSqlitePassDataset read GetItem write SetItem; default;
    property Count;
    Procedure Close;
  end;

  TSqlitePassDatasetType = (dtTable, dtQuery, dtView, dtSQLSelect, dtSQLDirect, dtUnknown);
 { Forward declaration }
 TSqlitePassDatasetIndexDefs=class;

 TSqlitePassDatasetVersionInfo = Class(TPersistent)
     Private
       FPackage     : String;
       FComponent   : String;
       procedure SetFComponent(const Value: String);
       procedure SetFPackage(const Value: String);
     Published
      constructor Create;
      Property Component     : String  Read FComponent       Write SetFComponent;
      Property Package       : String  Read FPackage         Write SetFPackage;
     end;




{ TSqlitePassDataset }

TSqlitePassDatasetExportMethod = (VisibleFields, AllFields);

{ Events }
TSqlitePassDatasetExportProgressEvent = procedure(Current, Total:Integer) of object;
TSqlitePassDatasetFilterRecordEvent = procedure(DataSet: TDataSet; var Accept: Boolean) of object;
TSqlitePassDatasetImportProgressEvent = procedure(Current:Integer) of object;
TSqlitePassDatasetGetFieldDataEvent = procedure(Field: TField; Const NativeFieldData: String; FieldDataBuffer: Pointer; var Handled: Boolean) of object;
TSqlitePassDatasetSetFieldDataEvent = Procedure(Field: TField; Const FieldDataBuffer: Pointer; Out NativeFieldData: String; var Handled: Boolean) of object;

{ Internal Object to Keep information on sorted fields }
TSqlitePassSortedField = Class
     public
     FieldName: String;
     FieldIndex: Integer; { Field Index }
     Descendant: Boolean; { True = 'DESC' }
     end;

{The TSqlitePassDataset is a link between your application and the database content.
 TSqlitePassDataset enables you to access tables, queries or even to create direct SQL queries to read and write data from/to your database.}
TSqlitePassDataset = class(TDataSet)
private
    FRecordset              : TSqlitePassRecordset;
    FParamCheck             : Boolean;
    FParams                 : TParams;
    FDatasetName            : String;
    FDatasetFullName        : String;
    FDatasetType            : TSqlitePassDatasetType;
    FDatabase               : TSqlitePassDatabase;
    { Locate }
    FLocateKeyFields        : string;     { Used to automatically update the locate found records list when dataset is sorted...}
    FLocateKeyValues        : Variant;    { Used to automatically update the locate found records list when dataset is sorted...}
    FLocateOptions          : TLocateOptions; { Used to automatically update the locate found records list when dataset is sorted...}
    FLocateRecordset        : TSqlitePassRecordset; { Recordset used to locate specific records }
    FLocateFoundRecords     : TList; { List to store location (RecordIndex) of found records }
    FLocateCurrentItem      : Integer; { Current position in found records list }
    FLocateMoveState        : TGetResult; { status of the last move in the list }
    FLocateSmartRefresh     : Boolean; { if True, Locate found record list is automatically updated }
    { LookUp }
    FHasCalcFields          : Boolean; { Flag to indicate whether or not TSqlitePassDataset fields contains Calc or lookup fields }
    FLookUpOptions          : TLocateOptions;
    FLookUpRecordset        : TSqlitePassRecordset;
    FLookUpCache            : Boolean;
    FLookUpMaxRecordCount   : Integer;
    FLookUpResultFields     : TList;
    FLookUpSmartRefresh     : Boolean; { if True, Lookup recordset is automatically updated }
    { Filtering }
    FMasterLink             : TMasterDataLink;
    FMasterFields           : String;
    FDetailFields           : TList;
    FMasterAutoActivate     : Boolean;
    FFilterRecordLowerLimit : Cardinal;
    FFilterRecordUpperLimit : Cardinal;
    { ----- }
    FRecordsCacheCapacity   : TSqlitePassRecordsCacheCapacity; { Number of records that can fit in each memory bloc allocated to store the field data }
    FReadOnly               : Boolean;
    FProcessMessages        : Boolean;
    { SQL Stmt }
    FSQL                    : TStringList; { The original SQL statement }
    FSQLSelectStmt          : TSqlitePassSelectStmt; { Parsed SQL Statement to add filtering, sorting... capabilities }
    FSQLChanges             : TSqlitePassSQLChanges; { Set to indicate that FSQLSelectStmt needs to be updated }
    FRefetchRows            : Boolean; { Flag to indicate we have to refetch rows from database }
    { Sorting }
    FSorted                 : Boolean; { Flag to indicate if TSqlitePassDataset is sorted }
    FSortedBy               : String;  { OrderBy SQL Stmt }
    FSortCalcFields         : Boolean; { If True, Calc and lookup fields are sorted }
    FSortedFields           : TList;   { A list of fields to be sorted }
    FSortMode               : TSqlitePassSortMode; { Flag to indicate if the sort uses SQL or QuickSort method. Calc and Lookup Fields can be sorted only when smInternal is selected}
    { ----- }
    FWriteMode              : TSqlitePassWriteMode;
    { ----- }
    FInInternalOpen         : Boolean; { Flag to indicate TSqlitePassDataset is executing InternalOpen }
    FDatabaseAutoActivate   : Boolean; { Flag to indicate whether or not TSqlitePassDataset will automatically open his associated Database }
    FVersionInfo            : TSqlitePassDatasetVersionInfo;
    FIndexDefs              : TSqlitePassDatasetIndexDefs;
    FExportMethod           : TSqlitePassDatasetExportMethod;

    { Additional Events }
    FOnExportProgress       : TSqlitePassDatasetExportProgressEvent;
    FOnImportProgress       : TSqlitePassDatasetImportProgressEvent;
    FOnFilterRecord         : TSqlitePassDatasetFilterRecordEvent;

    { --- Functions and Procedures }
    Function  GetActiveRecord: Integer;
    Procedure UpdateFSQL(Sender: TObject);
    Procedure UpdateFSQLStmt;
    procedure UpdateAutoIncValues;
    Procedure SetFDatasetType;
    procedure SetFSQL(const Value: TStringList);
    Function  GetFDatabase:TSqlitePassDatabase;
    Procedure SetFDatabase(Value:TSqlitePassDatabase);
    Procedure SetFDatasetName(const Value:String);
    Procedure SetFReadOnly(const Value:Boolean);
    Function  GetFReadOnly:Boolean;
    Procedure SetFFilterRecordLowerLimit(const Value: Cardinal);
    Procedure SetFFilterRecordUpperLimit(const Value: Cardinal);
    procedure SetFSortedBy(const Value: String);
    procedure SetFSorted(const Value: Boolean);
    function  GetUniqueFieldName(FieldDef: TSqlitePassSelectStmtFieldDef): String;
    procedure SetFLookUpCache(const Value: Boolean);
    procedure SetFWriteMode(Const Value: TSqlitePassWriteMode);
    { ----- Opening ----- }
    function CheckCanOpen(VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Boolean;
    { ----- Sorting ----- }
    procedure FillLookupFields; { Enable sorting on Calc and Lookup fields }
    procedure GetSortedFields(SortedBySql: String);
    procedure ClearSortedFields;
    procedure SortRecords(SortedBySql: String);
    procedure SetFSortCalcFields(const Value: Boolean);
    procedure SetFSortMode(const Value: TSqlitePassSortMode);
    { ----- Filtering ----- }
    function  GetMasterDataSource: TDataSource;
    procedure SetMasterDataSource(Value: TDataSource);
    procedure MasterChanged(Sender: TObject);
    procedure MasterDisabled(Sender: TObject);
    function  GetMasterFields: string;
    procedure SetMasterFields(const Value: string);
    procedure ProcessMasterDetailFields;
    { ----- Locate ----- }
    function CreateLocateSQL(const KeyFields: string; const KeyValues: Variant; LocateOptions: TLocateOptions; MaxRecordCount: Integer = 0): String;
    procedure LocateMoveToRecord;
    procedure ResetLocateList;
    procedure RefreshLocateList;
    function GetLocateRecordCount: Integer;
    { ----- Lookup ----- }
    function GetLookUpResultValues: Variant;
    procedure RefreshLookupRecordset;
    procedure ActivateLookUpCache;
protected
    { ----- Params ----- }
    Procedure SetParamsList(Value: TParams);
    Procedure UpdateParamsList;
    Procedure WriteParamData(Writer: TWriter);
    Function  GetParamsCount: Integer;
    Procedure DefineProperties(Filer: TFiler); override;
    Procedure ReadParamData(Reader: TReader);
    { ----- Override methods from TDataset ----- }
    Procedure Loaded; override;
    Procedure Notification(AComponent: TComponent; Operation: TOperation);Override;
    Function AllocRecordBuffer: PChar; override;
    Function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override; //Ok v1.00
    Function GetRecordSize: Word; override;
    Function IsCursorOpen: Boolean; override;
    Procedure ClearCalcFields(Buffer: PChar); override;
    Procedure FreeRecordBuffer(var Buffer: PChar); override;
    Procedure InternalOpen; override;
    Procedure InternalClose; override;
    Procedure InternalInitFieldDefs; override;
    Procedure InternalClearFieldDefs;
    Procedure InternalInitIndexDefs;
    Procedure InternalInitRecord(Buffer: PChar); override;
    Procedure InternalRefresh; override;
    Procedure UpdateInternalFieldsInfos;
    Procedure SetFieldData(Field: TField; Buffer: Pointer);override;
    Procedure UpdateIndexDefs; override;
    Function  GetCanModify: Boolean; override;

    { Fetch data from physical Database }
    Procedure RefreshActive;

    { ----- Navigation and Editing ----- }
    Procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    Procedure InternalCancel; override;
    Procedure InternalDelete; override;
    Procedure InternalEdit; override;
    Procedure InternalFirst;override;
    procedure InternalInsert; override;
    Procedure InternalLast;override;
    Procedure InternalPost; override;
    Procedure InternalSetToRecord(Buffer: PChar); override;
    { ----- Bookmarks ----- }
    Procedure InternalGotoBookmark(ABookmark: Pointer); override;
    // Procedure InternalMoveToBookmark(Bookmark: Pointer);
    Function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override; {From TDataset}
    Procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override; {From TDataset}
    // Keep the inherited - Function  GetBookmarkStr: TBookmarkStr; override;
    // Keep the inherited - Procedure SetBookmarkStr(const Value: TBookmarkStr); override;
    Procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    Procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    { ----- Filtering ----- }
    Procedure CheckFiltered(ApplyFilter: Boolean = False);
    function  FindRecord(Restart, GoForward: Boolean): Boolean; override;
    procedure DoOnNewRecord; override;
    procedure SetFiltered(Value: Boolean); override;
    procedure SetFilterText(Const Value: String); override;
    { ----- Miscalleous ----- }
    Procedure InternalHandleException; override;
    { ----- Records infos ----- }
    Function GetRecordCount: Integer; override;
    Function GetRecNo: Integer; override;
    Procedure SetRecNo(Value: Integer); override;
    { ----- Blobs ----- }
    procedure AllocateBLOBPointers(Buffer: PChar);
    procedure FreeBlobPointers(Buffer: PChar);
    procedure FreeRecordPointers(Buffer: PChar);
  public
    { ----- Constructor - Destructor ----- }
    Constructor Create(AOwner: TComponent); override;
    Destructor  Destroy; override;
    { ----- Bookmarks ----- }
    Function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    Function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
    { -----  ----- }
    Function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    { ----- Apply or Cancel records modification if WriteMode is wmPostpone ----- }
    Procedure ApplyChanges;
    Procedure CancelChanges;
    Procedure EmptyTable;
    { ----- Locate ----- }
    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
    function LocateFirst: Boolean;
    function LocateNext:  Boolean;
    function LocatePrior: Boolean;
    function LocateLast:  Boolean;
    { ----- Lookup ----- }
    function Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant; override;
    function LookupFirst: Variant;
    function LookupNext:  Variant;
    function LookupPrior: Variant;
    function LookUpLast:  Variant;
    { ----- Blobs ----- }
    Function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    procedure CloseBlob(Field: TField); override;
    { -----  ----- }
    Procedure ApplyFilter;
    Procedure Sort;
    { -----  Properties ----- }

    Property ParamCount: Integer read GetParamsCount;

    {This property is only read. It gives you information about the currently selected dataset and can be one of the following values : 

     @bold(dtUnkown) :
     The dataset type could not be recognized or the DatasetName property is empty.

     @bold(dtTable) :
     The dataset is a table.

     @bold(dtQuery) :
     The dataset is a query.

     @bold(dtView) :
     The dataset is a view.

     @bold(dtSqlDirect) :
     The SQL property has been modified or you entered a new SQL query. When the SQL text is changed, the DatasetName will automatically be set to '' assuming that the DatasetName and SQL text don't match anymore.}
    Property DatasetType: TSqlitePassDatasetType Read FDatasetType;
    Property LocateMoveState: TGetResult read FLocateMoveState;
    Property LocateRecordCount: Integer Read GetLocateRecordCount;
    Property SortMode: TSqlitePassSortMode Read FSortMode Write SetFSortMode;
    Property SQLSelectStmt: TSqlitePassSelectStmt Read FSQLSelectStmt;
  published

    {Selects the TsqlitePassDatabase component you want to depend on.}
    Property Database: TSqlitePassDatabase Read GetFDatabase Write SetFDatabase;

    Property ExportMethod: TSqlitePassDatasetExportMethod Read FExportMethod Write FExportMethod;

    {Classic table MasterFields behavior.

     At design time, a dialog will let you create or modify the relation between MasterFields and DetailFields.

     A relation is defined like this : MasterFieldName=DetailFieldName
     If you want set several relations, they must be separated by a ';'
     MasterFieldName1=DetailFieldName1;MasterFieldName2=DetailFieldName2}
    property MasterFields: string read GetMasterFields write SetMasterFields;

    {Classic table MasterSource behavior.}
    property MasterSource: TDataSource read GetMasterDataSource write SetMasterDataSource;
    property MasterSourceAutoActivate: Boolean read FMasterAutoActivate write FMasterAutoActivate;

    {Once you are connected to a database, enters a table name or a
     query name. At design time, a dialog will let you choose your dataset
     among all the available database datasets.}
    Property DatasetName: String Read FDatasetName Write SetFDatasetName;

    {FilterRecordLowerLimit is the lower limit of the range filter.
     If greater than 1, the -nth first records will not be retrieved.
     In other words, if FilterLowerLimit = 4, the fifth record will be the first one retrieved from the query.}
    Property FilterRecordLowerLimit: Cardinal Read FFilterRecordLowerLimit Write SetFFilterRecordLowerLimit;

    {FilterRecordUpperLimit is the upper limit of the range filter.

    If greater than 1, the -nth first records will be retrieved.
    In other words, if FilterUpperLimit = 4 and FilterLowerLimit = 0 then
    only the four first records will be retrieved from the query.

    If lesser than 0, the -nth last records will be retrieved.
    In other words, if FilterUpperLimit = -9 and FilterLowerLimit = 0 then
    only the nine last records will be retrieved from the query.}
    Property FilterRecordUpperLimit: Cardinal Read FFilterRecordUpperLimit Write setFFilterRecordUpperLimit;
    Property LocateSmartRefresh: Boolean Read FLocateSmartRefresh Write FLocateSmartRefresh;
    Property LookUpCache: Boolean Read FLookUpCache Write SetFLookUpCache;
    Property LookUpMaxRecordCount: Integer Read FLookUpMaxRecordCount Write FLookUpMaxRecordCount;
    Property LookUpOptions: TLocateOptions Read FLookUpOptions Write FLookUpOptions;
    Property LookUpSmartRefresh: Boolean Read FLookUpSmartRefresh Write FLookUpSmartRefresh;

    {SQL represents the SQL statement used to retrieve data from the database.

    For tables, it will automatically be set to :

    @code(SELECT * FROM TableName;)
    if all fields need to be retrieved from the table, or to :

    @code(SELECT field1, field2... FROM TableName;)
    if only some fields need to be retrieved from the table.

    For queries, it will reflect the query SQL statement.
    You can also directly write your own SQL statement to fit your needs or to interact directly with the database.
    In this case, the datasetname propery will be set to '' (empty) and the datasetType will be set to dtDirectSql.}
    Property SQL: TStringList Read FSQL Write SetFSQL;
    Property SortCaclFields: Boolean Read FSortCalcFields Write SetFSortCalcFields;

    {Determines whether or not the sortedBy property is activated.} 
    Property Sorted: Boolean Read FSorted Write SetFSorted;

    {The SortedBy property takes a SQL ORDERBY clause but without the
     'ORDER BY' expression at the begining.
     At design time, a dialog will let you create or modify the sort order.

     @italic(Example : 'car_names ASC, car_types DESC'.)}
    Property SortedBy: String Read FSortedBy Write SetFSortedBy;
    Property RecordsCacheCapacity: TSqlitePassRecordsCacheCapacity Read FRecordsCacheCapacity Write FRecordsCacheCapacity;
    Property Params: TParams read FParams Write SetParamsList Stored False;

    {The IndexDefs property gives you access to the indexes definitions for the selected table.
     Indexes are only available if the DatasetType is a 'dtTable' type.}
    Property IndexDefs: TSqlitePassDatasetIndexDefs  Read FIndexDefs Write FIndexDefs;
    Property ReadOnly: Boolean Read GetFReadOnly Write SetFReadOnly default False;

    {When set to True, opens automatically the database if needed.}
    Property DatabaseAutoActivate: Boolean Read FDatabaseAutoActivate  Write FDatabaseAutoActivate;

    Property VersionInfo: TSqlitePassDatasetVersionInfo Read FVersionInfo Write FVersionInfo;
    Property ParamCheck: Boolean Read FParamCheck Write FParamCheck;
    Property ProcessMessages: Boolean Read FProcessMessages Write FProcessMessages;
    Property WriteMode: TSqlitePassWriteMode Read FWriteMode Write SetFWriteMode;

    { Additionnal Events }
    Property OnExportProgress: TSqlitePassDatasetExportProgressEvent Read FOnExportProgress Write FOnExportProgress;
    Property OnImportProgress: TSqlitePassDatasetImportProgressEvent Read FOnImportProgress Write FOnImportProgress;
    Property OnFilterRecord  : TSqlitePassDatasetFilterRecordEvent read FOnFilterRecord write FOnFilterRecord;

    { --- }
    Property BeforeOpen;
    Property AfterOpen;
    Property BeforeClose;
    Property AfterClose;
    Property BeforeInsert;
    Property AfterInsert;
    Property BeforeEdit;
    Property AfterEdit;
    Property BeforePost;
    Property AfterPost;
    Property BeforeCancel;
    Property AfterCancel;
    Property BeforeDelete;
    Property AfterDelete;
    Property BeforeScroll;
    Property AfterScroll;
    Property OnCalcFields;
    Property OnDeleteError;
    Property OnEditError;
    Property OnNewRecord;
    Property OnPostError;
    { --- }
    Property AutoCalcFields;

    {Classic dataset behavior.
     When set to True, opens the dataset and displays data if dataware components are linked to the datasource.
     When set to False, closes the dataset and frees the memory used to store dataset records.}
    Property Active;

    {Classic dataset filter behavior. The filter property takes a SQL WHERE clause but without the WHERE word at the begining. You can also use wildcard characters as discribed in the Sqlite help.

    @italic(Example1 : country = 'France')
    @italic(Example1 : customer like '%cur%')}
    Property Filter;

    {Determines whether or not the different filters are activated.
     The TsqlitePassDataset component can handle three filter levels
     that will be applied in this priority order :

     1 : MasterFields/DetailFields property
     2 : Filter property
     3 : RecordLowerLimit/RecordUpperLimit properties}
    Property Filtered;
    Property FilterOptions;
  End;

{ ----- Blob fields stream ----- }

TSqlitePassBlobStream = class(TMemoryStream)
  private
    FActiveRecIndex : Integer;
    FField          : TBlobField;
    FDataSet        : TSqlitePassDataset;
    FOpened         : Boolean;
    FModified       : Boolean;
    FRecBlobStream  : TMemoryStream;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    Function Read(var Buffer; Count: Longint): Longint; override;
    Function Write(const Buffer; Count: Longint): Longint; override;
  End;

 TSqlitePassDatasetIndexDefs = class(TSqlitePassObjectList)
  private
  FDataset: TSqlitePassDataset;
  Protected
  function GetItem(Index: Integer): TSqlitePassIndex;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndex);
  public
  constructor Create(Owner: TSqlitePassDataset);
  destructor Destroy; override;
  procedure RefreshFromCache;
  property Items[Index: Integer]: TSqlitePassIndex read GetItem write SetItem; default;
  property Count;
  end;

 TSqlitePassDatasetIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatasetIndexDefs;
  public
  constructor Create(Owner: TSqlitePassDatasetIndexDefs);
  destructor Destroy; override;
  procedure Reindex;
  procedure DeleteIndex;
  end;

 TSqlitePassTranslator = class
  Private
    { Objects Reference }
    FDatabase: TSqlitePassDatabase;
    FFieldTypesTranslationRules: TSqlitePassFieldTypesTranslationRules;

     { Conversion functions from a SQLite Db Value to it's internal MemRecord storage }
    procedure IntegerToWordBool(Const IntegerValue: Integer; out Value: WordBool);
    procedure BooleanTextToWordBool(Const StrValue: String; out Value: WordBool);
    procedure DateTextToInteger(Const StrValue: String;
                               Const YearStart, YearLength, MonthStart, MonthLength, DayStart, DayLength: Word;
                               out Value: Integer); overload;
    procedure DateTextToInteger(Const StrValue: String; out Value: Integer); overload;
    procedure TimeTextToInteger(Const StrValue: String;
                               Const HourStart, HourLength, MinStart, MinLength, SecStart, SecLength, MSecStart, MSecLength: Word;
                               out Value: Integer); overload;
    procedure TimeTextToInteger(Const StrValue: String; out Value: Integer); overload;
    procedure JulianDateTimeToDouble(Const DoubleValue: Double; out Value: Double);
    procedure DateTimeTextToDouble(Const StrValue: String; out Value: Double);
    procedure UnixDateTimeToDouble(Const Int64Value: Int64; out Value: Double);
    procedure MacDateTimeToDouble(Const Int64Value: Int64; out Value: Double);

     { Conversion functions from an internal MemRecord storage to a SQLite Db Value }
    procedure WordBoolToInteger(Const Value: WordBool; out IntegerValue: Integer);
    procedure WordBoolToBooleanText(Const Value: WordBool; out StrValue: String);
    procedure IntegerToDateText(Const Value: Integer;
                               Const DatePattern: String;
                               Const YearStart, YearLength, MonthStart, MonthLength, DayStart, DayLength: Word;
                               out StrValue: String); overload;
    procedure IntegerToDateText(Const Value: Integer; out StrValue: String); overload;
    procedure IntegerToTimeText(Const Value: Integer;
                               Const TimePattern: String;
                               Const HourStart, HourLength, MinStart, MinLength, SecStart, SecLength, MSecStart, MSecLength: Word;
                               out StrValue: String); overload;
    procedure IntegerToTimeText(Const Value: Integer; out StrValue: String); overload;
    procedure DoubleToJulianDateTime(Const Value: Double; out DoubleValue: Double);
    procedure DoubleToDateTimeText(Const Value: Double; out StrValue: String);
    procedure DoubleToUnixDateTime(Const Value: Double; out Int64Value: Int64);
    procedure DoubleToMacDateTime(Const Value: Double; out Int64Value: Int64);

    { Informations on databases parts }
    procedure GetDatabaseTableDefs(TableDefs: TSqlitePassTableDefs); virtual;
    Procedure GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef); virtual;
    procedure GetDatabaseIndexDefs(IndexDefs: TSqlitePassDatabaseIndexDefs); virtual;
    procedure GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs); virtual;
    procedure GetDatabaseViews(Views: TSqlitePassViews); virtual;
    procedure GetDatabaseTriggers(Triggers: TSqlitePassTriggers); virtual;
    function IsSystemTable(TableName: String): Boolean; virtual;

   { Field Datatype Conversion }
    procedure ConvertNativeFieldTypeToPascalFieldType(FieldDef: TSqlitePassTableFieldDef);
    procedure GetFieldDefSize(FieldDef: TSqlitePassTableFieldDef);

    { Properties }
    property FieldTypesTranslationRules: TSqlitePassFieldTypesTranslationRules Read FFieldTypesTranslationRules;

  public
      { --- Constructor / Destructor --- }
    Constructor Create(Owner: TSqlitePassDatabase);
    Destructor Destroy;override;

  end;


  { Specific translator for Kexi databases }

  TSqlitePassTranslator_Kexi = class(TSqlitePassTranslator)
  public
   { Field Datatype Conversion }
    procedure ConvertNativeFieldTypeToPascalFieldType(FieldDef: TSqlitePassTableFieldDef);
    function GetNativeFieldTypeName(FieldType: Word): String;

    { Gets information on database parts }
    Procedure GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef); override;
    procedure GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs); override;

    { Utilities }
    function IsSystemTable(TableName: String): Boolean; override;
  end;

  { Empty declaration : nothing to override }
  TSqlitePassTranslator_SqliteExpert = class(TSqlitePassTranslator)
  end;

  { Empty declaration : nothing to override }
  TSqlitePassTranslator_SqliteAdmin = class(TSqlitePassTranslator)
  end;

StmtType = (stCreate, stCreateTable, stSelect, stInsert, stUpdate, stDelete, stUnknown);

TSqlitePassSQLTokenType=(ttKeyword, ttIdentifier, ttDatabaseIdentifier, ttTableIdentifier, ttFieldIdentifier, ttFunctionIdentifier,
                         ttStar, ttNumber, ttQuote, ttWhitespace, ttComment, ttColon, ttSemiColon, ttEqual, ttNotEqual, ttGreater, ttGreaterOrEqual,
                         ttLesser, ttLesserOrEqual, ttEOF, ttUnknown);
//tkFloat, tkInteger, tkHexDecimal, tkNumber, tkSymbol
TSqlitePassSQLTokenizerOptions =  set of (toClean, toUpperCase);


TSqlitePassSQLToken = class
 public
 { the SQL statement }
 Text: String;
 { the token type }
 TokenType: TSqlitePassSQLTokenType;
 { the matching token index
   for example, an '(' token should have a ')' token.}
 MatchingPair: Integer;
 { NestingLevel }
 NestingLevel: Integer;
 Constructor Create;
 end;

TSqlitePassSQLTokens = class(TSqlitePassObjectList)
  private
  function GetItem(Index: Integer): TSqlitePassSQLToken;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSQLToken);
  procedure ClearAndFreeItems;
  public
  destructor Destroy; override;
  property Count;
  property Items[Index: Integer]: TSqlitePassSQLToken read GetItem write SetItem; default;
  end;

{The TSqlitePassSQLSections is an utility object used to :
 Split a SQL statement into a list of string for each SQL Keyword }
TSqlitePassSQLSections = class(TStringList)
  public
  function  GetSectionText(SectionName: String): String;
  procedure SetSectionText(SectionName, Text: String);
  procedure SetSectionTextFromTokenizer(SectionName: String; Tokenizer: TSqlitePassSQLTokenizer);
  procedure ReplaceSectionText(SectionName: String; NewStmt: String);
  { Replace a given SQLSection Text by the one in the SQLSections Parameter
    if no SectionName is set, all the Strings are copied }
  procedure ReplaceSectionTextFromSQLSections(SQLSections: TSqlitePassSQLSections; SectionName: String);
  procedure ClearSectionText(SectionName: String; NewStmt: String);
  procedure SplitSQLStmtIntoSections(SQLStmt: String; Keywords: Array of String);
  function  GetSQLStmtFromSections: String;
  end;

{The TSqlitePassTokenizer is an utility object used to :
 Split a SQL statement or a piece of SQL statement
 into a list of tokens (TSqlitePassSQLToken)

 Move to the First, Prior, Next, Last Token
 Locate a specific token by text or type
 Replace a specific token located before an existing token
 Remove a specific token located after an existing token
 Insert a new token before an existing token
 Insert a new token after an existing token
 Set the string to be tokenized and rebuild a string representation from the tokens (Text property)...}
TSqlitePassSQLTokenizer = class
 private
 FText: String;
 { The current Token index }
 FCurrentTokenIndex: Integer;
 { A list of the tokens }
 FTokens: TSqlitePassSQLTokens;
 function GetFCount: Integer;
 function GetFCurrentToken: TSqlitePassSQLToken;
 function GetFEOF: Boolean;
 function GetFText: String;
 procedure SetFText(const Value: String);
 procedure Clean;
 public
 constructor Create;
 destructor Destroy; override;
 function First: Boolean;
 function Prior: Boolean; overload;
 function Prior(TokenType: TSqlitePassSQLTokenType): Boolean; overload;
 function Next: Boolean; overload;
 function Next(TokenType: TSqlitePassSQLTokenType): Boolean; overload;
 function Last: Boolean;
 procedure Replace(OldText, NewText: String);
 function Locate(TokenText: String): Boolean; overload;
 function Locate(TokenType: TSqlitePassSQLTokenType): Boolean; overload;
 procedure InsertBefore(TokenText, InsertedStmt: String);
 procedure InsertAfter(TokenText, InsertedStmt: String);
 procedure RemoveBefore(TokenText, RemovedStmt: String);
 procedure RemoveAfter(TokenText, RemovedStmt: String);
 procedure Remove(TokenText: String);
 procedure Tokenize(Options: TSqlitePassSQLTokenizerOptions = []);
 function GetTextAfter(TokenType: TSqlitePassSQLTokenType): String;
 { Sets the original string to be tokenized or builds and returns a string from the tokens }
 property Text: String Read GetFText Write SetFText;
 { Gives access to the currently selected token }
 property Token: TSqlitePassSQLToken Read GetFCurrentToken;
 property EOF: Boolean Read GetFEOF;
 property Count: Integer Read GetFCount;
 end;


TSqlitePassSQLStmtDef = class
 private
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  FSQLStmtDefs: TSqlitePassSQLStmtDefs;
  { The SQL statement that can be modified, using the section spliter or the tokenizer }
  FSQLText: String;
  { the Statement name }
  FSQLStmtName: String;
  function GetAttached: Boolean;
  function GetSQLStmtFullName: String;
  { Returns the statement type }
  function  GetStmtType: StmtType;
 protected
 public
  constructor Create(SQLStmts: TSqlitePassSqlStmtDefs);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassSQLStmtDef; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure DeleteSQLStmt;
  { Properties }
  property Attached: Boolean Read GetAttached;
  property SQL: String Read FSQLText Write FSQLText;
  property SQLStmtName: String Read FSQLStmtName Write FSQLStmtName;
  property SQLStmtFullName: String Read GetSQLStmtFullName;
  property SQLType: StmtType Read GetStmtType;  
 end;

 { SQL statements collection }
 TSqlitePassSQLStmtDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassSQLStmtDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSQLStmtDef);
  procedure UpdateSQLStmtTable(Const UpdateType: Integer; Const SQLStmtName: String; Const NewSQLStmtName: String; Const SQLStmt: TStringList);
  public
  constructor Create(Owner: TSqlitePassDatabase);
  destructor Destroy; override;
  procedure Assign(Source: TSqlitePassSQLStmtDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure AddSQLStmts(Source: TSqlitePassSQLStmtDefs; AttachedDatabase: TSqlitePassDatabaseAttached = nil);
  procedure CreateSQLStmt(Const SQLStmtName: String; Const SQLStmt: TStringList);
  procedure DeleteSQLStmt(Const SQLStmtName: String);
  function FindSQLStmt(const Value: string): TSqlitePassSQLStmtDef;
  procedure RenameSQLStmt(Const OldName, NewName: String);
  function SQLStmtByName(const Value: string): TSqlitePassSQLStmtDef;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassSQLStmtDef read GetItem write SetItem; default;
  end;

TSqlitePassSQLStmt = class
 private
  FDatabase: TSqlitePassDatabase;
  { The original SQL statement }
  FOriginalSQLText: String;
  { A TStringList used to split the original SQL Statement into Sections }
  FOriginalSQLSections: TSqlitePassSQLSections;
  { The SQL statement that can be modified, using the section spliter or the tokenizer }
  FSQLText: String;
  { A TStringList used to split the SQL Statement into Sections }
  FSQLSections: TSqlitePassSQLSections;
  { An Array of the SQL Keywords allowed in the SQL Statement
    used to split the statement in sections }
  FStmtKeywords: array of string;
  { A Tokenizer to split the statement in tokens }
  FTokenizer: TSqlitePassSQLTokenizer;
  { The operations that can be performed by the SQL Statement }
  FAvailableOperations: TSqlitePassAvailableOperations;
  { Statement Handle returned by sqlite engine }
  FStmtHandle: Pointer;
  { the Statement name }
  FStmtName: String;
  { Returns the statement type }
  function  GetStmtType: StmtType;
  function  GetFSQLText: String;
  procedure SetFSQLText(Value: String);
 protected
 public
  constructor Create(Database: TSqlitePassDatabase); overload;
  destructor Destroy; override;
  { }
  procedure Prepare; virtual;
  procedure UnPrepare;

  procedure QuoteString(Var S: String);
  procedure UnquoteString(Var S: String; All: Boolean = True);
  procedure RestoreOriginalStmt;
  { Properties }
  property AvailableOperations: TSqlitePassAvailableOperations Read FAvailableOperations;
  property StmtHandle: Pointer Read FStmtHandle;
  property StmtName: String Read FStmtName Write FStmtName;
  property SQLSections: TSqlitePassSQLSections Read FSQLSections Write FSQLSections;
  property SQL: String Read GetFSQLText Write SetFSQLText;
  property OriginalSQL: String Read FOriginalSQLText;
  property OriginalSQLSections: TSqlitePassSQLSections Read FOriginalSQLSections;
  property SQLType: StmtType Read GetStmtType;
  property Tokenizer: TSqlitePassSQLTokenizer Read FTokenizer Write FTokenizer;
 end;

 TSqlitePassSelectStmt = class(TSqlitePassSQLStmt)
 private
  { The dataset owner of the SQL Select statement }
  FDataset: TSqlitePassDataset;
  { An internal string used to add primary keys to the original SQL Select Statement }
  FPrimaryKeyStmt: String;
  { Number of updatable Tables }
  FPrimaryKeyCount: Integer;
  { A TableDefs collection representing the tables included in the SQL Select Statement }
  FTableDefs: TSqlitePassSelectStmtTableDefs;
  { A FieldDefs collection representing the fields included in the SQL Select Statement }
  FFieldDefs: TSqlitePassSelectStmtFieldDefs;
  { A FieldDefs 'sub'-collection representing the 'AutoInc' fields included in the SQL Select Statement
   in order to speed up thing in TSqlitePassDataset.UpdateAutoIncValues }
  FAutoIncFieldDefs: TSqlitePassSelectStmtFieldDefs;
  { Internal Flag to indicate wether or not a 'same' field is requested several time in query }
  FHasDuplicatedFields: Boolean;
 protected
  procedure AddPrimaryKeys;
  function FillTableDefs(Const TableName, FieldName: String; RecFieldIndex: Integer): Boolean;
  procedure FillTableDefsFromInternalCalculatedField(RecFieldIndex: Integer);
  procedure PrepareSQLStmts;
  procedure UnprepareSQLStmts;
 public
  constructor Create(Dataset: TSqlitePassDataset);
  destructor Destroy; override;
  { Buid an internal representation of the SQL Statement }
  procedure BuildSchema;
  { Prepare the statement and send it to the engine }
  procedure Prepare; override;
  { Release the prepared statement from the engine }
  procedure UnPrepare;
  procedure Post(PostType: TSqlitePassPostType);
  {}
  property FieldDefs: TSqlitePassSelectStmtFieldDefs Read FFieldDefs;
  property TableDefs: TSqlitePassSelectStmtTableDefs Read FTableDefs;
 end;


{ Internal cache holding the records returned by a table/query
  It is designed to provide data to TSqlitePassDataset }
TSqlitePassRecordset = Class
 Private
  FDatabase: TSqlitePassDatabase;
  FDataset: TSqlitePassDataset;

  { an array of pointers to the memory blocs used to store records values
  One record = [  BookmarkFlag
                + BookMarkData
                + RecordState
                + [Rowid_1..Rowid_n]
                + [FieldsValues_1..FieldsValues_n]
                + [Calc or Lookup FieldsValues_1..Calc or Lookup FieldsValues_n]
                + [NullFlagField_1..NullFlagField_n]
                + [ModifiedFlagField_1..ModifiedFlagField_n] ]}
  FRecordsMemoryBlocs: Array of PChar;
  FRecordsMemoryBlocsCount: Integer;

  { an array of pointers on records }
  FRecords: Array of PChar;
  FRecordsCapacity: Integer;
  FRecordsCount: Integer;
  FRecordsUsedCapacity:Integer;

  { an array of strings : Strings are not stored directly in the records, only PChars to strings }
  FStrings: Array of String;
  FStringsCount: Integer;
  FStringsCapacity: Integer;
  FStringsCapacityLimit: Integer;
  FRecycledStrings: TList;

  { a TList of TMemoryStream (necessary to free blobs) }
  FMemoryStreams: TList;
  FRecycledMemoryStreams: TList;

  { --- Record Offsets --- }
  { Offset between record start and BookMarkData Value }
  FBookMarkDataOffset: Word;
  { Offset between record start and RecordState Value }
  FRecordStateOffset: Word;
  { Offset between record start and RowIds Values
   (one rowid per table returned by the sql statement)}
  FRowIdsOffset: Word;
  { Offset between record start and Fields Values }
  FFieldsValuesOffset: Word;
  { Offset between record start and Null Fields Boolean Flags }
  FFieldsNullValuesOffset: Word;
  { Offset between record start and Modified Fields Boolean Flags }
  FFieldsModifiedValuesOffset: Word;

  { Number of Data Fields (Number of columns - 1 to substract the RowId column)
    returned by the query }
  FDataFieldsCount: Cardinal;

  { Number of Calc or Lookup Fields }
  FCalcFieldsCount: Cardinal;

  { Number of Fields (Number of columns - 1 to substract the RowId column)
    returned by the query, including Calc and Lookup fields stored in record buffer }
  FFieldsCount: Cardinal;

  { Number of tables concerned by the sql statement
    The FRowidTablesCount is used to store the record rowid for each table}
  FRowidTablesCount: Word;


  FInternalFieldsOffset: Array of Integer;

  { Size necessary to holds the fields values, or PChars for strings..blobs,
    including Calculated - Lookup fields for one record }
  FRecordSize: Integer;

  { a memory bloc to initialize values and states in a new record }
  FBlankRecordInitialization: PChar;

  { a memory bloc to initialize Calc and Lookup values as Null in a record }
  FCalcFieldsNullValuesInitBuffer: PChar;

  { a memory bloc to initialize or reset every field value in a record as UnModified }
  FUnModifiedRecordInitialization: PChar;

  FActive: Boolean;
  FCurrentRecordOldValues: PChar;
  FCurrentRecordIndex: Integer;
  MoveState: TGetResult;
  FOldCurrentRecordIndex: Integer;

  { when TsqlitePassDataset WriteMode is 'wmPostpone' then
    Records modification information is stored in two ways :
    1 - Inside the record itself in order to know the record state and verify if
        another modification is allowed (a record marked as deleted
        can't be modified anymore for example)
    2 - In a separate list (FModifiedRecords) to quickly apply updates/delete/insert operation to physical database}
  FPostPoned: Boolean;
  FModifiedRecords: Array of PChar;


  { --- Procedures / Functions --- }

//    Procedure Pack should remove all delete record (state := rsDeleted), rsModified, RsInserted, rsNoChange

  { --- Records --- }
  Function SetFRecordSize: Integer;
  function GetFieldSize(DataType: TFieldType): Integer;
  procedure GetRecords(Sql: String);
  Function GetFRecordsCount: Integer;
  Procedure ExchangeRecords(RecordIndex1, RecordIndex2: Integer);
  Procedure CopyRecord(Source, Dest: PChar);
  procedure InitInternalFieldsInfos(Fields: TFields);

  { --- Read and Write Data in Record Buffer --- }

  { Read and Write a record BookMarkFlag }
  Function GetBookMarkFlag(RecordIndex: Integer): TBookMarkFlag;
  Procedure SetBookMarkFlag(RecordIndex: Integer; Value: TBookMarkFlag);

  { Read and Write a record BookMarkData }
  Function GetBookMarkData(RecordIndex: Integer): Cardinal;
  Procedure SetBookMarkData(RecordIndex: Integer; Value: Cardinal);

  { Read and Write a record state }
  Function GetRecordState(RecordIndex: Integer):  TSqlitePassRecordState;
  Procedure SetRecordState(RecordIndex: Integer; Value: TSqlitePassRecordState);

  { Read and Write a record rowid }
  Function GetRowId(RecordIndex: Integer; TableNo: Word): Int64;
  Procedure SetRowId(RecordIndex: Integer; TableNo: Word; Value: Int64);

  { Returns a pointer (PChar) to the field value inside a record }
  Function GetFieldBuffer(RecordIndex, FieldIndex: Integer): PChar;

  { Returns the string value for FieldIndex }
  Function GetFieldAsString(RecordIndex, FieldIndex: Integer): String;
  Procedure SetFieldAsString(RecordIndex, FieldIndex: Integer; Value: String);

  { Returns the Blob - MemoryStream value for FieldIndex }
  Function GetFieldAsBlob(RecordIndex, FieldIndex: Integer): TMemoryStream;

  { Returns Variant for FieldIndex }
  Function GetFieldAsVariant(RecordIndex, FieldIndex: Integer; Field: TField): Variant;

  { Read and Write a record 'Null' Boolean Flag }
  Function FieldIsNull(RecordIndex, FieldIndex: Integer): Boolean;
  Procedure SetFieldNullValue(RecordIndex, FieldIndex: Integer; Value: ByteBool);

  { Read and Write a record 'Modified' Boolean Flag }
  Function FieldWasModified(RecordIndex, FieldIndex: Integer): Boolean;
  Procedure SetFieldWasModified(RecordIndex, FieldIndex: Integer; Value: ByteBool);

  { Used when no dataset is assigned }
  Function GetTextItem(RecordIndex, FieldIndex: Integer): String;
  Procedure SetTextItem(RecordIndex, FieldIndex: Integer; Value: String);
  Procedure AddTextItem(RecordIndex, FieldIndex: Integer; Value: String);

  { Sorting }
  Procedure SortRecords(SortedFields: TList);
  function CompareFieldValues(Const FieldIndex: Integer; RecordIndex1: Integer; RecordBuffer: PChar; Descending: Boolean = False): Integer;

  { --- Managing Records --- }
  function AddRecord: Integer;
  function AddTempRecord: Integer;
  procedure ClearRecord(Index: Integer);
  Procedure InsertRecord(Index: Integer);
  Procedure DeleteRecord(Index: Integer);
  procedure PostRecord;
  procedure Cancel;

  function AddMemoryStream(Buffer: PChar): Integer; overload;
  function AddMemoryStream(RecordIndex, FieldIndex: Integer): Integer; overload;
  procedure ClearMemoryStream(RecordIndex, FieldIndex: Integer);

  function AddString(Buffer: PChar; StrValue: String): Integer; overload;
  function AddString(RecordIndex, FieldIndex: Integer; StrValue: String): Integer; overload;
  procedure ClearString(RecordIndex, FieldIndex: Integer);

  { --- Current record old values storage / restore when editing --- }
  Procedure SaveCurrentRecordOldValues;
  Procedure RestoreCurrentRecordOldValues;

  { --- Opening / Closing --- }
  procedure Open(Sql: String; TablesCount: Word = 0);
  function SqliteValueToBuffer(Const PSqliteData: Pointer; Const ColumnIndex: Integer;
                               Const BufferFieldPos: Integer; Const DataType: TFieldType; Buffer: PChar): Integer;
  Procedure Close;
  { Reset Memory }
  procedure Reset;

  { --- Navigating --- }
  Procedure First;
  Procedure Prior;
  Procedure Next;
  Procedure Last;
  Procedure MoveBy(Offset: Integer);

  { --- Calculated and Lookup fields utilities --- }
  Procedure ClearCalcFields(RecordIndex: Integer);

  { --- Clear Postponed Records Changes --- }
  Procedure ClearModifiedRecords;

  { --- Properties --- }
  Property Active: Boolean Read FActive;
  Property RecordsCount: Integer read GetFRecordsCount;

  { CurrentRecordIndex is zero based }
  Property CurrentRecordIndex: Integer Read FCurrentRecordIndex Write FCurrentRecordIndex;
  //Property Records[RecordIndex: Integer]: PChar read GetRecord; default; *)
  Property ItemsTextValue[RecordIndex, FieldIndex: Integer]: String read GetTextItem write SetTextItem;

  Public
  { --- Constructor / Destructor --- }
  constructor Create(Dataset: TSqlitePassDataset); overload;
  constructor Create(Database: TSqlitePassDatabase); overload;
  Destructor Destroy; override;


  end;


 TSqlitePassTransaction = class
 Private
  FLevel: Integer;
  FEngine: TSqlitePassEngine;
  FInternalTransaction: Boolean;
 Public
  Constructor Create(Owner: TSqlitePassEngine);
  Destructor Destroy; override;
  Procedure CommitInternalTransaction;
  Procedure RollbackInternalTransaction;
  Procedure StartInternalTransaction;
  Procedure ResetTransactions(CommitTransaction: Boolean);

  {Ends the transaction and write data to the database.}
  Procedure Commit;

  Procedure CommitAll;

  {Ends the transaction and discards any change made to the database.}
  Procedure Rollback;

  Procedure RollbackAll;

  {Starts a new transaction unless one is already started.}
  Procedure Start;
  Property  Level: Integer Read FLevel Write FLevel;
 end;

 TSqlitePassEngine = Class
  Private
    FDatabase: TSqlitePassDatabase;
    FPreparedStmts: TStringList;
    FConnectionHandle: Pointer;
    FLibraryFile: String;
    FLibraryHandle: Integer;
    FLibraryLoaded: Boolean;
    FStartTime:Integer;
    FStopTime: Integer;
    FElapsedTime : Double;
    FQueryTimeOut: Integer;
    FTransaction: TSqlitePassTransaction;
  Public
    Constructor Create(Database: TSqlitePassDatabase);
    Destructor Destroy; override;
    Property QueryTimeout: Integer Read FQueryTimeout Write FQueryTimeout;
    Property Connection: Pointer read FConnectionHandle;
    { Chrono }
    Procedure StartChrono;
    Procedure StopChrono;
    Property ElapsedTime: Double read FElapsedTime;
    { ------------ Library ----------------- }
    Function LoadSqliteLibrary(LibraryFile: String): Boolean;
    { ------------- Database ---------------- }
    procedure OpenDatabase(FullName: String; LibraryFile: String = DefaultSQLiteLibrary);
    procedure CloseDatabase;
    { ----------- Transactions ------------- }

    {Transactions.
    Sqlite engine supports only one active transaction at the same time.
    In other words, you can't use nested transactions.
    The TSqlitePassTransaction object will automatically handle this,
    so any attempt to start a transaction while one is running will have no effect.
    Transactions are really helpfull to speed up and secure data operations.
    You should use them as often as you can.
    If no transaction is active, the TSqlitePassDatabase will always try to start
    a new transaction before writing to the database and commit it.
    This could be time consuming if you need to update or add many records at the
    same time and you should proceed like this :
    @code(
    Database.Transaction.Start;
    ...
    Your code here to update or create records;
    ...
    Database.Transaction.Commit;)}
    Property Transaction: TSqlitePassTransaction Read FTransaction Write FTransaction;
    { -------- SQL Statements ----------- }
    function PrepareStmt(var Stmt: pointer; const sql: String; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Boolean;
    function UnprepareStmt(var Stmt: pointer): Boolean;
    Procedure FinalizePendingStmts;
    { ------------- Sql Exec ---------------- }
    function ExecSQL(Sql: string; CallBackFunc: TSqlitePassExecCallBack=nil): Integer;
    function ExecQuery(Sql: string; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Integer;
    function GetLastInsertRowId: Int64;
    function CheckResult(i: Integer; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Integer;
   end;

implementation
{$i SqlitePassEngine.inc}
{$i SqlitePassDatabase.inc}
{$i SqlitePassDatabaseParts.inc}
{$i SqlitePassDataset.inc}
{$i SqlitePassSqlStmts.inc}
{$i SqlitePassRecordset.inc}
end.
