{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Basic SQL Parser and Statements Management
    Author : Luc DAVID Email: luckylazarus@free.fr
    2006 - 2009

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 05.01.2009

  --------------------------------------------------------------------------- }

  { TSqlitePassSQLSections }

procedure TSqlitePassSQLSections.ClearSectionText(SectionName,
  NewStmt: String);
begin
  Values[SectionName] := '';
end;

function TSqlitePassSQLSections.GetSectionText(SectionName: String): String;
begin
  Result := Values[SectionName];
end;

function TSqlitePassSQLSections.GetSQLStmtFromSections: String;
var
i: integer;
TempText: String;
begin
Result := '';
for i := 0 to Pred(Count) do
    begin
    TempText := GetSectionText(Names[i]);
    if TempText <> ''
       then Result := Result + kwSpace + TempText;
    end;
{ Clear unecessary ';' and Add ';' at the end of the SQL statement }
Result := StringReplace(Result, kwSqlEnd, '', [rfReplaceAll, rfIgnoreCase]);
if Result <> ''
   then Result := Result + kwSqlEnd;
end;

procedure TSqlitePassSQLSections.ReplaceSectionText(SectionName, NewStmt: String);
var
i: integer;
begin
i := IndexOfName(SectionName);
if i > -1
   then Strings[i] := SectionName + '=' + Trim(NewStmt);
end;

procedure TSqlitePassSQLSections.ReplaceSectionTextFromSQLSections(
  SQLSections: TSqlitePassSQLSections; SectionName: String);
begin
  if SQLSections = nil then Exit;
  if SectionName = ''
     then Assign(SQLSections)
     else SetSectionText(SectionName, SQLSections.GetSectionText(SectionName));
end;

procedure TSqlitePassSQLSections.SetSectionText(SectionName, Text: String);
begin
  Values[SectionName] := Text;
end;

procedure TSqlitePassSQLSections.SetSectionTextFromTokenizer(
  SectionName: String; Tokenizer: TSqlitePassSQLTokenizer);
begin
 Values[SectionName] := Tokenizer.Text;
end;

procedure TSqlitePassSQLSections.SplitSQLStmtIntoSections(SQLStmt: String; Keywords: Array of String);

  procedure split;
  var
  i: integer;
  SectionName, SectionText: String;
  Tokenizer: TSqlitePassSQLTokenizer;

  begin
  Tokenizer := TSqlitePassSQLTokenizer.Create;
  Try
  Tokenizer.Text := SQLStmt;

  { Fills the StringList with empty sections }
   Clear;
   For i := Low(Keywords) to High(Keywords) do Add(Keywords[i]+'=');

  { splits sections }
  Repeat
    SectionName := Tokenizer.Token.Text;
    SectionText := Tokenizer.Token.Text;
    SectionText := SectionText + kwSpace + Tokenizer.GetTextAfter(ttKeyWord);
    ReplaceSectionText(SectionName, SectionText);
  Until not Tokenizer.Next(ttKeyword)
  finally
  Tokenizer.Free;
  end;
  end;

begin
If (SQLStmt <> '') and (SizeOf(Keywords) > 0)
   then Split
end;



{ TSqlitePassSqlStmt }

constructor TSqlitePassSqlStmt.Create(Database: TSqlitePassDatabase);
begin
inherited Create;
FAvailableOperations := [];
FDatabase := Database;
FTokenizer := TSqlitePassSQLTokenizer.Create;
FSQLSections := TSqlitePassSQLSections.Create;
FOriginalSQLSections := TSqlitePassSQLSections.Create;
FStmtHandle := nil;
SQL := '';
end;


destructor TSqlitePassSqlStmt.Destroy;
begin
Unprepare;
FOriginalSQLSections.Free;
FSQLSections.Free;
FTokenizer.Free;
FDatabase := nil;
Inherited Destroy;
end;

procedure TSqlitePassSqlStmt.SetFSQLText(Value: String);
begin
FOriginalSQLText := Value;
FSQLText := Value;
GetStmtType;
FOriginalSQLSections.SplitSQLStmtIntoSections(FOriginalSQLText, FStmtKeywords);
FSQLSections.SplitSQLStmtIntoSections(FSQLText, FStmtKeywords);
end;

function TSqlitePassSqlStmt.GetFSQLText: String;
begin
If FSQLSections.Count > 0
   then FSQLText := FSQLSections.GetSQLStmtFromSections;
Result := FSQLText;
end;

function TSqlitePassSqlStmt.GetStmtType: StmtType;
var
i: Integer;
begin
SetLength(FStmtKeyWords, 0);
Result := SQLitePassDbo.GetStmtType(FSQLText);
if Result = stSelect then
   begin
   SetLength(FStmtKeyWords, High(SelectStmtKeywords)+1);
   for i := 0 to High(SelectStmtKeywords)
       do FStmtKeyWords[i] := SelectStmtKeywords[i];
   end;
end;

procedure TSqlitePassSqlStmt.Prepare;
begin
  FDatabase.FEngine.PrepareStmt(FStmtHandle, PChar(FSQLText));
end;

procedure TSqlitePassSqlStmt.UnPrepare;
begin
 if Assigned(FDatabase) and Assigned(FDatabase.FEngine) and Assigned(FStmtHandle) then
    begin
    FDatabase.Engine.UnprepareStmt(FStmtHandle);
    FStmtHandle := nil;
    end;
end;

procedure TSqlitePassSqlStmt.QuoteString(var S: String);
begin
If S <> '' then
   begin
   S := StringReplace(S, kwDot, FDatabase.Options.FQuoteOpen + kwDot + FDatabase.Options.FQuoteClose, [rfReplaceAll]);
   S := FDatabase.Options.FQuoteOpen + S + FDatabase.Options.FQuoteClose;
   end;
end;

Procedure TSqlitePassSqlStmt.UnquoteString(var S: String; All: Boolean = True);
var
i: integer;
begin
 if S = '' then Exit;

 if All
    then for i := 0 to High(QuoteCharsArray) do
         S := StringReplace(S, QuoteCharsArray[i], '', [rfReplaceAll])
    else begin
         S := StringReplace(S, FDatabase.Options.FQuoteOpen, '', [rfReplaceAll]);
         S := StringReplace(S, FDatabase.Options.FQuoteClose, '', [rfReplaceAll]);
         end;
end;

procedure TSqlitePassSqlStmt.RestoreOriginalStmt;
begin
 SQL := OriginalSQL;
end;


{ TSqlitePassSelectStmt }

constructor TSqlitePassSelectStmt.Create(Dataset: TSqlitePassDataset);
begin
if Assigned(Dataset.FDatabase)
   then Inherited Create(Dataset.FDatabase)
   else Inherited Create(TSqlitePassDatabase(nil));
FDataset := Dataset;
FTableDefs := TSqlitePassSelectStmtTableDefs.Create(Self);
FFieldDefs := TSqlitePassSelectStmtFieldDefs.Create(nil);
FAutoIncFieldDefs := TSqlitePassSelectStmtFieldDefs.Create(nil);
FPrimaryKeyStmt := '';
end;

destructor TSqlitePassSelectStmt.Destroy;
begin
Unprepare;
FAutoIncFieldDefs.Free;
FFieldDefs.Free;
FTableDefs.Free;
FDataset := nil;
inherited Destroy;
end;

{ Adds primary keys to a select statement if needed and then compiles the statement.
  Finally, calls PrepareSQLStmts to prepare and compile corresponding
  Insert-Update-Delete statements }
procedure TSqlitePassSelectStmt.Prepare;
begin
try
  FAvailableOperations := [];
  if (FSQLText = '') or (SQLType <> StSelect) then Exit;

  { Gets the tables involved in the query in order to add Primary Keys to the statement }
  AddPrimaryKeys;

  { Rebuilds the definitive Schema including Primary Keys }
  BuildSchema;

  { Prepares the SQL statements needed for inserting, updating or deleting fields data }
  PrepareSQLStmts;
except
  Raise;
end;
end;

Procedure TSqlitePassSelectStmt.UnPrepare;
begin
UnprepareSQLStmts;
Inherited UnPrepare;
end;

Procedure TSqlitePassSelectStmt.AddPrimaryKeys;
var
i, ColumnCount: Integer;
TablesName: TStringList;
TableName, SelectStmt: String;
begin
Try
  TablesName := TStringList.Create;
  FPrimaryKeyCount := 0;

  { Clears the previous prepared schema }
  UnprepareSQLStmts;
  UnPrepare;

  { Removes any previously added internal "_rowid_" primary keys from the select statemenent }
  SQLSections.ReplaceSectionTextFromSQLSections(OriginalSQLSections, kwSelect);

  { Compiles the select statement }
  FDatabase.FEngine.PrepareStmt(FStmtHandle, PChar(SQL));

  ColumnCount := SqliteDbv3_column_count(StmtHandle);
  for i := 0 to Pred(ColumnCount) do
      begin
      TableName := SqliteDbv3_column_table_name(StmtHandle, i);

      // TODO -> Mark as ReadOnly ?
      if TableName = ''
         then begin
              FPrimaryKeyStmt := '';
              Exit;
              end
         else if TablesName.IndexOf(TableName) = -1
                 then TablesName.Add(TableName);
      end;

  if TablesName.Count > 1
     then for i := 0 to Pred(TablesName.Count) do
          begin
          TableName := TablesName[i];
          QuoteString(TableName);
          FPrimaryKeyStmt := FPrimaryKeyStmt + kwSpace + TableName + '._Rowid_,';
          end
          else FPrimaryKeyStmt := ' _Rowid_,';

  Tokenizer.Text := SQLSections.GetSectionText(kwSelect);
  Tokenizer.InsertAfter(kwSelect, FPrimaryKeyStmt);
  SQLSections.SetSectionTextFromTokenizer(kwSelect, Tokenizer);
  FPrimaryKeyCount := TablesName.Count;
finally
  TablesName.Free;
  UnPrepare;
end;
end;

{ Builds a schema representation of the sql statement :
  A schema is made of tables and fieldDefs assigned from
  tables and fieldDefs info stored in the global database schema
}
procedure TSqlitePassSelectStmt.BuildSchema;
var
i, ColumnCount: Integer;
TableName, FieldName: String;
begin
Try
{ Clears the previous prepared schema }
UnprepareSQLStmts;
UnPrepare;

{ Compiles the select statement }
FDatabase.FEngine.PrepareStmt(FStmtHandle, PChar(SQL));

ColumnCount := SqliteDbv3_column_count(StmtHandle);

{ i starts at FPrimaryKeyCount because we skip the n first colonnes which are the RowId(s) }
for i := FPrimaryKeyCount to Pred(ColumnCount) do
     begin
     TableName := SqliteDbv3_column_table_name(StmtHandle, i);
     FieldName := SqliteDbv3_column_origin_name(StmtHandle, i);
     { 'i' is used to set the position of the field in Record buffer
     because field are created in the same order in buffer }

     if (TableName = '') or (FieldName = '')
        then FillTableDefsFromInternalCalculatedField(i - FPrimaryKeyCount)
        else FillTableDefs(TableName, FieldName, i - FPrimaryKeyCount);
     end;
Include(FAvailableOperations, aoRead);
except
Raise;
end;
end;


{ The default column-name which identifies rows in SQLite is 'rowid',
  but for tables which have a primary integer key, this doesn't work
  (even though the documentation says it ought to). Therefore it is necessary
  to identify, for the given database, which tables need to use a different
  column-name for this. This gives a hash of table->column names. }
function TSqlitePassSelectStmt.FillTableDefs(Const TableName, FieldName: String; RecFieldIndex: Integer): Boolean;
var
DatabaseTableDef: TSqlitePassTableDef;
StmtTableDef: TSqlitePassSelectStmtTableDef;
DatabaseFieldDef: TSqlitePassTableFieldDef;
StmtFieldDef: TSqlitePassSelectStmtFieldDef;

begin
Result := False;
  { Create a TableDef for a given table name }
   DatabaseTableDef := FDataset.Database.TableDefs.FindTable(TableName);
   if Assigned(DatabaseTableDef) then
      begin
      { if the TableDef doesn't already exist in select schema, we create it }
      StmtTableDef := TableDefs.FindTable(TableName);
      if StmtTableDef = nil then
         begin
         StmtTableDef := TSqlitePassSelectStmtTableDef.Create(FTableDefs);
         { Assigns doesn't fill FieldDefs }
         StmtTableDef.Assign(DatabaseTableDef);
         TableDefs.Add(StmtTableDef);
         end;

      { FieldDefs }
      DatabaseFieldDef := DatabaseTableDef.FieldDefs.FindFieldDef(TableName +  '.' + FieldName);
      if Assigned(DatabaseFieldDef) then
         begin
        { a same FieldDef can appear several times in a schema, so we always create one }
         StmtFieldDef := TSqlitePassSelectStmtFieldDef.Create(StmtTableDef.FFieldDefs);
         StmtFieldDef.Assign(DatabaseFieldDef);
         StmtFieldDef.FRecordFieldIndex := RecFieldIndex;
         StmtTableDef.FieldDefs.Add(StmtFieldDef);
         { Add this FieldDef to the SelectStmt fieldDefs too.
           That way End-Users will have a more convenient way to loop into FieldDefs }
         FFieldDefs.Add(StmtFieldDef);
         { Add this FieldDef to the AutoIncfieldDefs if it is an AutoInc Field }
         if StmtFieldDef.FAutoInc
            then FAutoIncFieldDefs.Add(StmtFieldDef);
         Result := True;
         end;
      end;
end;

procedure TSqlitePassSelectStmt.FillTableDefsFromInternalCalculatedField(RecFieldIndex: Integer);
var
//DatabaseTableDef: TSqlitePassTableDef;
StmtTableDef: TSqlitePassSelectStmtTableDef;
//DatabaseFieldDef: TSqlitePassTableFieldDef;
StmtFieldDef: TSqlitePassSelectStmtFieldDef;
begin
{ if the 'InternalCalcTable' TableDef doesn't already exist in select schema, we create it }
StmtTableDef := TableDefs.FindTable('InternalCalcTable');
if StmtTableDef = nil then
   begin
   StmtTableDef := TSqlitePassSelectStmtTableDef.Create(FTableDefs);
   StmtTableDef.FTableName := 'InternalCalcTable';
   StmtTableDef.FCanModify := False;
   TableDefs.Add(StmtTableDef);
   end;

{ a same FieldDef can appear several times in a schema, so we always create one }
StmtFieldDef := TSqlitePassSelectStmtFieldDef.Create(StmtTableDef.FFieldDefs);
With StmtFieldDef do
     begin
     FFieldName := SqliteDbv3_column_name(StmtHandle, RecFieldIndex);
     // FNativeDataType := ;
     FDataType := ftString;
     FPrimaryKey := False;
     // FieldTypesTranslationRules.SetFieldDataType(aFieldDef);
     Case FDataType of
          ftString : FSize := 30;
          ftBCD    : FPrecision := 2; // TODO Rework
          end;
     end;

StmtTableDef.FieldDefs.Add(StmtFieldDef);
{ Add this FieldDef to the SelectStmt fieldDefs too.
 That way End-Users will have a more convenient way to loop into FieldDefs }
FFieldDefs.Add(StmtFieldDef);
end;

procedure TSqlitePassSelectStmt.PrepareSQLStmts;
var
i: Integer;
begin
Include(FAvailableOperations, aoInsert);
Include(FAvailableOperations, aoUpdate);
Include(FAvailableOperations, aoDelete);
For i := 0 to Pred(FTableDefs.Count) do
    begin
    FTableDefs[i].PrepareInsertDeleteSQLStmts;
    If Not (aoInsert in FTableDefs[i].AvailableOperations)
       then Exclude(FAvailableOperations, aoInsert);
    If Not (aoUpdate in FTableDefs[i].AvailableOperations)
       then Exclude(FAvailableOperations, aoUpdate);
    If Not (aoDelete in FTableDefs[i].AvailableOperations)
       then Exclude(FAvailableOperations, aoDelete);
    end;
end;

Procedure TSqlitePassSelectStmt.UnprepareSQLStmts;
var
i: Integer;
begin
FPrimaryKeyStmt := '';
For i := 0 to Pred(FTableDefs.Count)
 do FTableDefs[i].UnprepareSQLStmts;
FAvailableOperations := [];
FTableDefs.ClearAndFreeItems;
FAutoIncFieldDefs.Clear;
FFieldDefs.Clear;
end;

procedure TSqlitePassSelectStmt.Post(PostType: TSqlitePassPostType);
var
i: Integer;
RowId: Int64;
begin
try
  FDataset.FDatabase.FEngine.Transaction.StartInternalTransaction;

  { If TDatatset.First is called then FCurrentRecordIndex = -1
    so we have to set it to 0 }
  If FDataset.FRecordset.FCurrentRecordIndex = -1
   then Inc(FDataset.FRecordset.FCurrentRecordIndex);

  For i := 0 to Pred(FTableDefs.Count) do
     begin
     RowId := FDataset.FRecordset.GetRowId(FDataset.FRecordset.FCurrentRecordIndex, i);
     Case PostType of
           ptInsert: FTableDefs[i].PostInsert;
           ptUpdate: FTableDefs[i].PostUpdate(RowId);
           ptDelete: FTableDefs[i].PostDelete(RowId);
           end;
     end;
   { Reset Modified Fields }
  FDataset.FRecordset.SetRecordState(FTableDefs.FSQLStmt.FDataset.FRecordset.FCurrentRecordIndex, rsUnmodified);
  FDataset.FDatabase.FEngine.Transaction.CommitInternalTransaction;
except
  FDataset.FDatabase.FEngine.Transaction.RollbackInternalTransaction;
  Raise;
end;
end;


{ TSqlitePassSelectStmtTableDefs }

constructor TSqlitePassSelectStmtTableDefs.Create(Owner: TSqlitePassSelectStmt);
begin
inherited Create;
FSQLStmt := Owner;
end;

destructor TSqlitePassSelectStmtTableDefs.Destroy;
begin
ClearAndFreeItems;
FSQLStmt := nil;
inherited Destroy;
end;

procedure TSqlitePassSelectStmtTableDefs.ClearAndFreeItems;
var
i: Integer;
TableDef: TSqlitePassSelectStmtTableDef;
begin
{ Clear the internal items }
for i := 0 to Pred(Count) do
  begin
  TableDef := GetItem(i);
  TableDef.Free;
  end;
{ Clear the list pointers }
inherited Clear;
end;

function TSqlitePassSelectStmtTableDefs.FindTable(
  const Value: string): TSqlitePassSelectStmtTableDef;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassSelectStmtTableDef(inherited Items[I]);
    if AnsiCompareText(Result.TableFullName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassSelectStmtTableDefs.GetItem(Index: Integer): TSqlitePassSelectStmtTableDef;
begin
Result:=TSqlitePassSelectStmtTableDef(Inherited Items[Index]);
end;

procedure TSqlitePassSelectStmtTableDefs.SetItem(Index: Integer;
  const Value: TSqlitePassSelectStmtTableDef);
begin
  Put(Index,Value);
end;

{ TSqlitePassSelectStmtTableDef }

constructor TSqlitePassSelectStmtTableDef.Create(Owner: TSqlitePassSelectStmtTableDefs);
begin
 inherited Create;
 FTableDefs := Owner;
 FFieldDefs := TSqlitePassSelectStmtFieldDefs.Create(Self);
end;


destructor TSqlitePassSelectStmtTableDef.Destroy;
begin
 UnprepareSQLStmts;
// FPrimaryKeyField := nil;
 FFieldDefs.Free;
 FTableDefs := nil;
 inherited Destroy;
end;

procedure TSqlitePassSelectStmtTableDef.Assign(Source: TSqlitePassTableDef);
begin
 FAttachedDatabase := Source.FAttachedDatabase;
 FAliasName  := Source.AliasName;
 FCanModify  := Source.CanModify;
 FOnConflict := Source.OnConlict;
 FSql        := Source.Sql;
 FTableName  := Source.TableName;
 FTemporary  := Source.Temporary;
end;


{ Prepare standard Insert, Delete sql statements to speed up
  engine operations. We will then just need to fill in the values when posting the data}
procedure TSqlitePassSelectStmtTableDef.PrepareInsertDeleteSQLStmts;
var
i: Integer;
FieldName, InsertStmtFields, InsertStmtValues: String;

begin
UnprepareSQLStmts;

{ Insert statement }
InsertStmtFields := '';
InsertStmtValues := '';
FieldName := '';
For i := 0 to Pred(FFieldDefs.Count) do
    begin
    FieldName := FieldDefs[i].FieldName;
    FTableDefs.FSQLStmt.QuoteString(FieldName);
    InsertStmtFields := InsertStmtFields + FieldName + ', ';
    InsertStmtValues := InsertStmtValues + ':val_' + IntToStr(i) + ', ';
    end;
System.Delete(InsertStmtFields,Length(InsertStmtFields)-1,2);
System.Delete(InsertStmtValues,Length(InsertStmtValues)-1,2);

FInsertStmtText := 'INSERT INTO "' + TableFullName + '" (' + InsertStmtFields +') VALUES (' + InsertStmtValues + ');';
FDeleteStmtText := 'DELETE FROM "' + TableFullName + '" WHERE RowId=?;';
Try
with FTableDefs.FSQLStmt.FDatabase.FEngine do
     begin
     if PrepareStmt(FInsertStmt, FInsertStmtText, vlSilent)
        then Include(FAvailableOperations, aoInsert);
     if PrepareStmt(FDeleteStmt, FDeleteStmtText, vlSilent)
        then Include(FAvailableOperations, aoDelete);
     Include(FAvailableOperations, aoUpdate);
     end;
except
{ Exception while retreiving data are trapped here
  so we can continu with the next available value }
On E: Exception do
end;
end;

Procedure TSqlitePassSelectStmtTableDef.UnprepareSQLStmts;
begin
{ Unprepare any previously prepared statement }
With FTableDefs.FSQLStmt.FDatabase.FEngine do
     begin
     UnprepareStmt(FInsertStmt);
     UnprepareStmt(FDeleteStmt);
     end;
FInsertStmtText := '';
FDeleteStmtText := '';
FAvailableOperations := [];
end;


{'INSERT INTO "' + TableFullName + '" (' + InsertStmtFields +') VALUES (' + InsertStmtValues + ');'; }
procedure TSqlitePassSelectStmtTableDef.PostInsert;
var
i: Integer;
begin
if not (aoInsert in FAvailableOperations) then Exit;
 Try
 FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_clear_binding(FInsertStmt));
 for i := 0 to Pred(FieldDefs.Count) do
     BufferToSqliteValue(FInsertStmt, FieldDefs[i].DataType,
                         FTableDefs.FSQLStmt.FDataset.FRecordset.FRecordsCount,
                         FieldDefs[i].FRecordFieldIndex,
                         Succ(i));

 FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(FInsertStmt), vlSilent);
 FLastInsertedRowId := FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.GetLastInsertRowId;
 finally
 SqliteDbv3_reset(FInsertStmt);
 end;
end;

{ 'UPDATE "' + TableFullName + '" SET ' + UpdateStmtFields + ' WHERE RowId=?;' }
procedure TSqlitePassSelectStmtTableDef.PostUpdate(RowId: Int64);
var
i, StmtFieldCount: Integer;
FUpdateStmt: Pointer;
FUpdateStmtText: String;

  function PrepareUpdateStmt: Boolean;
  var
  UpdateStmtFields, FieldName: String;
  i: integer;
  begin
  Result := False;
  UpdateStmtFields := '';
  FieldName := '';
  for i := 0 to Pred(FFieldDefs.Count) do
    begin
    FieldDefs[i].FWasModified := False;
    if FTableDefs.FSQLStmt.FDataset.FRecordset.FieldWasModified(FTableDefs.FSQLStmt.FDataset.FRecordset.FCurrentRecordIndex, FFieldDefs[i].FRecordFieldIndex) then
       begin
       FieldDefs[i].FWasModified := True;
       FieldName := FieldDefs[i].FieldName;
       FTableDefs.FSQLStmt.QuoteString(FieldName);
       UpdateStmtFields := UpdateStmtFields + FieldName + '= ?, ';
       Result := True;
       end;
    end;
  System.Delete(UpdateStmtFields,Length(UpdateStmtFields)-1,2);
  FUpdateStmtText := 'UPDATE "' + TableFullName + '" SET ' + UpdateStmtFields + ' WHERE RowId=?;';
  end;

begin
 if not (aoUpdate in FAvailableOperations) then Exit;
 if not PrepareUpdateStmt then Exit;
 FUpdateStmt := nil;
 StmtFieldCount := 0;
 if FTableDefs.FSQLStmt.FDatabase.FEngine.PrepareStmt(FUpdateStmt, FUpdateStmtText) then
    try
    for i := 0 to Pred(FFieldDefs.Count) do
        begin
        if FFieldDefs[i].FWasModified  then
           begin
           Inc(StmtFieldCount);
           BufferToSqliteValue(FUpdateStmt, FFieldDefs[i].DataType,
                         FTableDefs.FSQLStmt.FDataset.FRecordset.FCurrentRecordIndex,
                         FieldDefs[i].FRecordFieldIndex,
                         StmtFieldCount);
           end;
        end;
    FTableDefs.FSQLStmt.FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(FUpdateStmt, Succ(StmtFieldCount), RowId));
    FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(FUpdateStmt), vlShow);
    finally
    SqliteDbv3_finalize(FUpdateStmt);
    end;
end;

{ 'DELETE FROM "' + TableFullName + '" WHERE RowId=?;'; }
procedure TSqlitePassSelectStmtTableDef.PostDelete(RowId: Int64);
begin
 if not (aoDelete in FAvailableOperations) then Exit;
 try
 FTableDefs.FSQLStmt.FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(FDeleteStmt, 1, RowId));
 FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(FDeleteStmt), vlLog);
 finally
 SqliteDbv3_reset(FDeleteStmt);
 end;
end;


{ *** Conversion functions from an internal MemRecord storage to a SQLite Db Value *** }
{ Translates a value from internal buffer to an sqlite value.
  Then binds this value to parameter in INSERT - UPDATE - DELETE SQL statement
  StmtFieldNo is 1 based, FieldIndex is Zero based }
procedure TSqlitePassSelectStmtTableDef.BufferToSqliteValue(Const PreparedStmt: Pointer; Const DataType: TFieldType; RecordIndex, RecordFieldIndex, StmtFieldNo: Integer);
var
StrValue: String;
IntegerValue: Integer;
Int64Value: Int64;
DoubleValue: Double;
Buffer: PChar;
MemoryStream: TMemoryStream;
FDatabase: TSqlitePassDatabase;
begin

if not Assigned(PreparedStmt) then Exit;

if FTableDefs.FSQLStmt.FDataset.FRecordset.FieldIsNull(RecordIndex, RecordFieldIndex) then
   begin
   SqliteDbv3_bind_null(PreparedStmt, StmtFieldNo);
   Exit;
   end;

Buffer := FTableDefs.FSQLStmt.FDataset.FRecordset.GetFieldBuffer(RecordIndex, RecordFieldIndex);
{ Shorcut for coding... }
FDatabase := FTableDefs.FSQLStmt.FDataset.FDatabase;

case DataType of

  ftDate:
    Case FDatabase.DatatypeOptions.DateStorage of
      asInteger  : FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(PreparedStmt, StmtFieldNo, PInteger(Buffer)^));
      asText     : begin
                   FDatabase.FTranslator.IntegerToDateText(PInteger(Buffer)^, StrValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, StmtFieldNo, PChar(StrValue), Length(StrValue), SQLITE_TRANSIENT));
                   end;
      end;

  ftTime:
    Case FDatabase.DatatypeOptions.TimeStorage of
      asInteger  : FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(PreparedStmt, StmtFieldNo, PInteger(Buffer)^));
      asText     : begin
                   FDatabase.FTranslator.IntegerToTimeText(PInteger(Buffer)^, StrValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, StmtFieldNo, PChar(StrValue), Length(StrValue), SQLITE_TRANSIENT));
                   end;
      end;

 ftDateTime:
   Case FDatabase.DatatypeOptions.DateTimeStorage of
     dtsDateTime : FDatabase.Engine.CheckResult(SqliteDbv3_bind_double(PreparedStmt, StmtFieldNo, PDouble(Buffer)^));
     dtsJulian   : begin
                   FDatabase.FTranslator.DoubleToJulianDateTime(PDouble(Buffer)^, DoubleValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_double(PreparedStmt, StmtFieldNo, DoubleValue))
                   end;
     dtsText     : begin
                   FDatabase.FTranslator.DoubleToDateTimeText(PDouble(Buffer)^, StrValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, StmtFieldNo, PChar(StrValue), Length(StrValue), SQLITE_TRANSIENT));
                   end;
     dtsUnix     : begin
                   FDatabase.FTranslator.DoubleToUnixDateTime(PDouble(Buffer)^, Int64Value);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(PreparedStmt, StmtFieldNo, Int64Value));
                   end;
     dtsMac      : begin
                   FDatabase.FTranslator.DoubleToMacDateTime(PDouble(Buffer)^, Int64Value);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(PreparedStmt, StmtFieldNo, Int64Value));
                   end;
     end;

 ftInteger, ftSmallInt, ftWord:
   FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(PreparedStmt, StmtFieldNo, PInteger(Buffer)^));

 ftAutoInc, ftLargeint:
     FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(PreparedStmt, StmtFieldNo, PInt64(Buffer)^));

 ftBoolean:
   Case FDatabase.DatatypeOptions.BooleanStorage of
     asInteger   : begin
                   FDatabase.FTranslator.WordBoolToInteger(PWordBool(Buffer)^, IntegerValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(PreparedStmt, StmtFieldNo, IntegerValue));
                   end;
     asText      : begin
                   FDatabase.FTranslator.WordBoolToBooleanText(PWordBool(Buffer)^, StrValue);
                   FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, StmtFieldNo, PChar(StrValue), Length(StrValue), SQLITE_TRANSIENT));
                   end;
     end;

 ftFloat, ftBCD, ftCurrency:
   FDatabase.Engine.CheckResult(SqliteDbv3_bind_double(PreparedStmt, StmtFieldNo, PDouble(Buffer)^));

 ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString:
   begin
   StrValue := FTableDefs.FSQLStmt.FDataset.FRecordset.GetFieldAsString(RecordIndex, RecordFieldIndex);
   FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, StmtFieldNo, PChar(StrValue), Length(StrValue), SQLITE_TRANSIENT));
   end;

 ftBlob, ftGraphic, ftMemo, ftFmtMemo:
   begin
   MemoryStream := FTableDefs.FSQLStmt.FDataset.FRecordset.GetFieldAsBlob(RecordIndex, RecordFieldIndex);
   if Assigned(MemoryStream)
      then FDatabase.Engine.CheckResult(SqliteDbv3_bind_blob(PreparedStmt, StmtFieldNo, MemoryStream.Memory, MemoryStream.Size, SQLITE_TRANSIENT));
   end;

 else FDatabase.DatabaseError.RaiseExceptionFmt(Msg3011,[FieldTypeToString(DataType)]);
 end; { Case }
end;



{ TSqlitePassSelectStmtFieldDef }

constructor TSqlitePassSelectStmtFieldDef.Create(
  Owner: TSqlitePassSelectStmtFieldDefs);
begin
 Inherited Create;
 FFieldDefs := Owner;
end;

destructor TSqlitePassSelectStmtFieldDef.Destroy;
begin
 FFieldDefs := nil;
 Inherited Destroy;
end;

function TSqlitePassSelectStmtFieldDef.GetFFieldFullName: String;
begin
 Result := FFieldDefs.FTableDef.TableName + '.' + FFieldName;
end;

procedure TSqlitePassSelectStmtFieldDef.Assign(Source: TSqlitePassTableFieldDef);
begin
 inherited Assign(TSqlitePassGenericFieldDef(Source));
end;

{ TSqlitePassSelectStmtFieldDefs }

constructor TSqlitePassSelectStmtFieldDefs.Create(
  Owner: TSqlitePassSelectStmtTableDef);
begin
 Inherited Create;
 FTableDef := Owner;
end;

destructor TSqlitePassSelectStmtFieldDefs.Destroy;
begin
 { We clear Items only if FTableDef is Assigned otherwise it will be freed twice:
   TSqlitePassSelectStmtFieldDefs is used
     by TSqlitePassSelectStmtTableDef that assigns FTableDefs
     by TSqlitePassSelectStmt that does not assign FTableDefs
     but maintains a list pointing on the same FieldDefs as TSqlitePassSelectStmtTableDef
   }
 if Assigned(FTableDef) then
    begin
    ClearAndFreeItems;
    FTableDef := nil;
    end;
 Inherited Destroy;
end;

procedure TSqlitePassSelectStmtFieldDefs.ClearAndFreeItems;
var
i: Integer;
begin
{ Clears the internal items }
for i := 0 to Pred(Count)
    do Items[i].Free;
{ Clears the list pointers }
inherited Clear;
end;

function TSqlitePassSelectStmtFieldDefs.GetItem(
  Index: Integer): TSqlitePassSelectStmtFieldDef;
begin
  Result:=TSqlitePassSelectStmtFieldDef(Inherited Items[Index]);
end;

procedure TSqlitePassSelectStmtFieldDefs.SetItem(Index: Integer;
  const Value: TSqlitePassSelectStmtFieldDef);
begin
  Put(Index,Value);
end;

procedure TSqlitePassSelectStmtFieldDefs.Assign(Source: TSqlitePassTableFieldDefs);
var
i: integer;
NewFieldDef: TSqlitePassSelectStmtFieldDef;
begin
for i := 0 to Pred(Source.Count) do
   begin
   NewFieldDef := TSqlitePassSelectStmtFieldDef.Create(Self);
   NewFieldDef.Assign(Source[i]);
   Self.Add(NewFieldDef);
   end;
end;

{ TSqlitePassSQLToken }

constructor TSqlitePassSQLToken.Create;
begin
 TokenType := ttUnknown;
end;


{ TSqlitePassTokens }

procedure TSqlitePassSQLTokens.ClearAndFreeItems;
var
i: Integer;
begin
{ Clear the internal items }
for i := 0 to Pred(Count) do
    Items[i].Free;
{ Clear the list pointers }
inherited Clear;
end;

destructor TSqlitePassSQLTokens.Destroy;
begin
ClearAndFreeItems;
Inherited;
end;

function TSqlitePassSQLTokens.GetItem(Index: Integer): TSqlitePassSQLToken;
begin
 Result := TSqlitePassSQLToken(Inherited Items[Index]);
end;

procedure TSqlitePassSQLTokens.SetItem(Index: Integer;
  const Value: TSqlitePassSQLToken);
begin
  Put(Index,Value);
end;



{ TSqlitePassTokenizer }

constructor TSqlitePassSQLTokenizer.Create;
begin
  Inherited;
  FCurrentTokenIndex := -1;
  FTokens := TSqlitePassSQLTokens.Create;
end;

destructor TSqlitePassSQLTokenizer.Destroy;
begin
  FTokens.Free;
  Inherited;
end;


function TSqlitePassSQLTokenizer.First: Boolean;
begin
 if FTokens.Count > 0
    then FCurrentTokenIndex := 0
    else FCurrentTokenIndex := -1;
 Result := (FCurrentTokenIndex <> -1);
end;

function TSqlitePassSQLTokenizer.GetFCount: Integer;
begin
 Result := FTokens.Count;
end;

function TSqlitePassSQLTokenizer.GetFCurrentToken: TSqlitePassSQLToken;
begin
  if FCurrentTokenIndex > -1
     then Result := FTokens[FCurrentTokenIndex]
     else Result := nil;
end;

function TSqlitePassSQLTokenizer.GetFEOF: Boolean;
begin
  Result := (FTokens.Count > 0) and (FCurrentTokenIndex = Pred(FTokens.Count));
end;


function TSqlitePassSQLTokenizer.GetFText: String;
var
i: Integer;
begin
if FTokens.Count > 0 then
   begin
   Result := '';
   for i := 0 to Pred(FTokens.Count)
       do Result := Result + FTokens[i].Text;
   end;
end;

procedure TSqlitePassSQLTokenizer.InsertAfter(TokenText, InsertedStmt: String);
var
NewToken: TSqlitePassSQLToken;
NewText: String;
begin
If Locate(TokenText) then
   begin
   NewToken := TSqlitePassSQLToken.Create;
   NewToken.Text := InsertedStmt;
   FTokens.Insert(Succ(FCurrentTokenIndex), NewToken);
  end;
end;

procedure TSqlitePassSQLTokenizer.InsertBefore(TokenText,
  InsertedStmt: String);
var
NewToken: TSqlitePassSQLToken;
NewText: String;
begin
If Locate(TokenText) then
   begin
   NewToken := TSqlitePassSQLToken.Create;
   NewToken.Text := InsertedStmt;
   FTokens.Insert(FCurrentTokenIndex, NewToken);
  end;
end;

function TSqlitePassSQLTokenizer.Last: Boolean;
begin
 FCurrentTokenIndex := Pred(FTokens.Count);
 Result := (FCurrentTokenIndex <> -1);
end;

function TSqlitePassSQLTokenizer.Locate(TokenText: String): Boolean;
var
i: Integer;
begin
 Result := False;
 for i := 0 to Pred(FTokens.Count) do
     if AnsiCompareText(FTokens[i].Text, TokenText) = 0 then
        begin
        FCurrentTokenIndex := i;
        Result := True;
        Break;
        end;
end;

function TSqlitePassSQLTokenizer.Locate(
  TokenType: TSqlitePassSQLTokenType): Boolean;
var
i: Integer;
begin
 Result := False;
 for i := 0 to Pred(FTokens.Count) do
     if FTokens[i].TokenType = TokenType then
        begin
        FCurrentTokenIndex := i;
        Result := True;
        Break;
        end;
end;

function TSqlitePassSQLTokenizer.Next(TokenType: TSqlitePassSQLTokenType): Boolean;
begin
Result := False;
while FCurrentTokenIndex < Pred(FTokens.Count) do
  begin
  Inc(FCurrentTokenIndex);
  if FTokens[FCurrentTokenIndex].TokenType = TokenType then
     begin
     Result := True;
     Break;
     end;
  end;
end;

function TSqlitePassSQLTokenizer.Next: Boolean;
begin
Result := False;
if FCurrentTokenIndex < Pred(FTokens.Count) then
   begin
   Inc(FCurrentTokenIndex);
   Result := True;
   end;
end;

function TSqlitePassSQLTokenizer.Prior(
  TokenType: TSqlitePassSQLTokenType): Boolean;
begin
Result := False;
while FCurrentTokenIndex > 0 do
  begin
  Dec(FCurrentTokenIndex);
  if FTokens[FCurrentTokenIndex].TokenType = TokenType then
     begin
     Result := True;
     Break;
     end;
  end;
end;

function TSqlitePassSQLTokenizer.Prior: Boolean;
begin
if FCurrentTokenIndex > 0
   then Dec(FCurrentTokenIndex);
end;

procedure TSqlitePassSQLTokenizer.Remove(TokenText: String);
begin
 If Locate(TokenText) then FTokens.Delete(FCurrentTokenIndex);
end;

procedure TSqlitePassSQLTokenizer.RemoveAfter(TokenText, RemovedStmt: String);
begin
If Locate(TokenText) and (FCurrentTokenIndex < Pred(FTokens.Count))
   then FTokens.Delete(Succ(FCurrentTokenIndex));
end;

procedure TSqlitePassSQLTokenizer.RemoveBefore(TokenText, RemovedStmt: String);
begin
If Locate(TokenText) and (FCurrentTokenIndex > 0)
   then FTokens.Delete(Pred(FCurrentTokenIndex));
end;

procedure TSqlitePassSQLTokenizer.Replace(OldText, NewText: String);
begin
 If Locate(OldText) then FTokens[FCurrentTokenIndex].Text := NewText;
end;

procedure TSqlitePassSQLTokenizer.SetFText(const Value: String);
begin
FText := Value;
Tokenize;
end;

function TSqlitePassSQLTokenizer.GetTextAfter(TokenType: TSqlitePassSQLTokenType): String;
var
i, Index: Integer;
begin
  Index := Succ(FCurrentTokenIndex);
  Result := '';
  for i := Index to Pred(FTokens.Count) do
      begin
      if FTokens[i].TokenType = TokenType
         then Break;
      Result := Result + FTokens[i].Text;
      end;
end;

Procedure TSqlitePassSQLTokenizer.Clean;
var
LastChar,i: integer;
OldSQL, NewSQL: String;
begin
(*OldSQL := Trim(FSQLText);
OldSQL := StringReplace(OldSQL, kwLineEnd, KwSpace, [rfIgnoreCase]);
LastChar := Length(OldSQL);
If LastChar = 0 then exit;
i := 1;
NewSQL := '';
Repeat
  NewSQL := NewSQL + OldSQL[i];
  Inc(I);
  { Trim extra spaces }
  While (OldSQL[i-1] = kwSpace) and (OldSQL[i] = kwSpace) do Inc(i)
Until (i > LastChar) or (OldSQL[i] = kwSqlEnd);
Dec(i);
{ Add ';' at the end if needed }
if OldSql[i] <> kwSqlEnd
   then NewSQL := NewSQL + kwSqlEnd;
FSQLText := NewSQL;*)
end;


procedure TSqlitePassSQLTokenizer.Tokenize(Options: TSqlitePassSQLTokenizerOptions);
var
CurrentChar, TokenStart: PChar;
NewToken: TSqlitePassSQLToken;

  { List of Sub-Procedures to handle specialized processing depending on the token kind}
  Procedure ProcessEOF;
  begin
    NewToken.TokenType := ttEOF;
  end;

  Procedure ProcessLF;
  begin
   Inc(CurrentChar);
  end;

  Procedure ProcessCR;
  begin
   Inc(CurrentChar);
  end;

  Procedure ProcessSpace;
  begin
    NewToken.TokenType := ttWhiteSpace;
    Inc(CurrentChar);
  end;

  Procedure ProcessNumber;
  begin
    NewToken.TokenType := ttNumber;
    Inc(CurrentChar);
  end;

  Procedure ProcessIdentifier;

    function IsKeyword: Boolean;
    var
    i: Integer;
    TempText: String;
    begin
    Result := False;
    { Checks if the token is a SQL Keyword }
    For i := 0 to High(SelectStmtKeywords) do
        begin
        SetString(TempText, TokenStart, Length(SelectStmtKeywords[i]));
        if AnsiCompareText(TempText, SelectStmtKeywords[i]) = 0 then
           begin
           Result := True;
           Inc(CurrentChar, Length(SelectStmtKeywords[i]));
           Break;
           end;
        end;
    end;

  begin
    if IsKeyword
       then NewToken.TokenType := ttKeyword
       else begin
            While (CurrentChar^ in IdentifierChar) do Inc(CurrentChar);
            NewToken.TokenType := ttIdentifier;
            end;
  end;

  Procedure ProcessBraceOpen;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessBraceClose;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessRoundOpen;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessRoundClose;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessStar;
  begin
  NewToken.TokenType := ttStar;
  Inc(CurrentChar);
  end;

  Procedure ProcessPlus;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessComma;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessMinus;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessPoint;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessSlash;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessColon;
  begin
  NewToken.TokenType := ttColon;
  Inc(CurrentChar);
  end;

  Procedure ProcessSemiColon;
  begin
  NewToken.TokenType := ttSemiColon;
  Inc(CurrentChar);
  end;

  Procedure ProcessLower;
  begin
  case (CurrentChar+1)^ of
    '=':
      begin
        Inc(CurrentChar, 2);
        NewToken.TokenType := ttLesserOrEqual;
      end;
    '>':
      begin
        Inc(CurrentChar, 2);
        NewToken.TokenType := ttNotEqual;
      end
  else
    begin
      Inc(CurrentChar);
      NewToken.TokenType := ttLesser;
    end;
  end;
  end;

  Procedure ProcessEqual;
  begin
  NewToken.TokenType := ttEqual;
  Inc(CurrentChar);
  end;

  Procedure ProcessGreater;
  begin
  If (CurrentChar+1)^ = '='
      then begin
      NewToken.TokenType := ttGreaterOrEqual;
      Inc(CurrentChar, 2);
      end
      else begin
      NewToken.TokenType := ttGreater;
      Inc(CurrentChar);
      end;
  end;

  Procedure ProcessSquareOpen;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessSquareClose;
  begin
  Inc(CurrentChar);
  end;

  Procedure ProcessSingleQuote;
  begin
  NewToken.TokenType := ttQuote;
  Inc(CurrentChar);
  end;

  Procedure ProcessDoubleQuote;
  begin
  NewToken.TokenType := ttQuote;
  Inc(CurrentChar);
  end;

  Procedure ProcessUnknown;
  begin
  NewToken.TokenType := ttUnknown;
  Inc(CurrentChar);
  end;



{ Main procedure "Tokenize" }
begin
FTokens.ClearAndFreeItems;
if FText = '' then Exit;
CurrentChar := PChar(FText);

while Not (CurrentChar^ = #0) do
 begin
 { Add a new token to the token list }
 NewToken := TSqlitePassSQLToken.Create;
 FTokens.Add(NewToken);
 TokenStart:= CurrentChar;

 case CurrentChar^ of
   #0:  ProcessEOF;
   #10: ProcessLF;
   #13: ProcessCR;
   #1..#9, #11, #12, #14..#32: ProcessSpace;
   '0'..'9': ProcessNumber;
   'A'..'Z', 'a'..'z', '_', '', '', '': ProcessIdentifier;
   '{': ProcessBraceOpen;
   '}': ProcessBraceClose;
   '(': ProcessRoundOpen;
   ')': ProcessRoundClose;
   '*': ProcessStar;
   '+': ProcessPlus;
   ',': ProcessComma;
   '-': ProcessMinus;
   '.': ProcessPoint;
   '/': ProcessSlash;
   ':': ProcessColon;
   ';': ProcessSemiColon;
   '<': ProcessLower;
   '=': ProcessEqual;
   '>': ProcessGreater;
   '[': ProcessSquareOpen;
   ']': ProcessSquareClose;
   '''': ProcessSingleQuote;
   '"': ProcessDoubleQuote;
   else ProcessUnknown;
   end;
   SetString(NewToken.Text, TokenStart, CurrentChar-TokenStart);
end;
First;
end;
(*
      begin
        Inc(P); Inc(I);
        if I >= TL then
        begin
          Done:= True;
          Break;
        end;
      end;
      if Done then Break;

      { Is there any text before ? }
      if (TextStart <> nil) and (P > TextStart) then
      begin
        L:= P - TextStart;
        { Yes, copy to buffer }
        OnFoundText( CopyBuffer(TextStart, L) );
      end else
      begin
        TextStart:= nil;
      end;
      { No }

      TagStart:= P;
      while Not (P^ in [ '>', #0]) do
      begin
        // Find string in tag
        if (P^ = '"') or (P^ = '''') then
        begin
          C:= P^;
          Inc(P); Inc(I); // Skip current char " or '

          // Skip until string end
          while Not (P^ in [C, #0]) do
          begin
            Inc(P);Inc(I);
          end;
        end;

        Inc(P);Inc(I);
        if I >= TL then
        begin
          Done:= True;
          Break;
        end;
      end;
      if Done then Break;

      { Copy this tag to buffer }
      L:= P - TagStart + 1;

      OnFoundTag( uppercase(CopyBuffer(TagStart, L )), CopyBuffer(TagStart, L ) ); //L505: added uppercase
      Inc(P); Inc(I);
      if I >= TL then Break;
    until (Done);
  end;
  {$IFDEF DEBUGLN_ON}debugln('FastHtmlParser Exec End');{$ENDIF}

end; *)


